/* Generated by Cython 0.29.32 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "name": "ms_peak_picker._c.interval_t_vector",
        "sources": [
            "src/ms_peak_picker/_c/interval_t_vector.pyx"
        ]
    },
    "module_name": "ms_peak_picker._c.interval_t_vector"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_32"
#define CYTHON_HEX_VERSION 0x001D20F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_HEX >= 0x07030900)
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__ms_peak_picker___c__interval_t_vector
#define __PYX_HAVE_API__ms_peak_picker___c__interval_t_vector
/* Early includes */
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "src/ms_peak_picker/_c/interval_t_vector.pyx",
  "stringsource",
  "type.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector;
struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__;
struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t;
struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector;

/* "ms_peak_picker/_c/interval_t_vector.pxd":3
 * 
 * 
 * cdef struct interval_t:             # <<<<<<<<<<<<<<
 *     size_t start
 *     size_t end
 */
struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t {
  size_t start;
  size_t end;
};

/* "ms_peak_picker/_c/interval_t_vector.pxd":12
 * 
 * 
 * cdef struct interval_t_vector:             # <<<<<<<<<<<<<<
 *     interval_t* v
 *     size_t used
 */
struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector {
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *v;
  size_t used;
  size_t size;
};

/* "ms_peak_picker/_c/interval_t_vector.pyx":14
 * from cpython.slice cimport PySlice_GetIndicesEx
 * 
 * cdef enum VectorStateEnum:             # <<<<<<<<<<<<<<
 *     should_free = 1
 * 
 */
enum __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_VectorStateEnum {
  __pyx_e_14ms_peak_picker_2_c_17interval_t_vector_should_free = 1
};

/* "ms_peak_picker/_c/interval_t_vector.pxd":29
 * cdef void print_interval_t_vector(interval_t_vector* vec) nogil
 * 
 * cdef class IntervalVector(object):             # <<<<<<<<<<<<<<
 *     cdef __cythonbufferdefaults__ = {'ndim' : 1, 'mode':'c'}
 * 
 */
struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector {
  PyObject_HEAD
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_vtab;
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *impl;
  int flags;
};


/* "ms_peak_picker/_c/interval_t_vector.pyx":294
 *         return self.size()
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */
struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ {
  PyObject_HEAD
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self;
  size_t __pyx_t_0;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
};



/* "ms_peak_picker/_c/interval_t_vector.pyx":159
 * @cython.final
 * @cython.freelist(512)
 * cdef class IntervalVector(object):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector {
  int (*allocate_storage)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *);
  int (*allocate_storage_with_size)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, size_t);
  int (*free_storage)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *);
  int (*get_should_free)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *);
  void (*set_should_free)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, int);
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *(*get_data)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *);
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *(*_create)(size_t);
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *(*wrap)(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *);
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t (*get)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, size_t);
  void (*set)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, size_t, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t);
  size_t (*size)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *);
  int (*cappend)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t);
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *(*_slice)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, PyObject *);
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, int __pyx_skip_dispatch);
  int (*append)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, PyObject *, int __pyx_skip_dispatch);
  int (*extend)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, PyObject *, int __pyx_skip_dispatch);
  int (*reserve)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, size_t, int __pyx_skip_dispatch);
  int (*fill)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t, int __pyx_skip_dispatch);
  PyObject *(*_to_python)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t, int __pyx_skip_dispatch);
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t (*_to_c)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_vtabptr_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector;
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_allocate_storage(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *);
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_allocate_storage_with_size(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, size_t);
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_free_storage(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *);
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_get_should_free(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *);
static void __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_set_should_free(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, int);
static struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_get_data(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *);
static struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__create(size_t);
static struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_wrap(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *);
static struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_get(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, size_t);
static void __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_set(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, size_t, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t);
static size_t __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_size(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *);
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_cappend(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t);
static struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__slice(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, PyObject *);
static struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_copy(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, int __pyx_skip_dispatch);
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_append(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, PyObject *, int __pyx_skip_dispatch);
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_extend(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, PyObject *, int __pyx_skip_dispatch);
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_reserve(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, size_t, int __pyx_skip_dispatch);
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_fill(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t, int __pyx_skip_dispatch);
static PyObject *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__to_python(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t, int __pyx_skip_dispatch);
static struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__to_c(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, PyObject *, int __pyx_skip_dispatch);

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t;
static PyObject* __pyx_convert__to_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t s);
/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__create(size_t __pyx_v_size); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_wrap(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *__pyx_v_vector); /* proto*/
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_allocate_storage_with_size(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, size_t __pyx_v_size); /* proto*/
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_allocate_storage(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self); /* proto*/
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_free_storage(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self); /* proto*/
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_get_should_free(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self); /* proto*/
static void __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_set_should_free(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, int __pyx_v_flag); /* proto*/
static struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_get_data(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self); /* proto*/
static struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_get(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, size_t __pyx_v_i); /* proto*/
static void __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_set(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, size_t __pyx_v_i, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_value); /* proto*/
static size_t __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_size(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self); /* proto*/
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_cappend(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_value); /* proto*/
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_append(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_extend(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_values, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_reserve(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, size_t __pyx_v_size, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_fill(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_copy(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__slice(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_slice_spec); /* proto*/
static PyObject *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__to_python(CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__to_c(CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'libc' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.slice' */

/* Module declarations from 'ms_peak_picker._c.interval_t_vector' */
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ = 0;
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_initialize_interval_vector_t(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *, size_t); /*proto*/
static struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_make_interval_t_vector_with_size(size_t); /*proto*/
static struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_make_interval_t_vector(void); /*proto*/
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector_resize(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *); /*proto*/
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector_append(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t); /*proto*/
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector_reserve(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *, size_t); /*proto*/
static void __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_free_interval_t_vector(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *); /*proto*/
static struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_from_tuple(PyObject *); /*proto*/
static PyObject *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_tuple_from_interval(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t); /*proto*/
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_compare_value_interval_t(void const *, void const *); /*proto*/
static struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_convert__from_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t(PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "ms_peak_picker._c.interval_t_vector"
extern int __pyx_module_is_main_ms_peak_picker___c__interval_t_vector;
int __pyx_module_is_main_ms_peak_picker___c__interval_t_vector = 0;

/* Implementation of 'ms_peak_picker._c.interval_t_vector' */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_ValueError;
static const char __pyx_k_end[] = "end";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_iter[] = "__iter__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_seed[] = "seed";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_members[] = "members";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_IntervalVector[] = "IntervalVector";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_IntervalVector___iter[] = "IntervalVector.__iter__";
static const char __pyx_k_self___class_____name___members[] = "{self.__class__.__name__}({members})";
static const char __pyx_k_Does_not_support_slice_assignmen[] = "Does not support slice-assignment yet";
static const char __pyx_k_Must_provide_a_Sequence_like_obj[] = "Must provide a Sequence-like object";
static const char __pyx_k_No_value_specified_for_struct_at[] = "No value specified for struct attribute 'start'";
static const char __pyx_k_ms_peak_picker__c_interval_t_vec[] = "ms_peak_picker._c.interval_t_vector";
static const char __pyx_k_self_impl_cannot_be_converted_to[] = "self.impl cannot be converted to a Python object for pickling";
static const char __pyx_k_No_value_specified_for_struct_at_2[] = "No value specified for struct attribute 'end'";
static PyObject *__pyx_kp_s_Does_not_support_slice_assignmen;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_n_s_IntervalVector;
static PyObject *__pyx_n_s_IntervalVector___iter;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_kp_s_Must_provide_a_Sequence_like_obj;
static PyObject *__pyx_kp_s_No_value_specified_for_struct_at;
static PyObject *__pyx_kp_s_No_value_specified_for_struct_at_2;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_iter;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_members;
static PyObject *__pyx_n_s_ms_peak_picker__c_interval_t_vec;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_seed;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_kp_s_self___class_____name___members;
static PyObject *__pyx_kp_s_self_impl_cannot_be_converted_to;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static int __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector___init__(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_seed); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_2append(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_4extend(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_values); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_6reserve(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, size_t __pyx_v_size); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_8fill(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_value); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_10copy(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self); /* proto */
static void __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_12__dealloc__(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_14__len__(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_16__iter__(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_19__getitem__(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static int __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_21__setitem__(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_23__repr__(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_25_to_python(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_value); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_27_to_c(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_29__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_31__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
/* Late includes */

/* "ms_peak_picker/_c/interval_t_vector.pyx":17
 *     should_free = 1
 * 
 * cdef interval_t interval_from_tuple(tuple pair):             # <<<<<<<<<<<<<<
 *     cdef:
 *         interval_t value
 */

static struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_from_tuple(PyObject *__pyx_v_pair) {
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_value;
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("interval_from_tuple", 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":21
 *         interval_t value
 * 
 *     value.start = pair[0]             # <<<<<<<<<<<<<<
 *     value.end = pair[1]
 *     return value
 */
  if (unlikely(__pyx_v_pair == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 21, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v_pair, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_value.start = __pyx_t_2;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":22
 * 
 *     value.start = pair[0]
 *     value.end = pair[1]             # <<<<<<<<<<<<<<
 *     return value
 * 
 */
  if (unlikely(__pyx_v_pair == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 22, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v_pair, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_value.end = __pyx_t_2;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":23
 *     value.start = pair[0]
 *     value.end = pair[1]
 *     return value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":17
 *     should_free = 1
 * 
 * cdef interval_t interval_from_tuple(tuple pair):             # <<<<<<<<<<<<<<
 *     cdef:
 *         interval_t value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("ms_peak_picker._c.interval_t_vector.interval_from_tuple", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":26
 * 
 * 
 * cdef tuple tuple_from_interval(interval_t interval):             # <<<<<<<<<<<<<<
 *     return (interval.start, interval.end)
 * 
 */

static PyObject *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_tuple_from_interval(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_interval) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("tuple_from_interval", 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":27
 * 
 * cdef tuple tuple_from_interval(interval_t interval):
 *     return (interval.start, interval.end)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_interval.start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_interval.end); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":26
 * 
 * 
 * cdef tuple tuple_from_interval(interval_t interval):             # <<<<<<<<<<<<<<
 *     return (interval.start, interval.end)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.tuple_from_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":30
 * 
 * 
 * cdef bint interval_overlaps(interval_t* self, interval_t* interval) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         bint cond
 */

static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_overlaps(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *__pyx_v_self, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *__pyx_v_interval) {
  int __pyx_v_cond;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":33
 *     cdef:
 *         bint cond
 *     cond = ((self.start <= interval.start and self.end > interval.start) or (             # <<<<<<<<<<<<<<
 *              self.start >= interval.start and self.end < interval.end) or (
 *              self.start >= interval.start and self.end > interval.end and self.start < interval.end) or (
 */
  __pyx_t_2 = ((__pyx_v_self->start <= __pyx_v_interval->start) != 0);
  if (!__pyx_t_2) {
    goto __pyx_L4_next_or;
  } else {
  }
  __pyx_t_2 = ((__pyx_v_self->end > __pyx_v_interval->start) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_L4_next_or:;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":34
 *         bint cond
 *     cond = ((self.start <= interval.start and self.end > interval.start) or (
 *              self.start >= interval.start and self.end < interval.end) or (             # <<<<<<<<<<<<<<
 *              self.start >= interval.start and self.end > interval.end and self.start < interval.end) or (
 *              self.start <= interval.start and self.end > interval.start) or (
 */
  __pyx_t_2 = ((__pyx_v_self->start >= __pyx_v_interval->start) != 0);
  if (!__pyx_t_2) {
    goto __pyx_L6_next_or;
  } else {
  }
  __pyx_t_2 = ((__pyx_v_self->end < __pyx_v_interval->end) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_L6_next_or:;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":35
 *     cond = ((self.start <= interval.start and self.end > interval.start) or (
 *              self.start >= interval.start and self.end < interval.end) or (
 *              self.start >= interval.start and self.end > interval.end and self.start < interval.end) or (             # <<<<<<<<<<<<<<
 *              self.start <= interval.start and self.end > interval.start) or (
 *              self.start < interval.end and self.end > interval.end))
 */
  __pyx_t_2 = ((__pyx_v_self->start >= __pyx_v_interval->start) != 0);
  if (!__pyx_t_2) {
    goto __pyx_L8_next_or;
  } else {
  }
  __pyx_t_2 = ((__pyx_v_self->end > __pyx_v_interval->end) != 0);
  if (!__pyx_t_2) {
    goto __pyx_L8_next_or;
  } else {
  }
  __pyx_t_2 = ((__pyx_v_self->start < __pyx_v_interval->end) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_L8_next_or:;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":36
 *              self.start >= interval.start and self.end < interval.end) or (
 *              self.start >= interval.start and self.end > interval.end and self.start < interval.end) or (
 *              self.start <= interval.start and self.end > interval.start) or (             # <<<<<<<<<<<<<<
 *              self.start < interval.end and self.end > interval.end))
 *     return cond
 */
  __pyx_t_2 = ((__pyx_v_self->start <= __pyx_v_interval->start) != 0);
  if (!__pyx_t_2) {
    goto __pyx_L11_next_or;
  } else {
  }
  __pyx_t_2 = ((__pyx_v_self->end > __pyx_v_interval->start) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_L11_next_or:;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":37
 *              self.start >= interval.start and self.end > interval.end and self.start < interval.end) or (
 *              self.start <= interval.start and self.end > interval.start) or (
 *              self.start < interval.end and self.end > interval.end))             # <<<<<<<<<<<<<<
 *     return cond
 * 
 */
  __pyx_t_2 = ((__pyx_v_self->start < __pyx_v_interval->end) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_self->end > __pyx_v_interval->end) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L3_bool_binop_done:;
  __pyx_v_cond = __pyx_t_1;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":38
 *              self.start <= interval.start and self.end > interval.start) or (
 *              self.start < interval.end and self.end > interval.end))
 *     return cond             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_cond;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":30
 * 
 * 
 * cdef bint interval_overlaps(interval_t* self, interval_t* interval) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         bint cond
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":41
 * 
 * 
 * cdef bint interval_contains(interval_t* self, size_t point) nogil:             # <<<<<<<<<<<<<<
 *     return self.start <= point < self.end
 * 
 */

static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_contains(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *__pyx_v_self, size_t __pyx_v_point) {
  int __pyx_r;
  int __pyx_t_1;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":42
 * 
 * cdef bint interval_contains(interval_t* self, size_t point) nogil:
 *     return self.start <= point < self.end             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = (__pyx_v_self->start <= __pyx_v_point);
  if (__pyx_t_1) {
    __pyx_t_1 = (__pyx_v_point < __pyx_v_self->end);
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":41
 * 
 * 
 * cdef bint interval_contains(interval_t* self, size_t point) nogil:             # <<<<<<<<<<<<<<
 *     return self.start <= point < self.end
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":45
 * 
 * 
 * cdef int compare_value_interval_t(const void* a, const void* b) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         interval_t* av
 */

static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_compare_value_interval_t(void const *__pyx_v_a, void const *__pyx_v_b) {
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *__pyx_v_av;
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *__pyx_v_bv;
  int __pyx_r;
  int __pyx_t_1;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":49
 *         interval_t* av
 *         interval_t* bv
 *     av = (<interval_t*>a)             # <<<<<<<<<<<<<<
 *     bv = (<interval_t*>b)
 *     if av.start < bv.start:
 */
  __pyx_v_av = ((struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *)__pyx_v_a);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":50
 *         interval_t* bv
 *     av = (<interval_t*>a)
 *     bv = (<interval_t*>b)             # <<<<<<<<<<<<<<
 *     if av.start < bv.start:
 *         return -1
 */
  __pyx_v_bv = ((struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *)__pyx_v_b);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":51
 *     av = (<interval_t*>a)
 *     bv = (<interval_t*>b)
 *     if av.start < bv.start:             # <<<<<<<<<<<<<<
 *         return -1
 *     elif av.start == bv.start:
 */
  __pyx_t_1 = ((__pyx_v_av->start < __pyx_v_bv->start) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":52
 *     bv = (<interval_t*>b)
 *     if av.start < bv.start:
 *         return -1             # <<<<<<<<<<<<<<
 *     elif av.start == bv.start:
 *         return 0
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/interval_t_vector.pyx":51
 *     av = (<interval_t*>a)
 *     bv = (<interval_t*>b)
 *     if av.start < bv.start:             # <<<<<<<<<<<<<<
 *         return -1
 *     elif av.start == bv.start:
 */
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":53
 *     if av.start < bv.start:
 *         return -1
 *     elif av.start == bv.start:             # <<<<<<<<<<<<<<
 *         return 0
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_av->start == __pyx_v_bv->start) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":54
 *         return -1
 *     elif av.start == bv.start:
 *         return 0             # <<<<<<<<<<<<<<
 *     else:
 *         return 1
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/interval_t_vector.pyx":53
 *     if av.start < bv.start:
 *         return -1
 *     elif av.start == bv.start:             # <<<<<<<<<<<<<<
 *         return 0
 *     else:
 */
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":56
 *         return 0
 *     else:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":45
 * 
 * 
 * cdef int compare_value_interval_t(const void* a, const void* b) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         interval_t* av
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":59
 * 
 * 
 * cdef int compare_value_interval_t_reverse(const void* a, const void* b) nogil:             # <<<<<<<<<<<<<<
 *     return -compare_value_interval_t(a, b)
 * 
 */

static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_compare_value_interval_t_reverse(void const *__pyx_v_a, void const *__pyx_v_b) {
  int __pyx_r;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":60
 * 
 * cdef int compare_value_interval_t_reverse(const void* a, const void* b) nogil:
 *     return -compare_value_interval_t(a, b)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (-__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_compare_value_interval_t(__pyx_v_a, __pyx_v_b));
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":59
 * 
 * 
 * cdef int compare_value_interval_t_reverse(const void* a, const void* b) nogil:             # <<<<<<<<<<<<<<
 *     return -compare_value_interval_t(a, b)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":72
 * 
 * 
 * cdef int initialize_interval_vector_t(interval_t_vector* vec, size_t size) nogil:             # <<<<<<<<<<<<<<
 *     vec.v = <interval_t*>malloc(sizeof(interval_t) * size)
 *     vec.size = size
 */

static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_initialize_interval_vector_t(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *__pyx_v_vec, size_t __pyx_v_size) {
  int __pyx_r;
  int __pyx_t_1;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":73
 * 
 * cdef int initialize_interval_vector_t(interval_t_vector* vec, size_t size) nogil:
 *     vec.v = <interval_t*>malloc(sizeof(interval_t) * size)             # <<<<<<<<<<<<<<
 *     vec.size = size
 *     vec.used = 0
 */
  __pyx_v_vec->v = ((struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *)malloc(((sizeof(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t)) * __pyx_v_size)));

  /* "ms_peak_picker/_c/interval_t_vector.pyx":74
 * cdef int initialize_interval_vector_t(interval_t_vector* vec, size_t size) nogil:
 *     vec.v = <interval_t*>malloc(sizeof(interval_t) * size)
 *     vec.size = size             # <<<<<<<<<<<<<<
 *     vec.used = 0
 *     if vec.v == NULL:
 */
  __pyx_v_vec->size = __pyx_v_size;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":75
 *     vec.v = <interval_t*>malloc(sizeof(interval_t) * size)
 *     vec.size = size
 *     vec.used = 0             # <<<<<<<<<<<<<<
 *     if vec.v == NULL:
 *         vec.size = 0
 */
  __pyx_v_vec->used = 0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":76
 *     vec.size = size
 *     vec.used = 0
 *     if vec.v == NULL:             # <<<<<<<<<<<<<<
 *         vec.size = 0
 *         return 1
 */
  __pyx_t_1 = ((__pyx_v_vec->v == NULL) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":77
 *     vec.used = 0
 *     if vec.v == NULL:
 *         vec.size = 0             # <<<<<<<<<<<<<<
 *         return 1
 *     return 0
 */
    __pyx_v_vec->size = 0;

    /* "ms_peak_picker/_c/interval_t_vector.pyx":78
 *     if vec.v == NULL:
 *         vec.size = 0
 *         return 1             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/interval_t_vector.pyx":76
 *     vec.size = size
 *     vec.used = 0
 *     if vec.v == NULL:             # <<<<<<<<<<<<<<
 *         vec.size = 0
 *         return 1
 */
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":79
 *         vec.size = 0
 *         return 1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":72
 * 
 * 
 * cdef int initialize_interval_vector_t(interval_t_vector* vec, size_t size) nogil:             # <<<<<<<<<<<<<<
 *     vec.v = <interval_t*>malloc(sizeof(interval_t) * size)
 *     vec.size = size
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":82
 * 
 * 
 * cdef interval_t_vector* make_interval_t_vector_with_size(size_t size) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         interval_t_vector* vec
 */

static struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_make_interval_t_vector_with_size(size_t __pyx_v_size) {
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *__pyx_v_vec;
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *__pyx_r;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":86
 *         interval_t_vector* vec
 * 
 *     vec = <interval_t_vector*>malloc(sizeof(interval_t_vector))             # <<<<<<<<<<<<<<
 *     initialize_interval_vector_t(vec, size)
 * 
 */
  __pyx_v_vec = ((struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *)malloc((sizeof(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector))));

  /* "ms_peak_picker/_c/interval_t_vector.pyx":87
 * 
 *     vec = <interval_t_vector*>malloc(sizeof(interval_t_vector))
 *     initialize_interval_vector_t(vec, size)             # <<<<<<<<<<<<<<
 * 
 *     return vec
 */
  (void)(__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_initialize_interval_vector_t(__pyx_v_vec, __pyx_v_size));

  /* "ms_peak_picker/_c/interval_t_vector.pyx":89
 *     initialize_interval_vector_t(vec, size)
 * 
 *     return vec             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_vec;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":82
 * 
 * 
 * cdef interval_t_vector* make_interval_t_vector_with_size(size_t size) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         interval_t_vector* vec
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":92
 * 
 * 
 * cdef interval_t_vector* make_interval_t_vector() nogil:             # <<<<<<<<<<<<<<
 *     return make_interval_t_vector_with_size(INITIAL_SIZE)
 * 
 */

static struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_make_interval_t_vector(void) {
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *__pyx_r;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":93
 * 
 * cdef interval_t_vector* make_interval_t_vector() nogil:
 *     return make_interval_t_vector_with_size(INITIAL_SIZE)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_make_interval_t_vector_with_size(4);
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":92
 * 
 * 
 * cdef interval_t_vector* make_interval_t_vector() nogil:             # <<<<<<<<<<<<<<
 *     return make_interval_t_vector_with_size(INITIAL_SIZE)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":96
 * 
 * 
 * cdef int interval_t_vector_resize(interval_t_vector* vec) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t new_size
 */

static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector_resize(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *__pyx_v_vec) {
  size_t __pyx_v_new_size;
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *__pyx_v_v;
  int __pyx_r;
  int __pyx_t_1;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":100
 *         size_t new_size
 *         interval_t* v
 *     new_size = vec.size * GROWTH_RATE             # <<<<<<<<<<<<<<
 *     v = <interval_t*>realloc(vec.v, sizeof(interval_t) * new_size)
 *     if v == NULL:
 */
  __pyx_v_new_size = (__pyx_v_vec->size * 2);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":101
 *         interval_t* v
 *     new_size = vec.size * GROWTH_RATE
 *     v = <interval_t*>realloc(vec.v, sizeof(interval_t) * new_size)             # <<<<<<<<<<<<<<
 *     if v == NULL:
 *         printf("interval_t_vector_resize returned -1\n")
 */
  __pyx_v_v = ((struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *)realloc(__pyx_v_vec->v, ((sizeof(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t)) * __pyx_v_new_size)));

  /* "ms_peak_picker/_c/interval_t_vector.pyx":102
 *     new_size = vec.size * GROWTH_RATE
 *     v = <interval_t*>realloc(vec.v, sizeof(interval_t) * new_size)
 *     if v == NULL:             # <<<<<<<<<<<<<<
 *         printf("interval_t_vector_resize returned -1\n")
 *         return -1
 */
  __pyx_t_1 = ((__pyx_v_v == NULL) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":103
 *     v = <interval_t*>realloc(vec.v, sizeof(interval_t) * new_size)
 *     if v == NULL:
 *         printf("interval_t_vector_resize returned -1\n")             # <<<<<<<<<<<<<<
 *         return -1
 *     vec.v = v
 */
    (void)(printf(((char const *)"interval_t_vector_resize returned -1\n")));

    /* "ms_peak_picker/_c/interval_t_vector.pyx":104
 *     if v == NULL:
 *         printf("interval_t_vector_resize returned -1\n")
 *         return -1             # <<<<<<<<<<<<<<
 *     vec.v = v
 *     vec.size = new_size
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/interval_t_vector.pyx":102
 *     new_size = vec.size * GROWTH_RATE
 *     v = <interval_t*>realloc(vec.v, sizeof(interval_t) * new_size)
 *     if v == NULL:             # <<<<<<<<<<<<<<
 *         printf("interval_t_vector_resize returned -1\n")
 *         return -1
 */
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":105
 *         printf("interval_t_vector_resize returned -1\n")
 *         return -1
 *     vec.v = v             # <<<<<<<<<<<<<<
 *     vec.size = new_size
 *     return 0
 */
  __pyx_v_vec->v = __pyx_v_v;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":106
 *         return -1
 *     vec.v = v
 *     vec.size = new_size             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  __pyx_v_vec->size = __pyx_v_new_size;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":107
 *     vec.v = v
 *     vec.size = new_size
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":96
 * 
 * 
 * cdef int interval_t_vector_resize(interval_t_vector* vec) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t new_size
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":110
 * 
 * 
 * cdef int interval_t_vector_append(interval_t_vector* vec, interval_t value) nogil:             # <<<<<<<<<<<<<<
 *     if (vec.used + 1) >= vec.size:
 *         interval_t_vector_resize(vec)
 */

static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector_append(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *__pyx_v_vec, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":111
 * 
 * cdef int interval_t_vector_append(interval_t_vector* vec, interval_t value) nogil:
 *     if (vec.used + 1) >= vec.size:             # <<<<<<<<<<<<<<
 *         interval_t_vector_resize(vec)
 *     vec.v[vec.used] = value
 */
  __pyx_t_1 = (((__pyx_v_vec->used + 1) >= __pyx_v_vec->size) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":112
 * cdef int interval_t_vector_append(interval_t_vector* vec, interval_t value) nogil:
 *     if (vec.used + 1) >= vec.size:
 *         interval_t_vector_resize(vec)             # <<<<<<<<<<<<<<
 *     vec.v[vec.used] = value
 *     vec.used += 1
 */
    (void)(__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector_resize(__pyx_v_vec));

    /* "ms_peak_picker/_c/interval_t_vector.pyx":111
 * 
 * cdef int interval_t_vector_append(interval_t_vector* vec, interval_t value) nogil:
 *     if (vec.used + 1) >= vec.size:             # <<<<<<<<<<<<<<
 *         interval_t_vector_resize(vec)
 *     vec.v[vec.used] = value
 */
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":113
 *     if (vec.used + 1) >= vec.size:
 *         interval_t_vector_resize(vec)
 *     vec.v[vec.used] = value             # <<<<<<<<<<<<<<
 *     vec.used += 1
 *     return 0
 */
  (__pyx_v_vec->v[__pyx_v_vec->used]) = __pyx_v_value;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":114
 *         interval_t_vector_resize(vec)
 *     vec.v[vec.used] = value
 *     vec.used += 1             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  __pyx_v_vec->used = (__pyx_v_vec->used + 1);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":115
 *     vec.v[vec.used] = value
 *     vec.used += 1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":110
 * 
 * 
 * cdef int interval_t_vector_append(interval_t_vector* vec, interval_t value) nogil:             # <<<<<<<<<<<<<<
 *     if (vec.used + 1) >= vec.size:
 *         interval_t_vector_resize(vec)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":118
 * 
 * 
 * cdef void free_interval_t_vector(interval_t_vector* vec) nogil:             # <<<<<<<<<<<<<<
 *     free(vec.v)
 *     free(vec)
 */

static void __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_free_interval_t_vector(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *__pyx_v_vec) {

  /* "ms_peak_picker/_c/interval_t_vector.pyx":119
 * 
 * cdef void free_interval_t_vector(interval_t_vector* vec) nogil:
 *     free(vec.v)             # <<<<<<<<<<<<<<
 *     free(vec)
 * 
 */
  free(__pyx_v_vec->v);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":120
 * cdef void free_interval_t_vector(interval_t_vector* vec) nogil:
 *     free(vec.v)
 *     free(vec)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  free(__pyx_v_vec);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":118
 * 
 * 
 * cdef void free_interval_t_vector(interval_t_vector* vec) nogil:             # <<<<<<<<<<<<<<
 *     free(vec.v)
 *     free(vec)
 */

  /* function exit code */
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":123
 * 
 * 
 * cdef void print_interval_t_vector(interval_t_vector* vec) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static void __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_print_interval_t_vector(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *__pyx_v_vec) {
  size_t __pyx_v_i;
  int __pyx_t_1;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":126
 *     cdef:
 *         size_t i
 *     i = 0             # <<<<<<<<<<<<<<
 *     printf("[")
 *     while i < vec.used:
 */
  __pyx_v_i = 0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":127
 *         size_t i
 *     i = 0
 *     printf("[")             # <<<<<<<<<<<<<<
 *     while i < vec.used:
 *         printf("%0.6f", vec.v[i])
 */
  (void)(printf(((char const *)"[")));

  /* "ms_peak_picker/_c/interval_t_vector.pyx":128
 *     i = 0
 *     printf("[")
 *     while i < vec.used:             # <<<<<<<<<<<<<<
 *         printf("%0.6f", vec.v[i])
 *         if i != (vec.used - 1):
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_i < __pyx_v_vec->used) != 0);
    if (!__pyx_t_1) break;

    /* "ms_peak_picker/_c/interval_t_vector.pyx":129
 *     printf("[")
 *     while i < vec.used:
 *         printf("%0.6f", vec.v[i])             # <<<<<<<<<<<<<<
 *         if i != (vec.used - 1):
 *             printf(", ")
 */
    (void)(printf(((char const *)"%0.6f"), (__pyx_v_vec->v[__pyx_v_i])));

    /* "ms_peak_picker/_c/interval_t_vector.pyx":130
 *     while i < vec.used:
 *         printf("%0.6f", vec.v[i])
 *         if i != (vec.used - 1):             # <<<<<<<<<<<<<<
 *             printf(", ")
 *         i += 1
 */
    __pyx_t_1 = ((__pyx_v_i != (__pyx_v_vec->used - 1)) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/interval_t_vector.pyx":131
 *         printf("%0.6f", vec.v[i])
 *         if i != (vec.used - 1):
 *             printf(", ")             # <<<<<<<<<<<<<<
 *         i += 1
 *     printf("]\n")
 */
      (void)(printf(((char const *)", ")));

      /* "ms_peak_picker/_c/interval_t_vector.pyx":130
 *     while i < vec.used:
 *         printf("%0.6f", vec.v[i])
 *         if i != (vec.used - 1):             # <<<<<<<<<<<<<<
 *             printf(", ")
 *         i += 1
 */
    }

    /* "ms_peak_picker/_c/interval_t_vector.pyx":132
 *         if i != (vec.used - 1):
 *             printf(", ")
 *         i += 1             # <<<<<<<<<<<<<<
 *     printf("]\n")
 * 
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":133
 *             printf(", ")
 *         i += 1
 *     printf("]\n")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (void)(printf(((char const *)"]\n")));

  /* "ms_peak_picker/_c/interval_t_vector.pyx":123
 * 
 * 
 * cdef void print_interval_t_vector(interval_t_vector* vec) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":136
 * 
 * 
 * cdef void interval_t_vector_reset(interval_t_vector* vec) nogil:             # <<<<<<<<<<<<<<
 *     vec.used = 0
 * 
 */

static void __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector_reset(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *__pyx_v_vec) {

  /* "ms_peak_picker/_c/interval_t_vector.pyx":137
 * 
 * cdef void interval_t_vector_reset(interval_t_vector* vec) nogil:
 *     vec.used = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_vec->used = 0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":136
 * 
 * 
 * cdef void interval_t_vector_reset(interval_t_vector* vec) nogil:             # <<<<<<<<<<<<<<
 *     vec.used = 0
 * 
 */

  /* function exit code */
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":140
 * 
 * 
 * cdef int interval_t_vector_reserve(interval_t_vector* vec, size_t new_size) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         interval_t* v
 */

static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector_reserve(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *__pyx_v_vec, size_t __pyx_v_new_size) {
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *__pyx_v_v;
  int __pyx_r;
  int __pyx_t_1;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":143
 *     cdef:
 *         interval_t* v
 *     v = <interval_t*>realloc(vec.v, sizeof(interval_t) * new_size)             # <<<<<<<<<<<<<<
 *     if v == NULL:
 *         printf("interval_t_vector_resize returned -1\n")
 */
  __pyx_v_v = ((struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *)realloc(__pyx_v_vec->v, ((sizeof(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t)) * __pyx_v_new_size)));

  /* "ms_peak_picker/_c/interval_t_vector.pyx":144
 *         interval_t* v
 *     v = <interval_t*>realloc(vec.v, sizeof(interval_t) * new_size)
 *     if v == NULL:             # <<<<<<<<<<<<<<
 *         printf("interval_t_vector_resize returned -1\n")
 *         return -1
 */
  __pyx_t_1 = ((__pyx_v_v == NULL) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":145
 *     v = <interval_t*>realloc(vec.v, sizeof(interval_t) * new_size)
 *     if v == NULL:
 *         printf("interval_t_vector_resize returned -1\n")             # <<<<<<<<<<<<<<
 *         return -1
 *     vec.v = v
 */
    (void)(printf(((char const *)"interval_t_vector_resize returned -1\n")));

    /* "ms_peak_picker/_c/interval_t_vector.pyx":146
 *     if v == NULL:
 *         printf("interval_t_vector_resize returned -1\n")
 *         return -1             # <<<<<<<<<<<<<<
 *     vec.v = v
 *     vec.size = new_size
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/interval_t_vector.pyx":144
 *         interval_t* v
 *     v = <interval_t*>realloc(vec.v, sizeof(interval_t) * new_size)
 *     if v == NULL:             # <<<<<<<<<<<<<<
 *         printf("interval_t_vector_resize returned -1\n")
 *         return -1
 */
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":147
 *         printf("interval_t_vector_resize returned -1\n")
 *         return -1
 *     vec.v = v             # <<<<<<<<<<<<<<
 *     vec.size = new_size
 *     if new_size > vec.used:
 */
  __pyx_v_vec->v = __pyx_v_v;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":148
 *         return -1
 *     vec.v = v
 *     vec.size = new_size             # <<<<<<<<<<<<<<
 *     if new_size > vec.used:
 *         vec.used = new_size
 */
  __pyx_v_vec->size = __pyx_v_new_size;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":149
 *     vec.v = v
 *     vec.size = new_size
 *     if new_size > vec.used:             # <<<<<<<<<<<<<<
 *         vec.used = new_size
 *     return 0
 */
  __pyx_t_1 = ((__pyx_v_new_size > __pyx_v_vec->used) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":150
 *     vec.size = new_size
 *     if new_size > vec.used:
 *         vec.used = new_size             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    __pyx_v_vec->used = __pyx_v_new_size;

    /* "ms_peak_picker/_c/interval_t_vector.pyx":149
 *     vec.v = v
 *     vec.size = new_size
 *     if new_size > vec.used:             # <<<<<<<<<<<<<<
 *         vec.used = new_size
 *     return 0
 */
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":151
 *     if new_size > vec.used:
 *         vec.used = new_size
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":140
 * 
 * 
 * cdef int interval_t_vector_reserve(interval_t_vector* vec, size_t new_size) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         interval_t* v
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":162
 * 
 *     @staticmethod
 *     cdef IntervalVector _create(size_t size):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IntervalVector self
 */

static struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__create(size_t __pyx_v_size) {
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":165
 *         cdef:
 *             IntervalVector self
 *         self = IntervalVector.__new__(IntervalVector)             # <<<<<<<<<<<<<<
 *         self.flags = 0
 *         self.allocate_storage_with_size(size)
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector(((PyTypeObject *)__pyx_ptype_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_self = ((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":166
 *             IntervalVector self
 *         self = IntervalVector.__new__(IntervalVector)
 *         self.flags = 0             # <<<<<<<<<<<<<<
 *         self.allocate_storage_with_size(size)
 *         return self
 */
  __pyx_v_self->flags = 0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":167
 *         self = IntervalVector.__new__(IntervalVector)
 *         self.flags = 0
 *         self.allocate_storage_with_size(size)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  (void)(__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_allocate_storage_with_size(__pyx_v_self, __pyx_v_size));

  /* "ms_peak_picker/_c/interval_t_vector.pyx":168
 *         self.flags = 0
 *         self.allocate_storage_with_size(size)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":162
 * 
 *     @staticmethod
 *     cdef IntervalVector _create(size_t size):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IntervalVector self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":171
 * 
 *     @staticmethod
 *     cdef IntervalVector wrap(interval_t_vector* vector):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IntervalVector self
 */

static struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_wrap(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *__pyx_v_vector) {
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wrap", 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":174
 *         cdef:
 *             IntervalVector self
 *         self = IntervalVector.__new__(IntervalVector)             # <<<<<<<<<<<<<<
 *         self.flags = 0
 *         self.impl = vector
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector(((PyTypeObject *)__pyx_ptype_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_self = ((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":175
 *             IntervalVector self
 *         self = IntervalVector.__new__(IntervalVector)
 *         self.flags = 0             # <<<<<<<<<<<<<<
 *         self.impl = vector
 *         self.set_should_free(False)
 */
  __pyx_v_self->flags = 0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":176
 *         self = IntervalVector.__new__(IntervalVector)
 *         self.flags = 0
 *         self.impl = vector             # <<<<<<<<<<<<<<
 *         self.set_should_free(False)
 *         return self
 */
  __pyx_v_self->impl = __pyx_v_vector;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":177
 *         self.flags = 0
 *         self.impl = vector
 *         self.set_should_free(False)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_set_should_free(__pyx_v_self, 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":178
 *         self.impl = vector
 *         self.set_should_free(False)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, seed=None):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":171
 * 
 *     @staticmethod
 *     cdef IntervalVector wrap(interval_t_vector* vector):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IntervalVector self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":180
 *         return self
 * 
 *     def __init__(self, seed=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t n
 */

/* Python wrapper */
static int __pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_seed = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_seed,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seed);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 180, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_seed = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 180, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector___init__(((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_v_self), __pyx_v_seed);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector___init__(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_seed) {
  size_t __pyx_v_n;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":183
 *         cdef:
 *             size_t n
 *         self.flags = 0             # <<<<<<<<<<<<<<
 *         if seed is not None:
 *             if not PySequence_Check(seed):
 */
  __pyx_v_self->flags = 0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":184
 *             size_t n
 *         self.flags = 0
 *         if seed is not None:             # <<<<<<<<<<<<<<
 *             if not PySequence_Check(seed):
 *                 raise TypeError("Must provide a Sequence-like object")
 */
  __pyx_t_1 = (__pyx_v_seed != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":185
 *         self.flags = 0
 *         if seed is not None:
 *             if not PySequence_Check(seed):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Must provide a Sequence-like object")
 *             n = len(seed)
 */
    __pyx_t_2 = ((!(PySequence_Check(__pyx_v_seed) != 0)) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "ms_peak_picker/_c/interval_t_vector.pyx":186
 *         if seed is not None:
 *             if not PySequence_Check(seed):
 *                 raise TypeError("Must provide a Sequence-like object")             # <<<<<<<<<<<<<<
 *             n = len(seed)
 *             self.allocate_storage_with_size(n)
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 186, __pyx_L1_error)

      /* "ms_peak_picker/_c/interval_t_vector.pyx":185
 *         self.flags = 0
 *         if seed is not None:
 *             if not PySequence_Check(seed):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Must provide a Sequence-like object")
 *             n = len(seed)
 */
    }

    /* "ms_peak_picker/_c/interval_t_vector.pyx":187
 *             if not PySequence_Check(seed):
 *                 raise TypeError("Must provide a Sequence-like object")
 *             n = len(seed)             # <<<<<<<<<<<<<<
 *             self.allocate_storage_with_size(n)
 *             self.extend(seed)
 */
    __pyx_t_4 = PyObject_Length(__pyx_v_seed); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 187, __pyx_L1_error)
    __pyx_v_n = __pyx_t_4;

    /* "ms_peak_picker/_c/interval_t_vector.pyx":188
 *                 raise TypeError("Must provide a Sequence-like object")
 *             n = len(seed)
 *             self.allocate_storage_with_size(n)             # <<<<<<<<<<<<<<
 *             self.extend(seed)
 *         else:
 */
    (void)(__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_allocate_storage_with_size(__pyx_v_self, __pyx_v_n));

    /* "ms_peak_picker/_c/interval_t_vector.pyx":189
 *             n = len(seed)
 *             self.allocate_storage_with_size(n)
 *             self.extend(seed)             # <<<<<<<<<<<<<<
 *         else:
 *             self.allocate_storage()
 */
    __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_extend(__pyx_v_self, __pyx_v_seed, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 189, __pyx_L1_error)

    /* "ms_peak_picker/_c/interval_t_vector.pyx":184
 *             size_t n
 *         self.flags = 0
 *         if seed is not None:             # <<<<<<<<<<<<<<
 *             if not PySequence_Check(seed):
 *                 raise TypeError("Must provide a Sequence-like object")
 */
    goto __pyx_L3;
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":191
 *             self.extend(seed)
 *         else:
 *             self.allocate_storage()             # <<<<<<<<<<<<<<
 * 
 *     cdef int allocate_storage_with_size(self, size_t size) nogil:
 */
  /*else*/ {
    (void)(__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_allocate_storage(__pyx_v_self));
  }
  __pyx_L3:;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":180
 *         return self
 * 
 *     def __init__(self, seed=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t n
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":193
 *             self.allocate_storage()
 * 
 *     cdef int allocate_storage_with_size(self, size_t size) nogil:             # <<<<<<<<<<<<<<
 *         if self.impl != NULL:
 *             if self.flags & VectorStateEnum.should_free:
 */

static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_allocate_storage_with_size(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, size_t __pyx_v_size) {
  int __pyx_r;
  int __pyx_t_1;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":194
 * 
 *     cdef int allocate_storage_with_size(self, size_t size) nogil:
 *         if self.impl != NULL:             # <<<<<<<<<<<<<<
 *             if self.flags & VectorStateEnum.should_free:
 *                 free_interval_t_vector(self.impl)
 */
  __pyx_t_1 = ((__pyx_v_self->impl != NULL) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":195
 *     cdef int allocate_storage_with_size(self, size_t size) nogil:
 *         if self.impl != NULL:
 *             if self.flags & VectorStateEnum.should_free:             # <<<<<<<<<<<<<<
 *                 free_interval_t_vector(self.impl)
 *         self.impl = make_interval_t_vector_with_size(size)
 */
    __pyx_t_1 = ((__pyx_v_self->flags & __pyx_e_14ms_peak_picker_2_c_17interval_t_vector_should_free) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/interval_t_vector.pyx":196
 *         if self.impl != NULL:
 *             if self.flags & VectorStateEnum.should_free:
 *                 free_interval_t_vector(self.impl)             # <<<<<<<<<<<<<<
 *         self.impl = make_interval_t_vector_with_size(size)
 *         self.flags |= VectorStateEnum.should_free
 */
      __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_free_interval_t_vector(__pyx_v_self->impl);

      /* "ms_peak_picker/_c/interval_t_vector.pyx":195
 *     cdef int allocate_storage_with_size(self, size_t size) nogil:
 *         if self.impl != NULL:
 *             if self.flags & VectorStateEnum.should_free:             # <<<<<<<<<<<<<<
 *                 free_interval_t_vector(self.impl)
 *         self.impl = make_interval_t_vector_with_size(size)
 */
    }

    /* "ms_peak_picker/_c/interval_t_vector.pyx":194
 * 
 *     cdef int allocate_storage_with_size(self, size_t size) nogil:
 *         if self.impl != NULL:             # <<<<<<<<<<<<<<
 *             if self.flags & VectorStateEnum.should_free:
 *                 free_interval_t_vector(self.impl)
 */
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":197
 *             if self.flags & VectorStateEnum.should_free:
 *                 free_interval_t_vector(self.impl)
 *         self.impl = make_interval_t_vector_with_size(size)             # <<<<<<<<<<<<<<
 *         self.flags |= VectorStateEnum.should_free
 *         return self.impl == NULL
 */
  __pyx_v_self->impl = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_make_interval_t_vector_with_size(__pyx_v_size);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":198
 *                 free_interval_t_vector(self.impl)
 *         self.impl = make_interval_t_vector_with_size(size)
 *         self.flags |= VectorStateEnum.should_free             # <<<<<<<<<<<<<<
 *         return self.impl == NULL
 * 
 */
  __pyx_v_self->flags = (__pyx_v_self->flags | __pyx_e_14ms_peak_picker_2_c_17interval_t_vector_should_free);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":199
 *         self.impl = make_interval_t_vector_with_size(size)
 *         self.flags |= VectorStateEnum.should_free
 *         return self.impl == NULL             # <<<<<<<<<<<<<<
 * 
 *     cdef int allocate_storage(self) nogil:
 */
  __pyx_r = (__pyx_v_self->impl == NULL);
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":193
 *             self.allocate_storage()
 * 
 *     cdef int allocate_storage_with_size(self, size_t size) nogil:             # <<<<<<<<<<<<<<
 *         if self.impl != NULL:
 *             if self.flags & VectorStateEnum.should_free:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":201
 *         return self.impl == NULL
 * 
 *     cdef int allocate_storage(self) nogil:             # <<<<<<<<<<<<<<
 *         if self.impl != NULL:
 *             if self.flags & VectorStateEnum.should_free:
 */

static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_allocate_storage(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self) {
  int __pyx_r;
  int __pyx_t_1;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":202
 * 
 *     cdef int allocate_storage(self) nogil:
 *         if self.impl != NULL:             # <<<<<<<<<<<<<<
 *             if self.flags & VectorStateEnum.should_free:
 *                 free_interval_t_vector(self.impl)
 */
  __pyx_t_1 = ((__pyx_v_self->impl != NULL) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":203
 *     cdef int allocate_storage(self) nogil:
 *         if self.impl != NULL:
 *             if self.flags & VectorStateEnum.should_free:             # <<<<<<<<<<<<<<
 *                 free_interval_t_vector(self.impl)
 *         self.impl = make_interval_t_vector()
 */
    __pyx_t_1 = ((__pyx_v_self->flags & __pyx_e_14ms_peak_picker_2_c_17interval_t_vector_should_free) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/interval_t_vector.pyx":204
 *         if self.impl != NULL:
 *             if self.flags & VectorStateEnum.should_free:
 *                 free_interval_t_vector(self.impl)             # <<<<<<<<<<<<<<
 *         self.impl = make_interval_t_vector()
 *         self.flags |= VectorStateEnum.should_free
 */
      __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_free_interval_t_vector(__pyx_v_self->impl);

      /* "ms_peak_picker/_c/interval_t_vector.pyx":203
 *     cdef int allocate_storage(self) nogil:
 *         if self.impl != NULL:
 *             if self.flags & VectorStateEnum.should_free:             # <<<<<<<<<<<<<<
 *                 free_interval_t_vector(self.impl)
 *         self.impl = make_interval_t_vector()
 */
    }

    /* "ms_peak_picker/_c/interval_t_vector.pyx":202
 * 
 *     cdef int allocate_storage(self) nogil:
 *         if self.impl != NULL:             # <<<<<<<<<<<<<<
 *             if self.flags & VectorStateEnum.should_free:
 *                 free_interval_t_vector(self.impl)
 */
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":205
 *             if self.flags & VectorStateEnum.should_free:
 *                 free_interval_t_vector(self.impl)
 *         self.impl = make_interval_t_vector()             # <<<<<<<<<<<<<<
 *         self.flags |= VectorStateEnum.should_free
 *         return self.impl == NULL
 */
  __pyx_v_self->impl = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_make_interval_t_vector();

  /* "ms_peak_picker/_c/interval_t_vector.pyx":206
 *                 free_interval_t_vector(self.impl)
 *         self.impl = make_interval_t_vector()
 *         self.flags |= VectorStateEnum.should_free             # <<<<<<<<<<<<<<
 *         return self.impl == NULL
 * 
 */
  __pyx_v_self->flags = (__pyx_v_self->flags | __pyx_e_14ms_peak_picker_2_c_17interval_t_vector_should_free);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":207
 *         self.impl = make_interval_t_vector()
 *         self.flags |= VectorStateEnum.should_free
 *         return self.impl == NULL             # <<<<<<<<<<<<<<
 * 
 *     cdef int free_storage(self) nogil:
 */
  __pyx_r = (__pyx_v_self->impl == NULL);
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":201
 *         return self.impl == NULL
 * 
 *     cdef int allocate_storage(self) nogil:             # <<<<<<<<<<<<<<
 *         if self.impl != NULL:
 *             if self.flags & VectorStateEnum.should_free:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":209
 *         return self.impl == NULL
 * 
 *     cdef int free_storage(self) nogil:             # <<<<<<<<<<<<<<
 *         free_interval_t_vector(self.impl)
 * 
 */

static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_free_storage(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self) {
  int __pyx_r;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":210
 * 
 *     cdef int free_storage(self) nogil:
 *         free_interval_t_vector(self.impl)             # <<<<<<<<<<<<<<
 * 
 *     cdef bint get_should_free(self) nogil:
 */
  __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_free_interval_t_vector(__pyx_v_self->impl);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":209
 *         return self.impl == NULL
 * 
 *     cdef int free_storage(self) nogil:             # <<<<<<<<<<<<<<
 *         free_interval_t_vector(self.impl)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":212
 *         free_interval_t_vector(self.impl)
 * 
 *     cdef bint get_should_free(self) nogil:             # <<<<<<<<<<<<<<
 *         return self.flags & VectorStateEnum.should_free
 * 
 */

static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_get_should_free(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self) {
  int __pyx_r;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":213
 * 
 *     cdef bint get_should_free(self) nogil:
 *         return self.flags & VectorStateEnum.should_free             # <<<<<<<<<<<<<<
 * 
 *     cdef void set_should_free(self, bint flag) nogil:
 */
  __pyx_r = (__pyx_v_self->flags & __pyx_e_14ms_peak_picker_2_c_17interval_t_vector_should_free);
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":212
 *         free_interval_t_vector(self.impl)
 * 
 *     cdef bint get_should_free(self) nogil:             # <<<<<<<<<<<<<<
 *         return self.flags & VectorStateEnum.should_free
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":215
 *         return self.flags & VectorStateEnum.should_free
 * 
 *     cdef void set_should_free(self, bint flag) nogil:             # <<<<<<<<<<<<<<
 *         self.flags &= VectorStateEnum.should_free * flag
 * 
 */

static void __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_set_should_free(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, int __pyx_v_flag) {

  /* "ms_peak_picker/_c/interval_t_vector.pyx":216
 * 
 *     cdef void set_should_free(self, bint flag) nogil:
 *         self.flags &= VectorStateEnum.should_free * flag             # <<<<<<<<<<<<<<
 * 
 *     cdef interval_t* get_data(self) nogil:
 */
  __pyx_v_self->flags = (__pyx_v_self->flags & (__pyx_e_14ms_peak_picker_2_c_17interval_t_vector_should_free * __pyx_v_flag));

  /* "ms_peak_picker/_c/interval_t_vector.pyx":215
 *         return self.flags & VectorStateEnum.should_free
 * 
 *     cdef void set_should_free(self, bint flag) nogil:             # <<<<<<<<<<<<<<
 *         self.flags &= VectorStateEnum.should_free * flag
 * 
 */

  /* function exit code */
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":218
 *         self.flags &= VectorStateEnum.should_free * flag
 * 
 *     cdef interval_t* get_data(self) nogil:             # <<<<<<<<<<<<<<
 *         return self.impl.v
 * 
 */

static struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_get_data(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self) {
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *__pyx_r;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":219
 * 
 *     cdef interval_t* get_data(self) nogil:
 *         return self.impl.v             # <<<<<<<<<<<<<<
 * 
 *     cdef interval_t get(self, size_t i) nogil:
 */
  __pyx_r = __pyx_v_self->impl->v;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":218
 *         self.flags &= VectorStateEnum.should_free * flag
 * 
 *     cdef interval_t* get_data(self) nogil:             # <<<<<<<<<<<<<<
 *         return self.impl.v
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":221
 *         return self.impl.v
 * 
 *     cdef interval_t get(self, size_t i) nogil:             # <<<<<<<<<<<<<<
 *         return self.impl.v[i]
 * 
 */

static struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_get(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, size_t __pyx_v_i) {
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_r;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":222
 * 
 *     cdef interval_t get(self, size_t i) nogil:
 *         return self.impl.v[i]             # <<<<<<<<<<<<<<
 * 
 *     cdef void set(self, size_t i, interval_t value) nogil:
 */
  __pyx_r = (__pyx_v_self->impl->v[__pyx_v_i]);
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":221
 *         return self.impl.v
 * 
 *     cdef interval_t get(self, size_t i) nogil:             # <<<<<<<<<<<<<<
 *         return self.impl.v[i]
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":224
 *         return self.impl.v[i]
 * 
 *     cdef void set(self, size_t i, interval_t value) nogil:             # <<<<<<<<<<<<<<
 *         self.impl.v[i] = value
 * 
 */

static void __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_set(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, size_t __pyx_v_i, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_value) {

  /* "ms_peak_picker/_c/interval_t_vector.pyx":225
 * 
 *     cdef void set(self, size_t i, interval_t value) nogil:
 *         self.impl.v[i] = value             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t size(self) nogil:
 */
  (__pyx_v_self->impl->v[__pyx_v_i]) = __pyx_v_value;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":224
 *         return self.impl.v[i]
 * 
 *     cdef void set(self, size_t i, interval_t value) nogil:             # <<<<<<<<<<<<<<
 *         self.impl.v[i] = value
 * 
 */

  /* function exit code */
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":227
 *         self.impl.v[i] = value
 * 
 *     cdef size_t size(self) nogil:             # <<<<<<<<<<<<<<
 *         return self.impl.used
 * 
 */

static size_t __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_size(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self) {
  size_t __pyx_r;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":228
 * 
 *     cdef size_t size(self) nogil:
 *         return self.impl.used             # <<<<<<<<<<<<<<
 * 
 *     cdef int cappend(self, interval_t value) nogil:
 */
  __pyx_r = __pyx_v_self->impl->used;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":227
 *         self.impl.v[i] = value
 * 
 *     cdef size_t size(self) nogil:             # <<<<<<<<<<<<<<
 *         return self.impl.used
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":230
 *         return self.impl.used
 * 
 *     cdef int cappend(self, interval_t value) nogil:             # <<<<<<<<<<<<<<
 *         return interval_t_vector_append(self.impl, value)
 * 
 */

static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_cappend(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_value) {
  int __pyx_r;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":231
 * 
 *     cdef int cappend(self, interval_t value) nogil:
 *         return interval_t_vector_append(self.impl, value)             # <<<<<<<<<<<<<<
 * 
 *     cpdef int append(self, tuple value) except *:
 */
  __pyx_r = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector_append(__pyx_v_self->impl, __pyx_v_value);
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":230
 *         return self.impl.used
 * 
 *     cdef int cappend(self, interval_t value) nogil:             # <<<<<<<<<<<<<<
 *         return interval_t_vector_append(self.impl, value)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":233
 *         return interval_t_vector_append(self.impl, value)
 * 
 *     cpdef int append(self, tuple value) except *:             # <<<<<<<<<<<<<<
 *         cdef:
 *             interval_t cvalue
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_3append(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_append(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_cvalue;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("append", 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":236
 *         cdef:
 *             interval_t cvalue
 *         cvalue = self._to_c(value)             # <<<<<<<<<<<<<<
 *         return self.cappend(cvalue)
 * 
 */
  __pyx_t_1 = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__to_c(__pyx_v_self, __pyx_v_value, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 236, __pyx_L1_error)
  __pyx_v_cvalue = __pyx_t_1;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":237
 *             interval_t cvalue
 *         cvalue = self._to_c(value)
 *         return self.cappend(cvalue)             # <<<<<<<<<<<<<<
 * 
 *     cpdef int extend(self, object values) except *:
 */
  __pyx_r = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_cappend(__pyx_v_self, __pyx_v_cvalue);
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":233
 *         return interval_t_vector_append(self.impl, value)
 * 
 *     cpdef int append(self, tuple value) except *:             # <<<<<<<<<<<<<<
 *         cdef:
 *             interval_t cvalue
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_3append(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_2append[] = "IntervalVector.append(self, tuple value) -> int";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_3append(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyTuple_Type), 1, "value", 1))) __PYX_ERR(0, 233, __pyx_L1_error)
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_2append(((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_2append(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("append", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_append(__pyx_v_self, __pyx_v_value, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 233, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":239
 *         return self.cappend(cvalue)
 * 
 *     cpdef int extend(self, object values) except *:             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_5extend(PyObject *__pyx_v_self, PyObject *__pyx_v_values); /*proto*/
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_extend(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_values, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  PyObject *__pyx_v_fast_seq = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("extend", 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":243
 *             size_t i, n
 *             object fast_seq
 *         if not PySequence_Check(values):             # <<<<<<<<<<<<<<
 *             raise TypeError("Must provide a Sequence-like object")
 * 
 */
  __pyx_t_1 = ((!(PySequence_Check(__pyx_v_values) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":244
 *             object fast_seq
 *         if not PySequence_Check(values):
 *             raise TypeError("Must provide a Sequence-like object")             # <<<<<<<<<<<<<<
 * 
 *         fast_seq = PySequence_Fast(values, "Must provide a Sequence-like object")
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 244, __pyx_L1_error)

    /* "ms_peak_picker/_c/interval_t_vector.pyx":243
 *             size_t i, n
 *             object fast_seq
 *         if not PySequence_Check(values):             # <<<<<<<<<<<<<<
 *             raise TypeError("Must provide a Sequence-like object")
 * 
 */
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":246
 *             raise TypeError("Must provide a Sequence-like object")
 * 
 *         fast_seq = PySequence_Fast(values, "Must provide a Sequence-like object")             # <<<<<<<<<<<<<<
 *         n = PySequence_Fast_GET_SIZE(values)
 * 
 */
  __pyx_t_2 = PySequence_Fast(__pyx_v_values, ((char *)"Must provide a Sequence-like object")); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_fast_seq = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":247
 * 
 *         fast_seq = PySequence_Fast(values, "Must provide a Sequence-like object")
 *         n = PySequence_Fast_GET_SIZE(values)             # <<<<<<<<<<<<<<
 * 
 *         for i in range(n):
 */
  __pyx_v_n = PySequence_Fast_GET_SIZE(__pyx_v_values);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":249
 *         n = PySequence_Fast_GET_SIZE(values)
 * 
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             if self.append(<object>PySequence_Fast_GET_ITEM(fast_seq, i)) != 0:
 *                 return 1
 */
  __pyx_t_3 = __pyx_v_n;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "ms_peak_picker/_c/interval_t_vector.pyx":250
 * 
 *         for i in range(n):
 *             if self.append(<object>PySequence_Fast_GET_ITEM(fast_seq, i)) != 0:             # <<<<<<<<<<<<<<
 *                 return 1
 * 
 */
    __pyx_t_6 = PySequence_Fast_GET_ITEM(__pyx_v_fast_seq, __pyx_v_i);
    if (!(likely(PyTuple_CheckExact(((PyObject *)__pyx_t_6)))||((((PyObject *)__pyx_t_6)) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(((PyObject *)__pyx_t_6))->tp_name), 0))) __PYX_ERR(0, 250, __pyx_L1_error)
    __pyx_t_7 = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_append(__pyx_v_self, ((PyObject*)((PyObject *)__pyx_t_6)), 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 250, __pyx_L1_error)
    __pyx_t_1 = ((__pyx_t_7 != 0) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/interval_t_vector.pyx":251
 *         for i in range(n):
 *             if self.append(<object>PySequence_Fast_GET_ITEM(fast_seq, i)) != 0:
 *                 return 1             # <<<<<<<<<<<<<<
 * 
 *     cpdef int reserve(self, size_t size) nogil:
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/interval_t_vector.pyx":250
 * 
 *         for i in range(n):
 *             if self.append(<object>PySequence_Fast_GET_ITEM(fast_seq, i)) != 0:             # <<<<<<<<<<<<<<
 *                 return 1
 * 
 */
    }
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":239
 *         return self.cappend(cvalue)
 * 
 *     cpdef int extend(self, object values) except *:             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fast_seq);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_5extend(PyObject *__pyx_v_self, PyObject *__pyx_v_values); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_4extend[] = "IntervalVector.extend(self, values) -> int";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_5extend(PyObject *__pyx_v_self, PyObject *__pyx_v_values) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extend (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_4extend(((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_v_self), ((PyObject *)__pyx_v_values));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_4extend(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_values) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("extend", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_extend(__pyx_v_self, __pyx_v_values, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 239, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":253
 *                 return 1
 * 
 *     cpdef int reserve(self, size_t size) nogil:             # <<<<<<<<<<<<<<
 *         return interval_t_vector_reserve(self.impl, size)
 * 
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_7reserve(PyObject *__pyx_v_self, PyObject *__pyx_arg_size); /*proto*/
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_reserve(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, size_t __pyx_v_size, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":254
 * 
 *     cpdef int reserve(self, size_t size) nogil:
 *         return interval_t_vector_reserve(self.impl, size)             # <<<<<<<<<<<<<<
 * 
 *     cpdef int fill(self, interval_t value) nogil:
 */
  __pyx_r = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector_reserve(__pyx_v_self->impl, __pyx_v_size);
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":253
 *                 return 1
 * 
 *     cpdef int reserve(self, size_t size) nogil:             # <<<<<<<<<<<<<<
 *         return interval_t_vector_reserve(self.impl, size)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_7reserve(PyObject *__pyx_v_self, PyObject *__pyx_arg_size); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_6reserve[] = "IntervalVector.reserve(self, size_t size) -> int";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_7reserve(PyObject *__pyx_v_self, PyObject *__pyx_arg_size) {
  size_t __pyx_v_size;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reserve (wrapper)", 0);
  assert(__pyx_arg_size); {
    __pyx_v_size = __Pyx_PyInt_As_size_t(__pyx_arg_size); if (unlikely((__pyx_v_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 253, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.reserve", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_6reserve(((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_v_self), ((size_t)__pyx_v_size));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_6reserve(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, size_t __pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reserve", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_reserve(__pyx_v_self, __pyx_v_size, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.reserve", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":256
 *         return interval_t_vector_reserve(self.impl, size)
 * 
 *     cpdef int fill(self, interval_t value) nogil:             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_9fill(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_fill(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  int __pyx_r;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":259
 *         cdef:
 *             size_t i, n
 *         n = self.size()             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             self.set(i, value)
 */
  __pyx_v_n = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_size(__pyx_v_self);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":260
 *             size_t i, n
 *         n = self.size()
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             self.set(i, value)
 *         return 0
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "ms_peak_picker/_c/interval_t_vector.pyx":261
 *         n = self.size()
 *         for i in range(n):
 *             self.set(i, value)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_set(__pyx_v_self, __pyx_v_i, __pyx_v_value);
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":262
 *         for i in range(n):
 *             self.set(i, value)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cpdef IntervalVector copy(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":256
 *         return interval_t_vector_reserve(self.impl, size)
 * 
 *     cpdef int fill(self, interval_t value) nogil:             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_9fill(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_8fill[] = "IntervalVector.fill(self, interval_t value) -> int";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_9fill(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fill (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_convert__from_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t(__pyx_arg_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 256, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.fill", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_8fill(((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_8fill(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fill", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_fill(__pyx_v_self, __pyx_v_value, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.fill", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":264
 *         return 0
 * 
 *     cpdef IntervalVector copy(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IntervalVector dup
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_11copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_copy(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_dup = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":268
 *             IntervalVector dup
 *             size_t i, n
 *         n = self.size()             # <<<<<<<<<<<<<<
 *         dup = IntervalVector._create(n)
 *         for i in range(n):
 */
  __pyx_v_n = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_size(__pyx_v_self);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":269
 *             size_t i, n
 *         n = self.size()
 *         dup = IntervalVector._create(n)             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             dup.cappend(self.get(i))
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__create(__pyx_v_n)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_dup = ((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":270
 *         n = self.size()
 *         dup = IntervalVector._create(n)
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             dup.cappend(self.get(i))
 *         return dup
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "ms_peak_picker/_c/interval_t_vector.pyx":271
 *         dup = IntervalVector._create(n)
 *         for i in range(n):
 *             dup.cappend(self.get(i))             # <<<<<<<<<<<<<<
 *         return dup
 * 
 */
    (void)(__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_cappend(__pyx_v_dup, __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_get(__pyx_v_self, __pyx_v_i)));
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":272
 *         for i in range(n):
 *             dup.cappend(self.get(i))
 *         return dup             # <<<<<<<<<<<<<<
 * 
 *     cdef IntervalVector _slice(self, object slice_spec):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_dup));
  __pyx_r = __pyx_v_dup;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":264
 *         return 0
 * 
 *     cpdef IntervalVector copy(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IntervalVector dup
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_dup);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_11copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_10copy[] = "IntervalVector.copy(self) -> IntervalVector";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_11copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_10copy(((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_10copy(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":274
 *         return dup
 * 
 *     cdef IntervalVector _slice(self, object slice_spec):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IntervalVector dup
 */

static struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__slice(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_slice_spec) {
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_dup = 0;
  Py_ssize_t __pyx_v_start;
  Py_ssize_t __pyx_v_stop;
  Py_ssize_t __pyx_v_step;
  Py_ssize_t __pyx_v_slice_length;
  Py_ssize_t __pyx_v_i;
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_slice", 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":278
 *             IntervalVector dup
 *             Py_ssize_t length, start, stop, step, slice_length, i
 *         PySlice_GetIndicesEx(             # <<<<<<<<<<<<<<
 *             slice_spec, self.size(), &start, &stop, &step, &slice_length)
 *         dup = IntervalVector._create(slice_length)
 */
  __pyx_t_1 = PySlice_GetIndicesEx(__pyx_v_slice_spec, __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_size(__pyx_v_self), (&__pyx_v_start), (&__pyx_v_stop), (&__pyx_v_step), (&__pyx_v_slice_length)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 278, __pyx_L1_error)

  /* "ms_peak_picker/_c/interval_t_vector.pyx":280
 *         PySlice_GetIndicesEx(
 *             slice_spec, self.size(), &start, &stop, &step, &slice_length)
 *         dup = IntervalVector._create(slice_length)             # <<<<<<<<<<<<<<
 *         i = start
 *         while i < stop:
 */
  __pyx_t_2 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__create(__pyx_v_slice_length)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_dup = ((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":281
 *             slice_spec, self.size(), &start, &stop, &step, &slice_length)
 *         dup = IntervalVector._create(slice_length)
 *         i = start             # <<<<<<<<<<<<<<
 *         while i < stop:
 *             dup.cappend(self.get(i))
 */
  __pyx_v_i = __pyx_v_start;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":282
 *         dup = IntervalVector._create(slice_length)
 *         i = start
 *         while i < stop:             # <<<<<<<<<<<<<<
 *             dup.cappend(self.get(i))
 *             i += step
 */
  while (1) {
    __pyx_t_3 = ((__pyx_v_i < __pyx_v_stop) != 0);
    if (!__pyx_t_3) break;

    /* "ms_peak_picker/_c/interval_t_vector.pyx":283
 *         i = start
 *         while i < stop:
 *             dup.cappend(self.get(i))             # <<<<<<<<<<<<<<
 *             i += step
 *         return dup
 */
    (void)(__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_cappend(__pyx_v_dup, __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_get(__pyx_v_self, __pyx_v_i)));

    /* "ms_peak_picker/_c/interval_t_vector.pyx":284
 *         while i < stop:
 *             dup.cappend(self.get(i))
 *             i += step             # <<<<<<<<<<<<<<
 *         return dup
 * 
 */
    __pyx_v_i = (__pyx_v_i + __pyx_v_step);
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":285
 *             dup.cappend(self.get(i))
 *             i += step
 *         return dup             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_dup));
  __pyx_r = __pyx_v_dup;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":274
 *         return dup
 * 
 *     cdef IntervalVector _slice(self, object slice_spec):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IntervalVector dup
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector._slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_dup);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":287
 *         return dup
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.get_should_free():
 *             self.free_storage()
 */

/* Python wrapper */
static void __pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_13__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_13__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_12__dealloc__(((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_12__dealloc__(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":288
 * 
 *     def __dealloc__(self):
 *         if self.get_should_free():             # <<<<<<<<<<<<<<
 *             self.free_storage()
 * 
 */
  __pyx_t_1 = (__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_get_should_free(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":289
 *     def __dealloc__(self):
 *         if self.get_should_free():
 *             self.free_storage()             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
    (void)(__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_free_storage(__pyx_v_self));

    /* "ms_peak_picker/_c/interval_t_vector.pyx":288
 * 
 *     def __dealloc__(self):
 *         if self.get_should_free():             # <<<<<<<<<<<<<<
 *             self.free_storage()
 * 
 */
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":287
 *         return dup
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.get_should_free():
 *             self.free_storage()
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":291
 *             self.free_storage()
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.size()
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_15__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_15__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_14__len__(((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_14__len__(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":292
 * 
 *     def __len__(self):
 *         return self.size()             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __pyx_r = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_size(__pyx_v_self);
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":291
 *             self.free_storage()
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.size()
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_18generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "ms_peak_picker/_c/interval_t_vector.pyx":294
 *         return self.size()
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_17__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_17__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_16__iter__(((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_16__iter__(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self) {
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ *)__pyx_tp_new_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__(__pyx_ptype_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 294, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_18generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_IntervalVector___iter, __pyx_n_s_ms_peak_picker__c_interval_t_vec); if (unlikely(!gen)) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_18generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ *__pyx_cur_scope = ((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 294, __pyx_L1_error)

  /* "ms_peak_picker/_c/interval_t_vector.pyx":297
 *         cdef:
 *             size_t i, n
 *         n = self.size()             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             yield self._to_python(self.get(i))
 */
  __pyx_cur_scope->__pyx_v_n = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_size(__pyx_cur_scope->__pyx_v_self);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":298
 *             size_t i, n
 *         n = self.size()
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             yield self._to_python(self.get(i))
 * 
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_cur_scope->__pyx_v_i = __pyx_t_3;

    /* "ms_peak_picker/_c/interval_t_vector.pyx":299
 *         n = self.size()
 *         for i in range(n):
 *             yield self._to_python(self.get(i))             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, i):
 */
    __pyx_t_4 = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__to_python(__pyx_cur_scope->__pyx_v_self, __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_get(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_i), 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 299, __pyx_L1_error)
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":294
 *         return self.size()
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":301
 *             yield self._to_python(self.get(i))
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Py_ssize_t index
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_20__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_20__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_19__getitem__(((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_v_self), ((PyObject *)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_19__getitem__(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_i) {
  Py_ssize_t __pyx_v_index;
  size_t __pyx_v_n;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":305
 *             Py_ssize_t index
 *             size_t n
 *         if isinstance(i, slice):             # <<<<<<<<<<<<<<
 *             return self._slice(i)
 *         index = i
 */
  __pyx_t_1 = PySlice_Check(__pyx_v_i); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":306
 *             size_t n
 *         if isinstance(i, slice):
 *             return self._slice(i)             # <<<<<<<<<<<<<<
 *         index = i
 *         n = self.size()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__slice(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/interval_t_vector.pyx":305
 *             Py_ssize_t index
 *             size_t n
 *         if isinstance(i, slice):             # <<<<<<<<<<<<<<
 *             return self._slice(i)
 *         index = i
 */
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":307
 *         if isinstance(i, slice):
 *             return self._slice(i)
 *         index = i             # <<<<<<<<<<<<<<
 *         n = self.size()
 *         if index < 0:
 */
  __pyx_t_4 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_4 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 307, __pyx_L1_error)
  __pyx_v_index = __pyx_t_4;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":308
 *             return self._slice(i)
 *         index = i
 *         n = self.size()             # <<<<<<<<<<<<<<
 *         if index < 0:
 *             index = n + index
 */
  __pyx_v_n = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_size(__pyx_v_self);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":309
 *         index = i
 *         n = self.size()
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             index = n + index
 *         if index > n or index < 0:
 */
  __pyx_t_2 = ((__pyx_v_index < 0) != 0);
  if (__pyx_t_2) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":310
 *         n = self.size()
 *         if index < 0:
 *             index = n + index             # <<<<<<<<<<<<<<
 *         if index > n or index < 0:
 *             raise IndexError(index)
 */
    __pyx_v_index = (__pyx_v_n + __pyx_v_index);

    /* "ms_peak_picker/_c/interval_t_vector.pyx":309
 *         index = i
 *         n = self.size()
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             index = n + index
 *         if index > n or index < 0:
 */
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":311
 *         if index < 0:
 *             index = n + index
 *         if index > n or index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError(index)
 *         return self._to_python(self.get(index))
 */
  __pyx_t_1 = ((__pyx_v_index > __pyx_v_n) != 0);
  if (!__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_1 = ((__pyx_v_index < 0) != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L6_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":312
 *             index = n + index
 *         if index > n or index < 0:
 *             raise IndexError(index)             # <<<<<<<<<<<<<<
 *         return self._to_python(self.get(index))
 * 
 */
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 312, __pyx_L1_error)

    /* "ms_peak_picker/_c/interval_t_vector.pyx":311
 *         if index < 0:
 *             index = n + index
 *         if index > n or index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError(index)
 *         return self._to_python(self.get(index))
 */
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":313
 *         if index > n or index < 0:
 *             raise IndexError(index)
 *         return self._to_python(self.get(index))             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, i, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__to_python(__pyx_v_self, __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_get(__pyx_v_self, __pyx_v_index), 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":301
 *             yield self._to_python(self.get(i))
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Py_ssize_t index
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":315
 *         return self._to_python(self.get(index))
 * 
 *     def __setitem__(self, i, value):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Py_ssize_t index
 */

/* Python wrapper */
static int __pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_22__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_22__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_21__setitem__(((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_v_self), ((PyObject *)__pyx_v_i), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_21__setitem__(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_value) {
  Py_ssize_t __pyx_v_index;
  size_t __pyx_v_n;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":319
 *             Py_ssize_t index
 *             size_t n
 *         if isinstance(i, slice):             # <<<<<<<<<<<<<<
 *             raise TypeError("Does not support slice-assignment yet")
 *         n = self.size()
 */
  __pyx_t_1 = PySlice_Check(__pyx_v_i); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (unlikely(__pyx_t_2)) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":320
 *             size_t n
 *         if isinstance(i, slice):
 *             raise TypeError("Does not support slice-assignment yet")             # <<<<<<<<<<<<<<
 *         n = self.size()
 *         index = i
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 320, __pyx_L1_error)

    /* "ms_peak_picker/_c/interval_t_vector.pyx":319
 *             Py_ssize_t index
 *             size_t n
 *         if isinstance(i, slice):             # <<<<<<<<<<<<<<
 *             raise TypeError("Does not support slice-assignment yet")
 *         n = self.size()
 */
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":321
 *         if isinstance(i, slice):
 *             raise TypeError("Does not support slice-assignment yet")
 *         n = self.size()             # <<<<<<<<<<<<<<
 *         index = i
 *         if index < 0:
 */
  __pyx_v_n = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_size(__pyx_v_self);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":322
 *             raise TypeError("Does not support slice-assignment yet")
 *         n = self.size()
 *         index = i             # <<<<<<<<<<<<<<
 *         if index < 0:
 *             index = n + index
 */
  __pyx_t_4 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_4 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 322, __pyx_L1_error)
  __pyx_v_index = __pyx_t_4;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":323
 *         n = self.size()
 *         index = i
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             index = n + index
 *         if index > self.size() or index < 0:
 */
  __pyx_t_2 = ((__pyx_v_index < 0) != 0);
  if (__pyx_t_2) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":324
 *         index = i
 *         if index < 0:
 *             index = n + index             # <<<<<<<<<<<<<<
 *         if index > self.size() or index < 0:
 *             raise IndexError(i)
 */
    __pyx_v_index = (__pyx_v_n + __pyx_v_index);

    /* "ms_peak_picker/_c/interval_t_vector.pyx":323
 *         n = self.size()
 *         index = i
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             index = n + index
 *         if index > self.size() or index < 0:
 */
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":325
 *         if index < 0:
 *             index = n + index
 *         if index > self.size() or index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError(i)
 *         self.set(index, self._to_c(value))
 */
  __pyx_t_1 = ((__pyx_v_index > __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_size(__pyx_v_self)) != 0);
  if (!__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_1 = ((__pyx_v_index < 0) != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L6_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "ms_peak_picker/_c/interval_t_vector.pyx":326
 *             index = n + index
 *         if index > self.size() or index < 0:
 *             raise IndexError(i)             # <<<<<<<<<<<<<<
 *         self.set(index, self._to_c(value))
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 326, __pyx_L1_error)

    /* "ms_peak_picker/_c/interval_t_vector.pyx":325
 *         if index < 0:
 *             index = n + index
 *         if index > self.size() or index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError(i)
 *         self.set(index, self._to_c(value))
 */
  }

  /* "ms_peak_picker/_c/interval_t_vector.pyx":327
 *         if index > self.size() or index < 0:
 *             raise IndexError(i)
 *         self.set(index, self._to_c(value))             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_5 = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__to_c(__pyx_v_self, __pyx_v_value, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 327, __pyx_L1_error)
  __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_set(__pyx_v_self, __pyx_v_index, __pyx_t_5);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":315
 *         return self._to_python(self.get(index))
 * 
 *     def __setitem__(self, i, value):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Py_ssize_t index
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":329
 *         self.set(index, self._to_c(value))
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "{self.__class__.__name__}({members})".format(self=self, members=list(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_24__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_24__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_23__repr__(((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_23__repr__(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":330
 * 
 *     def __repr__(self):
 *         return "{self.__class__.__name__}({members})".format(self=self, members=list(self))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_self___class_____name___members, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_self, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 330, __pyx_L1_error)
  __pyx_t_3 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_members, __pyx_t_3) < 0) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":329
 *         self.set(index, self._to_c(value))
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "{self.__class__.__name__}({members})".format(self=self, members=list(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":336
 * 
 * 
 *     cpdef object _to_python(self, interval_t value):             # <<<<<<<<<<<<<<
 *         return tuple_from_interval(value)
 * 
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_26_to_python(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static PyObject *__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__to_python(CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_to_python", 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":337
 * 
 *     cpdef object _to_python(self, interval_t value):
 *         return tuple_from_interval(value)             # <<<<<<<<<<<<<<
 * 
 *     cpdef interval_t _to_c(self, object value) except *:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_tuple_from_interval(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":336
 * 
 * 
 *     cpdef object _to_python(self, interval_t value):             # <<<<<<<<<<<<<<
 *         return tuple_from_interval(value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector._to_python", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_26_to_python(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_25_to_python[] = "IntervalVector._to_python(self, interval_t value)";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_26_to_python(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_to_python (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_convert__from_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t(__pyx_arg_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 336, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector._to_python", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_25_to_python(((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_25_to_python(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_to_python", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__to_python(__pyx_v_self, __pyx_v_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector._to_python", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/interval_t_vector.pyx":339
 *         return tuple_from_interval(value)
 * 
 *     cpdef interval_t _to_c(self, object value) except *:             # <<<<<<<<<<<<<<
 *         return interval_from_tuple(value)
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_28_to_c(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__to_c(CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_to_c", 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":340
 * 
 *     cpdef interval_t _to_c(self, object value) except *:
 *         return interval_from_tuple(value)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 340, __pyx_L1_error)
  __pyx_r = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_from_tuple(((PyObject*)__pyx_v_value));
  goto __pyx_L0;

  /* "ms_peak_picker/_c/interval_t_vector.pyx":339
 *         return tuple_from_interval(value)
 * 
 *     cpdef interval_t _to_c(self, object value) except *:             # <<<<<<<<<<<<<<
 *         return interval_from_tuple(value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector._to_c", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_28_to_c(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_27_to_c[] = "IntervalVector._to_c(self, value) -> interval_t";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_28_to_c(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_to_c (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_27_to_c(((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_27_to_c(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_to_c", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__to_c(__pyx_v_self, __pyx_v_value, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 339, __pyx_L1_error)
  __pyx_t_2 = __pyx_convert__to_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector._to_c", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_30__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_29__reduce_cython__[] = "IntervalVector.__reduce_cython__(self)";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_30__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_29__reduce_cython__(((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_29__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_32__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_31__setstate_cython__[] = "IntervalVector.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_32__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_31__setstate_cython__(((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_31__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.interval_t_vector.IntervalVector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "FromPyStructUtility":11
 * 
 * @cname("__pyx_convert__from_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t")
 * cdef struct_type __pyx_convert__from_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t(obj) except *:             # <<<<<<<<<<<<<<
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):
 */

static struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_convert__from_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t(PyObject *__pyx_v_obj) {
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_v_result;
  PyObject *__pyx_v_value = NULL;
  struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert__from_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t", 0);

  /* "FromPyStructUtility":13
 * cdef struct_type __pyx_convert__from_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t(obj) except *:
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):             # <<<<<<<<<<<<<<
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 */
  __pyx_t_1 = ((!(PyMapping_Check(__pyx_v_obj) != 0)) != 0);
  if (__pyx_t_1) {

    /* "FromPyStructUtility":14
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    __pyx_t_2 = PyErr_Format(__pyx_builtin_TypeError, ((char const *)"Expected %.16s, got %.200s"), ((char *)"a mapping"), Py_TYPE(__pyx_v_obj)->tp_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "FromPyStructUtility":13
 * cdef struct_type __pyx_convert__from_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t(obj) except *:
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):             # <<<<<<<<<<<<<<
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 */
  }

  /* "FromPyStructUtility":16
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['start']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "FromPyStructUtility":17
 * 
 *     try:
 *         value = obj['start']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'start'")
 */
      __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_start); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_value = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "FromPyStructUtility":16
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['start']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "FromPyStructUtility":18
 *     try:
 *         value = obj['start']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'start'")
 *     result.start = value
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(1, 18, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);

      /* "FromPyStructUtility":19
 *         value = obj['start']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'start'")             # <<<<<<<<<<<<<<
 *     result.start = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 19, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(1, 19, __pyx_L6_except_error)
    }
    goto __pyx_L6_except_error;
    __pyx_L6_except_error:;

    /* "FromPyStructUtility":16
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['start']
 *     except KeyError:
 */
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L9_try_end:;
  }

  /* "FromPyStructUtility":20
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'start'")
 *     result.start = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['end']
 */
  __pyx_t_10 = __Pyx_PyInt_As_size_t(__pyx_v_value); if (unlikely((__pyx_t_10 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 20, __pyx_L1_error)
  __pyx_v_result.start = __pyx_t_10;

  /* "FromPyStructUtility":21
 *         raise ValueError("No value specified for struct attribute 'start'")
 *     result.start = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['end']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "FromPyStructUtility":22
 *     result.start = value
 *     try:
 *         value = obj['end']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'end'")
 */
      __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_end); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 22, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "FromPyStructUtility":21
 *         raise ValueError("No value specified for struct attribute 'start'")
 *     result.start = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['end']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L17_try_end;
    __pyx_L12_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":23
 *     try:
 *         value = obj['end']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'end'")
 *     result.end = value
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_2) < 0) __PYX_ERR(1, 23, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_2);

      /* "FromPyStructUtility":24
 *         value = obj['end']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'end'")             # <<<<<<<<<<<<<<
 *     result.end = value
 *     return result
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 24, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(1, 24, __pyx_L14_except_error)
    }
    goto __pyx_L14_except_error;
    __pyx_L14_except_error:;

    /* "FromPyStructUtility":21
 *         raise ValueError("No value specified for struct attribute 'start'")
 *     result.start = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['end']
 *     except KeyError:
 */
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L17_try_end:;
  }

  /* "FromPyStructUtility":25
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'end'")
 *     result.end = value             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_t_10 = __Pyx_PyInt_As_size_t(__pyx_v_value); if (unlikely((__pyx_t_10 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 25, __pyx_L1_error)
  __pyx_v_result.end = __pyx_t_10;

  /* "FromPyStructUtility":26
 *         raise ValueError("No value specified for struct attribute 'end'")
 *     result.end = value
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "FromPyStructUtility":11
 * 
 * @cname("__pyx_convert__from_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t")
 * cdef struct_type __pyx_convert__from_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t(obj) except *:             # <<<<<<<<<<<<<<
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector;

static struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *__pyx_freelist_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector[512];
static int __pyx_freecount_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector = 0;

static PyObject *__pyx_tp_new_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *p;
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector)))) {
    o = (PyObject*)__pyx_freelist_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector[--__pyx_freecount_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector];
    memset(o, 0, sizeof(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  p = ((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)o);
  p->__pyx_vtab = __pyx_vtabptr_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector;
  return o;
}

static void __pyx_tp_dealloc_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector(PyObject *o) {
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_13__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector < 512) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector)))) {
    __pyx_freelist_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector[__pyx_freecount_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector++] = ((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}
static PyObject *__pyx_sq_item_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_22__setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyMethodDef __pyx_methods_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector[] = {
  {"append", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_3append, METH_O, __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_2append},
  {"extend", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_5extend, METH_O, __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_4extend},
  {"reserve", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_7reserve, METH_O, __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_6reserve},
  {"fill", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_9fill, METH_O, __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_8fill},
  {"copy", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_11copy, METH_NOARGS, __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_10copy},
  {"_to_python", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_26_to_python, METH_O, __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_25_to_python},
  {"_to_c", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_28_to_c, METH_O, __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_27_to_c},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_30__reduce_cython__, METH_NOARGS, __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_29__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_32__setstate_cython__, METH_O, __pyx_doc_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_31__setstate_cython__},
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_IntervalVector = {
  __pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_15__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_IntervalVector = {
  __pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_15__len__, /*mp_length*/
  __pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_20__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector = {
  PyVarObject_HEAD_INIT(0, 0)
  "ms_peak_picker._c.interval_t_vector.IntervalVector", /*tp_name*/
  sizeof(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_24__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_IntervalVector, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_IntervalVector, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  "IntervalVector(seed=None)", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_17__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ *__pyx_freelist_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__[8];
static int __pyx_freecount_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ = 0;

static PyObject *__pyx_tp_new_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__)))) {
    o = (PyObject*)__pyx_freelist_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__[--__pyx_freecount_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__];
    memset(o, 0, sizeof(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__(PyObject *o) {
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ *p = (struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__)))) {
    __pyx_freelist_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__[__pyx_freecount_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__++] = ((struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ *p = (struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "ms_peak_picker._c.interval_t_vector.__pyx_scope_struct____iter__", /*tp_name*/
  sizeof(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_interval_t_vector(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_interval_t_vector},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "interval_t_vector",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_Does_not_support_slice_assignmen, __pyx_k_Does_not_support_slice_assignmen, sizeof(__pyx_k_Does_not_support_slice_assignmen), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_n_s_IntervalVector, __pyx_k_IntervalVector, sizeof(__pyx_k_IntervalVector), 0, 0, 1, 1},
  {&__pyx_n_s_IntervalVector___iter, __pyx_k_IntervalVector___iter, sizeof(__pyx_k_IntervalVector___iter), 0, 0, 1, 1},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_kp_s_Must_provide_a_Sequence_like_obj, __pyx_k_Must_provide_a_Sequence_like_obj, sizeof(__pyx_k_Must_provide_a_Sequence_like_obj), 0, 0, 1, 0},
  {&__pyx_kp_s_No_value_specified_for_struct_at, __pyx_k_No_value_specified_for_struct_at, sizeof(__pyx_k_No_value_specified_for_struct_at), 0, 0, 1, 0},
  {&__pyx_kp_s_No_value_specified_for_struct_at_2, __pyx_k_No_value_specified_for_struct_at_2, sizeof(__pyx_k_No_value_specified_for_struct_at_2), 0, 0, 1, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_iter, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_members, __pyx_k_members, sizeof(__pyx_k_members), 0, 0, 1, 1},
  {&__pyx_n_s_ms_peak_picker__c_interval_t_vec, __pyx_k_ms_peak_picker__c_interval_t_vec, sizeof(__pyx_k_ms_peak_picker__c_interval_t_vec), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_seed, __pyx_k_seed, sizeof(__pyx_k_seed), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_kp_s_self___class_____name___members, __pyx_k_self___class_____name___members, sizeof(__pyx_k_self___class_____name___members), 0, 0, 1, 0},
  {&__pyx_kp_s_self_impl_cannot_be_converted_to, __pyx_k_self_impl_cannot_be_converted_to, sizeof(__pyx_k_self_impl_cannot_be_converted_to), 0, 0, 1, 0},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 186, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 249, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 312, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(1, 18, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(1, 19, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":186
 *         if seed is not None:
 *             if not PySequence_Check(seed):
 *                 raise TypeError("Must provide a Sequence-like object")             # <<<<<<<<<<<<<<
 *             n = len(seed)
 *             self.allocate_storage_with_size(n)
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_Must_provide_a_Sequence_like_obj); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "ms_peak_picker/_c/interval_t_vector.pyx":320
 *             size_t n
 *         if isinstance(i, slice):
 *             raise TypeError("Does not support slice-assignment yet")             # <<<<<<<<<<<<<<
 *         n = self.size()
 *         index = i
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_Does_not_support_slice_assignmen); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_self_impl_cannot_be_converted_to); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "(tree fragment)":4
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_self_impl_cannot_be_converted_to); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "FromPyStructUtility":19
 *         value = obj['start']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'start'")             # <<<<<<<<<<<<<<
 *     result.start = value
 *     try:
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(1, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "FromPyStructUtility":24
 *         value = obj['end']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'end'")             # <<<<<<<<<<<<<<
 *     result.end = value
 *     return result
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_2); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("interval_overlaps", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_overlaps, "int (struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("interval_contains", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_contains, "int (struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *, size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("initialize_interval_vector_t", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_initialize_interval_vector_t, "int (struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *, size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("make_interval_t_vector_with_size", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_make_interval_t_vector_with_size, "struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *(size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("make_interval_t_vector", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_make_interval_t_vector, "struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *(void)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("interval_t_vector_resize", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector_resize, "int (struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("interval_t_vector_append", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector_append, "int (struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("interval_t_vector_reserve", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector_reserve, "int (struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *, size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("interval_t_vector_reset", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector_reset, "void (struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("free_interval_t_vector", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_free_interval_t_vector, "void (struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("print_interval_t_vector", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_print_interval_t_vector, "void (struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector = &__pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.allocate_storage = (int (*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_allocate_storage;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.allocate_storage_with_size = (int (*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, size_t))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_allocate_storage_with_size;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.free_storage = (int (*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_free_storage;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.get_should_free = (int (*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_get_should_free;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.set_should_free = (void (*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, int))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_set_should_free;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.get_data = (struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t *(*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_get_data;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector._create = (struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *(*)(size_t))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__create;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.wrap = (struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *(*)(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t_vector *))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_wrap;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.get = (struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t (*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, size_t))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_get;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.set = (void (*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, size_t, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_set;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.size = (size_t (*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_size;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.cappend = (int (*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_cappend;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector._slice = (struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *(*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, PyObject *))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__slice;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.copy = (struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *(*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_copy;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.append = (int (*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, PyObject *, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_append;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.extend = (int (*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, PyObject *, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_extend;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.reserve = (int (*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, size_t, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_reserve;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.fill = (int (*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector_fill;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector._to_python = (PyObject *(*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__to_python;
  __pyx_vtable_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector._to_c = (struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t (*)(struct __pyx_obj_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector *, PyObject *, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_17interval_t_vector_14IntervalVector__to_c;
  if (PyType_Ready(&__pyx_type_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector) < 0) __PYX_ERR(0, 159, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.tp_dictoffset && __pyx_type_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  if (__Pyx_SetVtable(__pyx_type_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector.tp_dict, __pyx_vtabptr_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector) < 0) __PYX_ERR(0, 159, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IntervalVector, (PyObject *)&__pyx_type_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector) < 0) __PYX_ERR(0, 159, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector) < 0) __PYX_ERR(0, 159, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector = &__pyx_type_14ms_peak_picker_2_c_17interval_t_vector_IntervalVector;
  if (PyType_Ready(&__pyx_type_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__) < 0) __PYX_ERR(0, 294, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__.tp_dictoffset && __pyx_type_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__ = &__pyx_type_14ms_peak_picker_2_c_17interval_t_vector___pyx_scope_struct____iter__;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initinterval_t_vector(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initinterval_t_vector(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_interval_t_vector(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_interval_t_vector(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_interval_t_vector(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'interval_t_vector' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_interval_t_vector(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("interval_t_vector", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_ms_peak_picker___c__interval_t_vector) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "ms_peak_picker._c.interval_t_vector")) {
      if (unlikely(PyDict_SetItemString(modules, "ms_peak_picker._c.interval_t_vector", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "ms_peak_picker/_c/interval_t_vector.pyx":1
 * cimport cython             # <<<<<<<<<<<<<<
 * 
 * from libc.stdlib cimport malloc, realloc, free
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "FromPyStructUtility":11
 * 
 * @cname("__pyx_convert__from_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t")
 * cdef struct_type __pyx_convert__from_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t(obj) except *:             # <<<<<<<<<<<<<<
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init ms_peak_picker._c.interval_t_vector", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init ms_peak_picker._c.interval_t_vector");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

static PyObject* __pyx_convert__to_py_struct____pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t(struct __pyx_t_14ms_peak_picker_2_c_17interval_t_vector_interval_t s) {
  PyObject* res;
  PyObject* member;
  res = __Pyx_PyDict_NewPresized(2); if (unlikely(!res)) return NULL;
  member = __Pyx_PyInt_FromSize_t(s.start); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_start, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __Pyx_PyInt_FromSize_t(s.end); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_end, member) < 0)) goto bad;
  Py_DECREF(member);
  return res;
  bad:
  Py_XDECREF(member);
  Py_DECREF(res);
  return NULL;
}
/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (Py_REFCNT(self) > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
