/* Generated by Cython 0.29.32 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "name": "ms_peak_picker._c.peak_set",
        "sources": [
            "src/ms_peak_picker/_c/peak_set.pyx"
        ]
    },
    "module_name": "ms_peak_picker._c.peak_set"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_32"
#define CYTHON_HEX_VERSION 0x001D20F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_HEX >= 0x07030900)
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__ms_peak_picker___c__peak_set
#define __PYX_HAVE_API__ms_peak_picker___c__peak_set
/* Early includes */
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "pythread.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "src/ms_peak_picker/_c/peak_set.pyx",
  "src/ms_peak_picker/_c/peak_set.pxd",
  "stringsource",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for;
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell;
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__has_peak;

/* "ms_peak_picker/_c/peak_set.pxd":42
 *     cdef FittedPeak _get_nearest_peak(self, double mz, double* errout)
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakSet clone(self)
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak {
  int __pyx_n;
  double tolerance;
};

/* "ms_peak_picker/_c/peak_set.pxd":50
 *     cdef int _between_bounds(self, double m1, double m2, size_t* startp, size_t* endp)
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef FittedPeak getitem(self, size_t i)
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for {
  int __pyx_n;
  double error_tolerance;
};

/* "ms_peak_picker/_c/peak_set.pxd":64
 * 
 * 
 * cdef struct index_cell:             # <<<<<<<<<<<<<<
 *     double center_value
 *     size_t start
 */
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell {
  double center_value;
  size_t start;
  size_t end;
};

/* "ms_peak_picker/_c/peak_set.pxd":70
 * 
 * 
 * cdef struct index_list:             # <<<<<<<<<<<<<<
 *     index_cell* index
 *     size_t size
 */
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list {
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell *index;
  size_t size;
  double low;
  double high;
};

/* "ms_peak_picker/_c/peak_set.pxd":90
 *     cpdef _allocate_index(self, size_t interval_index_size)
 * 
 *     cpdef FittedPeak has_peak_hinted(self, double mz, double tolerance=*, size_t hint_start=*, size_t hint_end=*)             # <<<<<<<<<<<<<<
 * 
 * cdef size_t double_binary_search_ppm(double* array, double value, double tolerance, size_t n)
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted {
  int __pyx_n;
  double tolerance;
  size_t hint_start;
  size_t hint_end;
};

/* "ms_peak_picker/_c/peak_set.pyx":750
 * 
 *     @cython.cdivision(True)
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=1e-5):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n, s
 */
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__has_peak {
  int __pyx_n;
  double tolerance;
};

/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtab;
  double mz;
  double intensity;
  double area;
};


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;
  double signal_to_noise;
  double full_width_at_half_max;
  double left_width;
  double right_width;
  long peak_count;
  long index;
};


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet {
  PyObject_HEAD
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtab;
  PyObject *peaks;
  int indexed;
};


/* "ms_peak_picker/_c/peak_set.pxd":83
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;
  double *mz_index;
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *interval_index;
};



/* "ms_peak_picker/_c/peak_set.pyx":19
 * 
 * @cython.freelist(1000000)
 * cdef class PeakBase:             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakBase clone(self):
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase;


/* "ms_peak_picker/_c/peak_set.pyx":28
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     """Represent a single centroided mass spectral peak.
 * 
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak {
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_create)(double, double, double, double, double, double, long, long, double);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak;


/* "ms_peak_picker/_c/peak_set.pyx":140
 * 
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     """A sequence of :class:`FittedPeak` instances, ordered by m/z,
 *     providing efficient search and retrieval of individual peaks or
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_create)(PyObject *);
  PyObject *(*reindex)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
  size_t (*_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_get_nearest_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double *);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak *__pyx_optional_args);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double);
  int (*_between_bounds)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double, size_t *, size_t *);
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for *__pyx_optional_args);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*getitem)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, size_t);
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *);
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet;


/* "ms_peak_picker/_c/peak_set.pyx":685
 * cdef size_t INTERVAL_INDEX_SIZE = 0
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;
  PyObject *(*_allocate_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, size_t, int __pyx_skip_dispatch);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak_hinted)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted *__pyx_optional_args);
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* pyobject_as_double.proto */
static double __Pyx__PyObject_AsDouble(PyObject* obj);
#if CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyObject_AsDouble(obj)\
(likely(PyFloat_CheckExact(obj)) ? PyFloat_AS_DOUBLE(obj) :\
 likely(PyInt_CheckExact(obj)) ?\
 PyFloat_AsDouble(obj) : __Pyx__PyObject_AsDouble(obj))
#else
#define __Pyx_PyObject_AsDouble(obj)\
((likely(PyFloat_CheckExact(obj))) ?\
 PyFloat_AS_DOUBLE(obj) : __Pyx__PyObject_AsDouble(obj))
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* IncludeStringH.proto */
#include <string.h>

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell;
static PyObject* __pyx_convert__to_py_struct____pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell(struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell s);
/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_f_14ms_peak_picker_2_c_8peak_set_8PeakBase_clone(CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_f_14ms_peak_picker_2_c_8peak_set_8PeakBase_copy(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set_10FittedPeak__create(double __pyx_v_mz, double __pyx_v_intensity, double __pyx_v_signal_to_noise, double __pyx_v_full_width_at_half_max, double __pyx_v_left_width, double __pyx_v_right_width, long __pyx_v_peak_count, long __pyx_v_index, double __pyx_v_area); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_f_14ms_peak_picker_2_c_8peak_set_10FittedPeak_clone(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_14ms_peak_picker_2_c_8peak_set_10FittedPeak__eq(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__create(PyObject *__pyx_v_peaks); /* proto*/
static size_t __pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_get_size(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_reindex(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static size_t __pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__index(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__get_nearest_peak(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_mz, double *__pyx_v_errout); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_mz, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak *__pyx_optional_args); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_getitem(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, size_t __pyx_v_i); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_clone(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_copy(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__between(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_m1, double __pyx_v_m2); /* proto*/
static int __pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__between_bounds(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_m1, double __pyx_v_m2, size_t *__pyx_v_startp, size_t *__pyx_v_endp); /* proto*/
static PyObject *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_mz, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for *__pyx_optional_args); /* proto*/
static int __pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__eq(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__create(PyObject *__pyx_v_peaks); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_clone(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__allocate_index(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, size_t __pyx_v_interval_index_size, int __pyx_skip_dispatch); /* proto*/
static size_t __pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__index(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__has_peak(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, double __pyx_v_mz, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__has_peak *__pyx_optional_args); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__get_nearest_peak(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, double __pyx_v_mz, double *__pyx_v_errout); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, double __pyx_v_mz, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted *__pyx_optional_args); /* proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython.version' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'ms_peak_picker._c.peak_set' */
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = 0;
static size_t __pyx_v_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE;
static double __pyx_v_14ms_peak_picker_2_c_8peak_set_INF;
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak = 0;
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set_binary_search_ppm_error(PyObject *, double, double, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set__binary_search_ppm_error(PyObject *, double, size_t, size_t, double); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set__binary_search_nearest_match(PyObject *, double, size_t, size_t, double *); /*proto*/
static int __pyx_f_14ms_peak_picker_2_c_8peak_set_check_index(struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *); /*proto*/
static size_t __pyx_f_14ms_peak_picker_2_c_8peak_set_double_binary_search_ppm(double *, double, double, size_t); /*proto*/
static size_t __pyx_f_14ms_peak_picker_2_c_8peak_set_double_binary_search_nearest_match(double *, double, size_t); /*proto*/
static size_t __pyx_f_14ms_peak_picker_2_c_8peak_set_double_binary_search_ppm_with_hint(double *, double, double, size_t, size_t); /*proto*/
static double __pyx_f_14ms_peak_picker_2_c_8peak_set_ppm_error(double, double); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set__sweep_solution_ppm_error(PyObject *, double, size_t, size_t, double); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set__sweep_nearest_match(PyObject *, double, size_t, size_t, double *); /*proto*/
static size_t __pyx_f_14ms_peak_picker_2_c_8peak_set_double_binary_search_nearest_match_with_hint(double *, double, size_t, size_t); /*proto*/
static double *__pyx_f_14ms_peak_picker_2_c_8peak_set_build_linear_spaced_array(double, double, size_t); /*proto*/
static void __pyx_f_14ms_peak_picker_2_c_8peak_set_free_index_list(struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *); /*proto*/
static int __pyx_f_14ms_peak_picker_2_c_8peak_set_build_interval_index(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *, size_t); /*proto*/
static size_t __pyx_f_14ms_peak_picker_2_c_8peak_set_interpolate_index(struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *, double); /*proto*/
static int __pyx_f_14ms_peak_picker_2_c_8peak_set_find_search_interval(struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *, double, size_t *, size_t *); /*proto*/
static PyObject *__pyx_f_14ms_peak_picker_2_c_8peak_set___pyx_unpickle_PeakBase__set_state(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "ms_peak_picker._c.peak_set"
extern int __pyx_module_is_main_ms_peak_picker___c__peak_set;
int __pyx_module_is_main_ms_peak_picker___c__peak_set = 0;

/* Implementation of 'ms_peak_picker._c.peak_set' */
static PyObject *__pyx_builtin_sorted;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_NotImplemented;
static PyObject *__pyx_builtin_ValueError;
static const char __pyx_k_eq[] = "_eq";
static const char __pyx_k_m1[] = "m1";
static const char __pyx_k_m2[] = "m2";
static const char __pyx_k_mz[] = "mz";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_area[] = "area";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_clone[] = "clone";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_peaks[] = "peaks";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_sorted[] = "sorted";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_PeakSet[] = "PeakSet";
static const char __pyx_k_index_2[] = "_index";
static const char __pyx_k_reindex[] = "reindex";
static const char __pyx_k_PeakBase[] = "PeakBase";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_hint_end[] = "hint_end";
static const char __pyx_k_operator[] = "operator";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_intensity[] = "intensity";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_tolerance[] = "tolerance";
static const char __pyx_k_FittedPeak[] = "FittedPeak";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_attrgetter[] = "attrgetter";
static const char __pyx_k_hint_start[] = "hint_start";
static const char __pyx_k_left_width[] = "left_width";
static const char __pyx_k_peak_count[] = "peak_count";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_right_width[] = "right_width";
static const char __pyx_k_center_value[] = "center_value";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_all_peaks_for[] = "all_peaks_for";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_NotImplemented[] = "NotImplemented";
static const char __pyx_k_PeakSetIndexed[] = "PeakSetIndexed";
static const char __pyx_k_allocate_index[] = "_allocate_index";
static const char __pyx_k_PeakSet_d_Peaks[] = "<PeakSet %d Peaks>";
static const char __pyx_k_error_tolerance[] = "error_tolerance";
static const char __pyx_k_has_peak_hinted[] = "has_peak_hinted";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_signal_to_noise[] = "signal_to_noise";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_INTERVAL_INDEX_SIZE[] = "INTERVAL_INDEX_SIZE";
static const char __pyx_k_pyx_unpickle_PeakBase[] = "__pyx_unpickle_PeakBase";
static const char __pyx_k_full_width_at_half_max[] = "full_width_at_half_max";
static const char __pyx_k_ms_peak_picker__c_peak_set[] = "ms_peak_picker._c.peak_set";
static const char __pyx_k_FittedPeak_mz_0_3f_intensity_0_3[] = "FittedPeak(mz=%0.3f, intensity=%0.3f, signal_to_noise=%0.3f, peak_count=%d, index=%d, full_width_at_half_max=%0.3f, area=%0.3f)";
static const char __pyx_k_Hinted_start_index_d_cannot_be_g[] = "Hinted start index %d cannot be greater than the set size %d";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0x4f329c4, 0x9e542aa, 0x4c1f989) = (area, intensity, mz))";
static PyObject *__pyx_n_s_FittedPeak;
static PyObject *__pyx_kp_s_FittedPeak_mz_0_3f_intensity_0_3;
static PyObject *__pyx_kp_s_Hinted_start_index_d_cannot_be_g;
static PyObject *__pyx_n_s_INTERVAL_INDEX_SIZE;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
static PyObject *__pyx_n_s_NotImplemented;
static PyObject *__pyx_n_s_PeakBase;
static PyObject *__pyx_n_s_PeakSet;
static PyObject *__pyx_n_s_PeakSetIndexed;
static PyObject *__pyx_kp_s_PeakSet_d_Peaks;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_all_peaks_for;
static PyObject *__pyx_n_s_allocate_index;
static PyObject *__pyx_n_s_area;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_attrgetter;
static PyObject *__pyx_n_s_center_value;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_clone;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_eq;
static PyObject *__pyx_n_s_error_tolerance;
static PyObject *__pyx_n_s_full_width_at_half_max;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_has_peak_hinted;
static PyObject *__pyx_n_s_hint_end;
static PyObject *__pyx_n_s_hint_start;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_index;
static PyObject *__pyx_n_s_index_2;
static PyObject *__pyx_n_s_inf;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_intensity;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_left_width;
static PyObject *__pyx_n_s_m1;
static PyObject *__pyx_n_s_m2;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_ms_peak_picker__c_peak_set;
static PyObject *__pyx_n_s_mz;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_operator;
static PyObject *__pyx_n_s_peak_count;
static PyObject *__pyx_n_s_peaks;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_capi;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_PeakBase;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_reindex;
static PyObject *__pyx_n_s_right_width;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_signal_to_noise;
static PyObject *__pyx_n_s_sorted;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_tolerance;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_clone(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_2copy(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_2mz___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self); /* proto */
static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_2mz_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_9intensity___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self); /* proto */
static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_9intensity_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_4area___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self); /* proto */
static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_4area_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_4__reduce_cython__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_6__setstate_cython__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak___init__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_mz, PyObject *__pyx_v_intensity, PyObject *__pyx_v_signal_to_noise, PyObject *__pyx_v_peak_count, PyObject *__pyx_v_index, PyObject *__pyx_v_full_width_at_half_max, PyObject *__pyx_v_area, PyObject *__pyx_v_left_width, PyObject *__pyx_v_right_width); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_2clone(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_4__repr__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_6__getstate__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_8__setstate__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10__reduce__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self); /* proto */
static Py_hash_t __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_12__hash__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_14_eq(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_16__richcmp__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_15signal_to_noise___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self); /* proto */
static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_15signal_to_noise_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_22full_width_at_half_max___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self); /* proto */
static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_22full_width_at_half_max_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10left_width___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self); /* proto */
static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10left_width_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_11right_width___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self); /* proto */
static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_11right_width_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10peak_count___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self); /* proto */
static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10peak_count_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_5index___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self); /* proto */
static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_5index_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet___init__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, PyObject *__pyx_v_peaks); /* proto */
static Py_ssize_t __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_2__len__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_4reindex(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_6_index(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_8has_peak(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_mz, double __pyx_v_tolerance); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_10get_nearest_peak(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_mz); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_12__repr__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_14__getitem__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_16clone(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_18copy(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_20between(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_m1, double __pyx_v_m2); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_22all_peaks_for(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_24__getstate__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_26__setstate__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_28__reduce__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_30_eq(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_32__richcmp__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self); /* proto */
static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks_4__del__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_7indexed___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self); /* proto */
static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_7indexed_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_binary_search_ppm_error(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_array, double __pyx_v_value, double __pyx_v_tolerance); /* proto */
static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed___init__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, PyObject *__pyx_v_peaks); /* proto */
static void __pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_2__dealloc__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_4clone(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_6_allocate_index(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, size_t __pyx_v_interval_index_size); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_8_index(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_10get_nearest_peak(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, double __pyx_v_mz); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_12has_peak_hinted(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, double __pyx_v_mz, double __pyx_v_tolerance, size_t __pyx_v_hint_start, size_t __pyx_v_hint_end); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_14test_interval(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_16find_interval_for(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_18check_interval(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, size_t __pyx_v_i); /* proto */
static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_2__pyx_unpickle_PeakBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_14ms_peak_picker_2_c_8peak_set_PeakBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14ms_peak_picker_2_c_8peak_set_FittedPeak(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14ms_peak_picker_2_c_8peak_set_PeakSet(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_79821193;
static PyObject *__pyx_int_83044804;
static PyObject *__pyx_int_166019754;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_codeobj__5;
/* Late includes */

/* "ms_peak_picker/_c/peak_set.pyx":11
 * from cpython cimport PyObject
 * 
 * cdef double ppm_error(double x, double y):             # <<<<<<<<<<<<<<
 *     return (x - y) / y
 * 
 */

static double __pyx_f_14ms_peak_picker_2_c_8peak_set_ppm_error(double __pyx_v_x, double __pyx_v_y) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ppm_error", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":12
 * 
 * cdef double ppm_error(double x, double y):
 *     return (x - y) / y             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = (__pyx_v_x - __pyx_v_y);
  if (unlikely(__pyx_v_y == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_r = (__pyx_t_1 / __pyx_v_y);
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":11
 * from cpython cimport PyObject
 * 
 * cdef double ppm_error(double x, double y):             # <<<<<<<<<<<<<<
 *     return (x - y) / y
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("ms_peak_picker._c.peak_set.ppm_error", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":21
 * cdef class PeakBase:
 * 
 *     cpdef PeakBase clone(self):             # <<<<<<<<<<<<<<
 *         return PeakBase()
 * 
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_1clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_f_14ms_peak_picker_2_c_8peak_set_8PeakBase_clone(CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_clone); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_1clone)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase))))) __PYX_ERR(0, 21, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_peak_picker/_c/peak_set.pyx":22
 * 
 *     cpdef PeakBase clone(self):
 *         return PeakBase()             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakBase copy(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":21
 * cdef class PeakBase:
 * 
 *     cpdef PeakBase clone(self):             # <<<<<<<<<<<<<<
 *         return PeakBase()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakBase.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_1clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_8PeakBase_clone[] = "PeakBase.clone(self) -> PeakBase";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_1clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_clone(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_clone(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set_8PeakBase_clone(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakBase.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":24
 *         return PeakBase()
 * 
 *     cpdef PeakBase copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone()
 * 
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_3copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_f_14ms_peak_picker_2_c_8peak_set_8PeakBase_copy(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_3copy)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 24, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase))))) __PYX_ERR(0, 24, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_peak_picker/_c/peak_set.pyx":25
 * 
 *     cpdef PeakBase copy(self):
 *         return self.clone()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_self->__pyx_vtab)->clone(__pyx_v_self, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":24
 *         return PeakBase()
 * 
 *     cpdef PeakBase copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakBase.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_3copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_8PeakBase_2copy[] = "PeakBase.copy(self) -> PeakBase";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_3copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_2copy(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_2copy(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set_8PeakBase_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakBase.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pxd":4
 * cdef class PeakBase(object):
 *     cdef:
 *         public double mz             # <<<<<<<<<<<<<<
 *         public double intensity
 *         public double area
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_2mz_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_2mz_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_2mz___get__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_2mz___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakBase.mz.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_2mz_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_2mz_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_2mz_2__set__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_2mz_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L1_error)
  __pyx_v_self->mz = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakBase.mz.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pxd":5
 *     cdef:
 *         public double mz
 *         public double intensity             # <<<<<<<<<<<<<<
 *         public double area
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_9intensity_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_9intensity_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_9intensity___get__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_9intensity___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->intensity); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakBase.intensity.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_9intensity_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_9intensity_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_9intensity_2__set__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_9intensity_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 5, __pyx_L1_error)
  __pyx_v_self->intensity = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakBase.intensity.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pxd":6
 *         public double mz
 *         public double intensity
 *         public double area             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakBase clone(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_4area_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_4area_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_4area___get__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_4area___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->area); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakBase.area.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_4area_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_4area_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_4area_2__set__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_4area_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 6, __pyx_L1_error)
  __pyx_v_self->area = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakBase.area.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_8PeakBase_4__reduce_cython__[] = "PeakBase.__reduce_cython__(self)";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_4__reduce_cython__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_4__reduce_cython__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.area, self.intensity, self.mz)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->area); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->intensity); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.area, self.intensity, self.mz)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_4 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v__dict = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":7
 *     state = (self.area, self.intensity, self.mz)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_5 = (__pyx_v__dict != Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v__dict);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.area, self.intensity, self.mz)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_PeakBase, (type(self), 0x4f329c4, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PeakBase, (type(self), 0x4f329c4, None), state
 *     else:
 */
  __pyx_t_6 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_PeakBase, (type(self), 0x4f329c4, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_PeakBase, (type(self), 0x4f329c4, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_PeakBase); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_83044804);
    __Pyx_GIVEREF(__pyx_int_83044804);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_83044804);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_4, 2, Py_None);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PeakBase, (type(self), 0x4f329c4, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_PeakBase, (type(self), 0x4f329c4, None), state
 *     else:
 *         return __pyx_unpickle_PeakBase, (type(self), 0x4f329c4, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PeakBase__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_PeakBase); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_83044804);
    __Pyx_GIVEREF(__pyx_int_83044804);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_83044804);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PeakBase, (type(self), 0x4f329c4, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PeakBase__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_8PeakBase_6__setstate_cython__[] = "PeakBase.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_6__setstate_cython__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_8PeakBase_6__setstate_cython__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_PeakBase, (type(self), 0x4f329c4, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PeakBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_14ms_peak_picker_2_c_8peak_set___pyx_unpickle_PeakBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PeakBase, (type(self), 0x4f329c4, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PeakBase__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":59
 *         The right-sided width at half of max
 *     """
 *     def __init__(self, mz, intensity, signal_to_noise, peak_count, index, full_width_at_half_max,             # <<<<<<<<<<<<<<
 *                  area, left_width=0, right_width=0):
 *             self.mz = mz
 */

/* Python wrapper */
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_mz = 0;
  PyObject *__pyx_v_intensity = 0;
  PyObject *__pyx_v_signal_to_noise = 0;
  PyObject *__pyx_v_peak_count = 0;
  PyObject *__pyx_v_index = 0;
  PyObject *__pyx_v_full_width_at_half_max = 0;
  PyObject *__pyx_v_area = 0;
  PyObject *__pyx_v_left_width = 0;
  PyObject *__pyx_v_right_width = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_intensity,&__pyx_n_s_signal_to_noise,&__pyx_n_s_peak_count,&__pyx_n_s_index,&__pyx_n_s_full_width_at_half_max,&__pyx_n_s_area,&__pyx_n_s_left_width,&__pyx_n_s_right_width,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    values[7] = ((PyObject *)__pyx_int_0);
    values[8] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_intensity)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 7, 9, 1); __PYX_ERR(0, 59, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_signal_to_noise)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 7, 9, 2); __PYX_ERR(0, 59, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_count)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 7, 9, 3); __PYX_ERR(0, 59, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 7, 9, 4); __PYX_ERR(0, 59, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_full_width_at_half_max)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 7, 9, 5); __PYX_ERR(0, 59, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_area)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 7, 9, 6); __PYX_ERR(0, 59, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left_width);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right_width);
          if (value) { values[8] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 59, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_mz = values[0];
    __pyx_v_intensity = values[1];
    __pyx_v_signal_to_noise = values[2];
    __pyx_v_peak_count = values[3];
    __pyx_v_index = values[4];
    __pyx_v_full_width_at_half_max = values[5];
    __pyx_v_area = values[6];
    __pyx_v_left_width = values[7];
    __pyx_v_right_width = values[8];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 7, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 59, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak___init__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self), __pyx_v_mz, __pyx_v_intensity, __pyx_v_signal_to_noise, __pyx_v_peak_count, __pyx_v_index, __pyx_v_full_width_at_half_max, __pyx_v_area, __pyx_v_left_width, __pyx_v_right_width);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak___init__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_mz, PyObject *__pyx_v_intensity, PyObject *__pyx_v_signal_to_noise, PyObject *__pyx_v_peak_count, PyObject *__pyx_v_index, PyObject *__pyx_v_full_width_at_half_max, PyObject *__pyx_v_area, PyObject *__pyx_v_left_width, PyObject *__pyx_v_right_width) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  long __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":61
 *     def __init__(self, mz, intensity, signal_to_noise, peak_count, index, full_width_at_half_max,
 *                  area, left_width=0, right_width=0):
 *             self.mz = mz             # <<<<<<<<<<<<<<
 *             self.intensity = intensity
 *             self.signal_to_noise = signal_to_noise
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_mz); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L1_error)
  __pyx_v_self->__pyx_base.mz = __pyx_t_1;

  /* "ms_peak_picker/_c/peak_set.pyx":62
 *                  area, left_width=0, right_width=0):
 *             self.mz = mz
 *             self.intensity = intensity             # <<<<<<<<<<<<<<
 *             self.signal_to_noise = signal_to_noise
 *             self.peak_count = peak_count
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_intensity); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 62, __pyx_L1_error)
  __pyx_v_self->__pyx_base.intensity = __pyx_t_1;

  /* "ms_peak_picker/_c/peak_set.pyx":63
 *             self.mz = mz
 *             self.intensity = intensity
 *             self.signal_to_noise = signal_to_noise             # <<<<<<<<<<<<<<
 *             self.peak_count = peak_count
 *             self.index = index
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_signal_to_noise); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 63, __pyx_L1_error)
  __pyx_v_self->signal_to_noise = __pyx_t_1;

  /* "ms_peak_picker/_c/peak_set.pyx":64
 *             self.intensity = intensity
 *             self.signal_to_noise = signal_to_noise
 *             self.peak_count = peak_count             # <<<<<<<<<<<<<<
 *             self.index = index
 *             self.full_width_at_half_max = full_width_at_half_max
 */
  __pyx_t_2 = __Pyx_PyInt_As_long(__pyx_v_peak_count); if (unlikely((__pyx_t_2 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 64, __pyx_L1_error)
  __pyx_v_self->peak_count = __pyx_t_2;

  /* "ms_peak_picker/_c/peak_set.pyx":65
 *             self.signal_to_noise = signal_to_noise
 *             self.peak_count = peak_count
 *             self.index = index             # <<<<<<<<<<<<<<
 *             self.full_width_at_half_max = full_width_at_half_max
 *             self.area = area
 */
  __pyx_t_2 = __Pyx_PyInt_As_long(__pyx_v_index); if (unlikely((__pyx_t_2 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 65, __pyx_L1_error)
  __pyx_v_self->index = __pyx_t_2;

  /* "ms_peak_picker/_c/peak_set.pyx":66
 *             self.peak_count = peak_count
 *             self.index = index
 *             self.full_width_at_half_max = full_width_at_half_max             # <<<<<<<<<<<<<<
 *             self.area = area
 *             self.left_width = left_width
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_full_width_at_half_max); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 66, __pyx_L1_error)
  __pyx_v_self->full_width_at_half_max = __pyx_t_1;

  /* "ms_peak_picker/_c/peak_set.pyx":67
 *             self.index = index
 *             self.full_width_at_half_max = full_width_at_half_max
 *             self.area = area             # <<<<<<<<<<<<<<
 *             self.left_width = left_width
 *             self.right_width = right_width
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_area); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 67, __pyx_L1_error)
  __pyx_v_self->__pyx_base.area = __pyx_t_1;

  /* "ms_peak_picker/_c/peak_set.pyx":68
 *             self.full_width_at_half_max = full_width_at_half_max
 *             self.area = area
 *             self.left_width = left_width             # <<<<<<<<<<<<<<
 *             self.right_width = right_width
 * 
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_left_width); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 68, __pyx_L1_error)
  __pyx_v_self->left_width = __pyx_t_1;

  /* "ms_peak_picker/_c/peak_set.pyx":69
 *             self.area = area
 *             self.left_width = left_width
 *             self.right_width = right_width             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_right_width); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 69, __pyx_L1_error)
  __pyx_v_self->right_width = __pyx_t_1;

  /* "ms_peak_picker/_c/peak_set.pyx":59
 *         The right-sided width at half of max
 *     """
 *     def __init__(self, mz, intensity, signal_to_noise, peak_count, index, full_width_at_half_max,             # <<<<<<<<<<<<<<
 *                  area, left_width=0, right_width=0):
 *             self.mz = mz
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":72
 * 
 *     @staticmethod
 *     cdef FittedPeak _create(double mz, double intensity, double signal_to_noise,             # <<<<<<<<<<<<<<
 *                             double full_width_at_half_max, double left_width,
 *                             double right_width, long peak_count, long index,
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set_10FittedPeak__create(double __pyx_v_mz, double __pyx_v_intensity, double __pyx_v_signal_to_noise, double __pyx_v_full_width_at_half_max, double __pyx_v_left_width, double __pyx_v_right_width, long __pyx_v_peak_count, long __pyx_v_index, double __pyx_v_area) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_inst = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":79
 *             FittedPeak inst
 * 
 *         inst = FittedPeak.__new__(FittedPeak)             # <<<<<<<<<<<<<<
 *         inst.mz = mz
 *         inst.intensity = intensity
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_14ms_peak_picker_2_c_8peak_set_FittedPeak(((PyTypeObject *)__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_inst = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":80
 * 
 *         inst = FittedPeak.__new__(FittedPeak)
 *         inst.mz = mz             # <<<<<<<<<<<<<<
 *         inst.intensity = intensity
 *         inst.signal_to_noise = signal_to_noise
 */
  __pyx_v_inst->__pyx_base.mz = __pyx_v_mz;

  /* "ms_peak_picker/_c/peak_set.pyx":81
 *         inst = FittedPeak.__new__(FittedPeak)
 *         inst.mz = mz
 *         inst.intensity = intensity             # <<<<<<<<<<<<<<
 *         inst.signal_to_noise = signal_to_noise
 *         inst.full_width_at_half_max = full_width_at_half_max
 */
  __pyx_v_inst->__pyx_base.intensity = __pyx_v_intensity;

  /* "ms_peak_picker/_c/peak_set.pyx":82
 *         inst.mz = mz
 *         inst.intensity = intensity
 *         inst.signal_to_noise = signal_to_noise             # <<<<<<<<<<<<<<
 *         inst.full_width_at_half_max = full_width_at_half_max
 *         inst.left_width = left_width
 */
  __pyx_v_inst->signal_to_noise = __pyx_v_signal_to_noise;

  /* "ms_peak_picker/_c/peak_set.pyx":83
 *         inst.intensity = intensity
 *         inst.signal_to_noise = signal_to_noise
 *         inst.full_width_at_half_max = full_width_at_half_max             # <<<<<<<<<<<<<<
 *         inst.left_width = left_width
 *         inst.right_width = right_width
 */
  __pyx_v_inst->full_width_at_half_max = __pyx_v_full_width_at_half_max;

  /* "ms_peak_picker/_c/peak_set.pyx":84
 *         inst.signal_to_noise = signal_to_noise
 *         inst.full_width_at_half_max = full_width_at_half_max
 *         inst.left_width = left_width             # <<<<<<<<<<<<<<
 *         inst.right_width = right_width
 *         inst.peak_count = peak_count
 */
  __pyx_v_inst->left_width = __pyx_v_left_width;

  /* "ms_peak_picker/_c/peak_set.pyx":85
 *         inst.full_width_at_half_max = full_width_at_half_max
 *         inst.left_width = left_width
 *         inst.right_width = right_width             # <<<<<<<<<<<<<<
 *         inst.peak_count = peak_count
 *         inst.index = index
 */
  __pyx_v_inst->right_width = __pyx_v_right_width;

  /* "ms_peak_picker/_c/peak_set.pyx":86
 *         inst.left_width = left_width
 *         inst.right_width = right_width
 *         inst.peak_count = peak_count             # <<<<<<<<<<<<<<
 *         inst.index = index
 *         inst.area = area
 */
  __pyx_v_inst->peak_count = __pyx_v_peak_count;

  /* "ms_peak_picker/_c/peak_set.pyx":87
 *         inst.right_width = right_width
 *         inst.peak_count = peak_count
 *         inst.index = index             # <<<<<<<<<<<<<<
 *         inst.area = area
 *         return inst
 */
  __pyx_v_inst->index = __pyx_v_index;

  /* "ms_peak_picker/_c/peak_set.pyx":88
 *         inst.peak_count = peak_count
 *         inst.index = index
 *         inst.area = area             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_v_inst->__pyx_base.area = __pyx_v_area;

  /* "ms_peak_picker/_c/peak_set.pyx":89
 *         inst.index = index
 *         inst.area = area
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakBase clone(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_inst));
  __pyx_r = __pyx_v_inst;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":72
 * 
 *     @staticmethod
 *     cdef FittedPeak _create(double mz, double intensity, double signal_to_noise,             # <<<<<<<<<<<<<<
 *                             double full_width_at_half_max, double left_width,
 *                             double right_width, long peak_count, long index,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":91
 *         return inst
 * 
 *     cpdef PeakBase clone(self):             # <<<<<<<<<<<<<<
 *         return FittedPeak._create(
 *             self.mz, self.intensity, self.signal_to_noise,
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_3clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_f_14ms_peak_picker_2_c_8peak_set_10FittedPeak_clone(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_clone); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_3clone)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 91, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase))))) __PYX_ERR(0, 91, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_peak_picker/_c/peak_set.pyx":92
 * 
 *     cpdef PeakBase clone(self):
 *         return FittedPeak._create(             # <<<<<<<<<<<<<<
 *             self.mz, self.intensity, self.signal_to_noise,
 *             self.full_width_at_half_max, self.left_width, self.right_width,
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));

  /* "ms_peak_picker/_c/peak_set.pyx":95
 *             self.mz, self.intensity, self.signal_to_noise,
 *             self.full_width_at_half_max, self.left_width, self.right_width,
 *             self.peak_count, self.index, self.area)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set_10FittedPeak__create(__pyx_v_self->__pyx_base.mz, __pyx_v_self->__pyx_base.intensity, __pyx_v_self->signal_to_noise, __pyx_v_self->full_width_at_half_max, __pyx_v_self->left_width, __pyx_v_self->right_width, __pyx_v_self->peak_count, __pyx_v_self->index, __pyx_v_self->__pyx_base.area)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":91
 *         return inst
 * 
 *     cpdef PeakBase clone(self):             # <<<<<<<<<<<<<<
 *         return FittedPeak._create(
 *             self.mz, self.intensity, self.signal_to_noise,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_3clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_10FittedPeak_2clone[] = "FittedPeak.clone(self) -> PeakBase";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_3clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_2clone(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_2clone(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set_10FittedPeak_clone(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":97
 *             self.peak_count, self.index, self.area)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return ("FittedPeak(mz=%0.3f, intensity=%0.3f, signal_to_noise=%0.3f, peak_count=%d, "
 *                 "index=%d, full_width_at_half_max=%0.3f, area=%0.3f)") % (
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_4__repr__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_4__repr__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":98
 * 
 *     def __repr__(self):
 *         return ("FittedPeak(mz=%0.3f, intensity=%0.3f, signal_to_noise=%0.3f, peak_count=%d, "             # <<<<<<<<<<<<<<
 *                 "index=%d, full_width_at_half_max=%0.3f, area=%0.3f)") % (
 *                 self.mz, self.intensity, self.signal_to_noise,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "ms_peak_picker/_c/peak_set.pyx":100
 *         return ("FittedPeak(mz=%0.3f, intensity=%0.3f, signal_to_noise=%0.3f, peak_count=%d, "
 *                 "index=%d, full_width_at_half_max=%0.3f, area=%0.3f)") % (
 *                 self.mz, self.intensity, self.signal_to_noise,             # <<<<<<<<<<<<<<
 *                 self.peak_count, self.index, self.full_width_at_half_max,
 *                 self.area)
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.intensity); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->signal_to_noise); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "ms_peak_picker/_c/peak_set.pyx":101
 *                 "index=%d, full_width_at_half_max=%0.3f, area=%0.3f)") % (
 *                 self.mz, self.intensity, self.signal_to_noise,
 *                 self.peak_count, self.index, self.full_width_at_half_max,             # <<<<<<<<<<<<<<
 *                 self.area)
 * 
 */
  __pyx_t_4 = __Pyx_PyInt_From_long(__pyx_v_self->peak_count); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_long(__pyx_v_self->index); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_self->full_width_at_half_max); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "ms_peak_picker/_c/peak_set.pyx":102
 *                 self.mz, self.intensity, self.signal_to_noise,
 *                 self.peak_count, self.index, self.full_width_at_half_max,
 *                 self.area)             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __pyx_t_7 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.area); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "ms_peak_picker/_c/peak_set.pyx":100
 *         return ("FittedPeak(mz=%0.3f, intensity=%0.3f, signal_to_noise=%0.3f, peak_count=%d, "
 *                 "index=%d, full_width_at_half_max=%0.3f, area=%0.3f)") % (
 *                 self.mz, self.intensity, self.signal_to_noise,             # <<<<<<<<<<<<<<
 *                 self.peak_count, self.index, self.full_width_at_half_max,
 *                 self.area)
 */
  __pyx_t_8 = PyTuple_New(7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_8, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_8, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_8, 6, __pyx_t_7);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":99
 *     def __repr__(self):
 *         return ("FittedPeak(mz=%0.3f, intensity=%0.3f, signal_to_noise=%0.3f, peak_count=%d, "
 *                 "index=%d, full_width_at_half_max=%0.3f, area=%0.3f)") % (             # <<<<<<<<<<<<<<
 *                 self.mz, self.intensity, self.signal_to_noise,
 *                 self.peak_count, self.index, self.full_width_at_half_max,
 */
  __pyx_t_7 = __Pyx_PyString_Format(__pyx_kp_s_FittedPeak_mz_0_3f_intensity_0_3, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":97
 *             self.peak_count, self.index, self.area)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return ("FittedPeak(mz=%0.3f, intensity=%0.3f, signal_to_noise=%0.3f, peak_count=%d, "
 *                 "index=%d, full_width_at_half_max=%0.3f, area=%0.3f)") % (
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":104
 *                 self.area)
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return (self.mz, self.intensity, self.signal_to_noise,
 *                 self.peak_count, self.index, self.full_width_at_half_max,
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_7__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_10FittedPeak_6__getstate__[] = "FittedPeak.__getstate__(self)";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_7__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_6__getstate__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_6__getstate__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":105
 * 
 *     def __getstate__(self):
 *         return (self.mz, self.intensity, self.signal_to_noise,             # <<<<<<<<<<<<<<
 *                 self.peak_count, self.index, self.full_width_at_half_max,
 *                 self.area, self.left_width, self.right_width)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.intensity); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->signal_to_noise); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "ms_peak_picker/_c/peak_set.pyx":106
 *     def __getstate__(self):
 *         return (self.mz, self.intensity, self.signal_to_noise,
 *                 self.peak_count, self.index, self.full_width_at_half_max,             # <<<<<<<<<<<<<<
 *                 self.area, self.left_width, self.right_width)
 * 
 */
  __pyx_t_4 = __Pyx_PyInt_From_long(__pyx_v_self->peak_count); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_long(__pyx_v_self->index); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_self->full_width_at_half_max); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "ms_peak_picker/_c/peak_set.pyx":107
 *         return (self.mz, self.intensity, self.signal_to_noise,
 *                 self.peak_count, self.index, self.full_width_at_half_max,
 *                 self.area, self.left_width, self.right_width)             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */
  __pyx_t_7 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.area); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyFloat_FromDouble(__pyx_v_self->left_width); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_self->right_width); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "ms_peak_picker/_c/peak_set.pyx":105
 * 
 *     def __getstate__(self):
 *         return (self.mz, self.intensity, self.signal_to_noise,             # <<<<<<<<<<<<<<
 *                 self.peak_count, self.index, self.full_width_at_half_max,
 *                 self.area, self.left_width, self.right_width)
 */
  __pyx_t_10 = PyTuple_New(9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_10, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_10, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_10, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 7, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 8, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":104
 *                 self.area)
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return (self.mz, self.intensity, self.signal_to_noise,
 *                 self.peak_count, self.index, self.full_width_at_half_max,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":109
 *                 self.area, self.left_width, self.right_width)
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         (self.mz, self.intensity, self.signal_to_noise,
 *          self.peak_count, self.index, self.full_width_at_half_max,
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_9__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_10FittedPeak_8__setstate__[] = "FittedPeak.__setstate__(self, state)";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_9__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_8__setstate__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self), ((PyObject *)__pyx_v_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_8__setstate__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  double __pyx_t_12;
  double __pyx_t_13;
  double __pyx_t_14;
  long __pyx_t_15;
  long __pyx_t_16;
  double __pyx_t_17;
  double __pyx_t_18;
  double __pyx_t_19;
  double __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":112
 *         (self.mz, self.intensity, self.signal_to_noise,
 *          self.peak_count, self.index, self.full_width_at_half_max,
 *          self.area, self.left_width, self.right_width) = state             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {
    PyObject* sequence = __pyx_v_state;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 9)) {
      if (size > 9) __Pyx_RaiseTooManyValuesError(9);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 110, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 3); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 4); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 5); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyTuple_GET_ITEM(sequence, 7); 
      __pyx_t_9 = PyTuple_GET_ITEM(sequence, 8); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 3); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 4); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 5); 
      __pyx_t_7 = PyList_GET_ITEM(sequence, 6); 
      __pyx_t_8 = PyList_GET_ITEM(sequence, 7); 
      __pyx_t_9 = PyList_GET_ITEM(sequence, 8); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_t_9);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[9] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7,&__pyx_t_8,&__pyx_t_9};
      for (i=0; i < 9; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 110, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[9] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7,&__pyx_t_8,&__pyx_t_9};
    __pyx_t_10 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = Py_TYPE(__pyx_t_10)->tp_iternext;
    for (index=0; index < 9; index++) {
      PyObject* item = __pyx_t_11(__pyx_t_10); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_10), 9) < 0) __PYX_ERR(0, 110, __pyx_L1_error)
    __pyx_t_11 = NULL;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_11 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 110, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":110
 * 
 *     def __setstate__(self, state):
 *         (self.mz, self.intensity, self.signal_to_noise,             # <<<<<<<<<<<<<<
 *          self.peak_count, self.index, self.full_width_at_half_max,
 *          self.area, self.left_width, self.right_width) = state
 */
  __pyx_t_12 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_12 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_13 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_13 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_15 = __Pyx_PyInt_As_long(__pyx_t_4); if (unlikely((__pyx_t_15 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_16 = __Pyx_PyInt_As_long(__pyx_t_5); if (unlikely((__pyx_t_16 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_6); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_7); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_19 = __pyx_PyFloat_AsDouble(__pyx_t_8); if (unlikely((__pyx_t_19 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_20 = __pyx_PyFloat_AsDouble(__pyx_t_9); if (unlikely((__pyx_t_20 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_self->__pyx_base.mz = __pyx_t_12;
  __pyx_v_self->__pyx_base.intensity = __pyx_t_13;
  __pyx_v_self->signal_to_noise = __pyx_t_14;

  /* "ms_peak_picker/_c/peak_set.pyx":111
 *     def __setstate__(self, state):
 *         (self.mz, self.intensity, self.signal_to_noise,
 *          self.peak_count, self.index, self.full_width_at_half_max,             # <<<<<<<<<<<<<<
 *          self.area, self.left_width, self.right_width) = state
 * 
 */
  __pyx_v_self->peak_count = __pyx_t_15;
  __pyx_v_self->index = __pyx_t_16;
  __pyx_v_self->full_width_at_half_max = __pyx_t_17;

  /* "ms_peak_picker/_c/peak_set.pyx":112
 *         (self.mz, self.intensity, self.signal_to_noise,
 *          self.peak_count, self.index, self.full_width_at_half_max,
 *          self.area, self.left_width, self.right_width) = state             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __pyx_v_self->__pyx_base.area = __pyx_t_18;
  __pyx_v_self->left_width = __pyx_t_19;
  __pyx_v_self->right_width = __pyx_t_20;

  /* "ms_peak_picker/_c/peak_set.pyx":109
 *                 self.area, self.left_width, self.right_width)
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         (self.mz, self.intensity, self.signal_to_noise,
 *          self.peak_count, self.index, self.full_width_at_half_max,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":114
 *          self.area, self.left_width, self.right_width) = state
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return FittedPeak, (self.mz, self.intensity, self.signal_to_noise, self.peak_count,
 *                             self.index, self.full_width_at_half_max, self.left_width,
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_11__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10__reduce__[] = "FittedPeak.__reduce__(self)";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_11__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10__reduce__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10__reduce__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":115
 * 
 *     def __reduce__(self):
 *         return FittedPeak, (self.mz, self.intensity, self.signal_to_noise, self.peak_count,             # <<<<<<<<<<<<<<
 *                             self.index, self.full_width_at_half_max, self.left_width,
 *                             self.right_width)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.intensity); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->signal_to_noise); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_long(__pyx_v_self->peak_count); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "ms_peak_picker/_c/peak_set.pyx":116
 *     def __reduce__(self):
 *         return FittedPeak, (self.mz, self.intensity, self.signal_to_noise, self.peak_count,
 *                             self.index, self.full_width_at_half_max, self.left_width,             # <<<<<<<<<<<<<<
 *                             self.right_width)
 * 
 */
  __pyx_t_5 = __Pyx_PyInt_From_long(__pyx_v_self->index); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_self->full_width_at_half_max); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyFloat_FromDouble(__pyx_v_self->left_width); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "ms_peak_picker/_c/peak_set.pyx":117
 *         return FittedPeak, (self.mz, self.intensity, self.signal_to_noise, self.peak_count,
 *                             self.index, self.full_width_at_half_max, self.left_width,
 *                             self.right_width)             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
  __pyx_t_8 = PyFloat_FromDouble(__pyx_v_self->right_width); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "ms_peak_picker/_c/peak_set.pyx":115
 * 
 *     def __reduce__(self):
 *         return FittedPeak, (self.mz, self.intensity, self.signal_to_noise, self.peak_count,             # <<<<<<<<<<<<<<
 *                             self.index, self.full_width_at_half_max, self.left_width,
 *                             self.right_width)
 */
  __pyx_t_9 = PyTuple_New(8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_9, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_9, 7, __pyx_t_8);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak));
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak));
  PyTuple_SET_ITEM(__pyx_t_8, 0, ((PyObject *)__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_9);
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":114
 *          self.area, self.left_width, self.right_width) = state
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return FittedPeak, (self.mz, self.intensity, self.signal_to_noise, self.peak_count,
 *                             self.index, self.full_width_at_half_max, self.left_width,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":119
 *                             self.right_width)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash(self.mz)
 * 
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_13__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_13__hash__(PyObject *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_12__hash__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_12__hash__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":120
 * 
 *     def __hash__(self):
 *         return hash(self.mz)             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _eq(self, FittedPeak other):
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_hash_t)-1))) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":119
 *                             self.right_width)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash(self.mz)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":122
 *         return hash(self.mz)
 * 
 *     cpdef bint _eq(self, FittedPeak other):             # <<<<<<<<<<<<<<
 *         return (abs(self.mz - other.mz) < 1e-5) and (
 *             abs(self.intensity - other.intensity) < 1e-5) and (
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_15_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_14ms_peak_picker_2_c_8peak_set_10FittedPeak__eq(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_eq", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_eq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 122, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_15_eq)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_other)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 122, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 122, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_peak_picker/_c/peak_set.pyx":123
 * 
 *     cpdef bint _eq(self, FittedPeak other):
 *         return (abs(self.mz - other.mz) < 1e-5) and (             # <<<<<<<<<<<<<<
 *             abs(self.intensity - other.intensity) < 1e-5) and (
 *             abs(self.signal_to_noise - other.signal_to_noise) < 1e-5) and (
 */
  __pyx_t_6 = ((fabs((__pyx_v_self->__pyx_base.mz - __pyx_v_other->__pyx_base.mz)) < 1e-5) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L3_bool_binop_done;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":124
 *     cpdef bint _eq(self, FittedPeak other):
 *         return (abs(self.mz - other.mz) < 1e-5) and (
 *             abs(self.intensity - other.intensity) < 1e-5) and (             # <<<<<<<<<<<<<<
 *             abs(self.signal_to_noise - other.signal_to_noise) < 1e-5) and (
 *             abs(self.full_width_at_half_max - other.full_width_at_half_max) < 1e-5)
 */
  __pyx_t_6 = ((fabs((__pyx_v_self->__pyx_base.intensity - __pyx_v_other->__pyx_base.intensity)) < 1e-5) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L3_bool_binop_done;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":125
 *         return (abs(self.mz - other.mz) < 1e-5) and (
 *             abs(self.intensity - other.intensity) < 1e-5) and (
 *             abs(self.signal_to_noise - other.signal_to_noise) < 1e-5) and (             # <<<<<<<<<<<<<<
 *             abs(self.full_width_at_half_max - other.full_width_at_half_max) < 1e-5)
 * 
 */
  __pyx_t_6 = ((fabs((__pyx_v_self->signal_to_noise - __pyx_v_other->signal_to_noise)) < 1e-5) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L3_bool_binop_done;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":126
 *             abs(self.intensity - other.intensity) < 1e-5) and (
 *             abs(self.signal_to_noise - other.signal_to_noise) < 1e-5) and (
 *             abs(self.full_width_at_half_max - other.full_width_at_half_max) < 1e-5)             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, other, int code):
 */
  __pyx_t_6 = ((fabs((__pyx_v_self->full_width_at_half_max - __pyx_v_other->full_width_at_half_max)) < 1e-5) != 0);
  __pyx_t_5 = __pyx_t_6;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":122
 *         return hash(self.mz)
 * 
 *     cpdef bint _eq(self, FittedPeak other):             # <<<<<<<<<<<<<<
 *         return (abs(self.mz - other.mz) < 1e-5) and (
 *             abs(self.intensity - other.intensity) < 1e-5) and (
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("ms_peak_picker._c.peak_set.FittedPeak._eq", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_15_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_10FittedPeak_14_eq[] = "FittedPeak._eq(self, FittedPeak other) -> bool";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_15_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_eq (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak, 1, "other", 0))) __PYX_ERR(0, 122, __pyx_L1_error)
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_14_eq(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self), ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_14_eq(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_eq", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_14ms_peak_picker_2_c_8peak_set_10FittedPeak__eq(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak._eq", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":128
 *             abs(self.full_width_at_half_max - other.full_width_at_half_max) < 1e-5)
 * 
 *     def __richcmp__(self, other, int code):             # <<<<<<<<<<<<<<
 *         if code == 2:
 *             if other is None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_17__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_17__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_16__richcmp__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((int)__pyx_v_code));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_16__richcmp__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":129
 * 
 *     def __richcmp__(self, other, int code):
 *         if code == 2:             # <<<<<<<<<<<<<<
 *             if other is None:
 *                 return False
 */
  switch (__pyx_v_code) {
    case 2:

    /* "ms_peak_picker/_c/peak_set.pyx":130
 *     def __richcmp__(self, other, int code):
 *         if code == 2:
 *             if other is None:             # <<<<<<<<<<<<<<
 *                 return False
 *             return self._eq(other)
 */
    __pyx_t_1 = (__pyx_v_other == Py_None);
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (__pyx_t_2) {

      /* "ms_peak_picker/_c/peak_set.pyx":131
 *         if code == 2:
 *             if other is None:
 *                 return False             # <<<<<<<<<<<<<<
 *             return self._eq(other)
 *         elif code == 3:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":130
 *     def __richcmp__(self, other, int code):
 *         if code == 2:
 *             if other is None:             # <<<<<<<<<<<<<<
 *                 return False
 *             return self._eq(other)
 */
    }

    /* "ms_peak_picker/_c/peak_set.pyx":132
 *             if other is None:
 *                 return False
 *             return self._eq(other)             # <<<<<<<<<<<<<<
 *         elif code == 3:
 *             if other is None:
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak))))) __PYX_ERR(0, 132, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self->__pyx_base.__pyx_vtab)->_eq(__pyx_v_self, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_other), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":129
 * 
 *     def __richcmp__(self, other, int code):
 *         if code == 2:             # <<<<<<<<<<<<<<
 *             if other is None:
 *                 return False
 */
    break;
    case 3:

    /* "ms_peak_picker/_c/peak_set.pyx":134
 *             return self._eq(other)
 *         elif code == 3:
 *             if other is None:             # <<<<<<<<<<<<<<
 *                 return True
 *             return not self._eq(other)
 */
    __pyx_t_2 = (__pyx_v_other == Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":135
 *         elif code == 3:
 *             if other is None:
 *                 return True             # <<<<<<<<<<<<<<
 *             return not self._eq(other)
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":134
 *             return self._eq(other)
 *         elif code == 3:
 *             if other is None:             # <<<<<<<<<<<<<<
 *                 return True
 *             return not self._eq(other)
 */
    }

    /* "ms_peak_picker/_c/peak_set.pyx":136
 *             if other is None:
 *                 return True
 *             return not self._eq(other)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak))))) __PYX_ERR(0, 136, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyBool_FromLong((!(((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self->__pyx_base.__pyx_vtab)->_eq(__pyx_v_self, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_other), 0) != 0))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":133
 *                 return False
 *             return self._eq(other)
 *         elif code == 3:             # <<<<<<<<<<<<<<
 *             if other is None:
 *                 return True
 */
    break;
    default: break;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":128
 *             abs(self.full_width_at_half_max - other.full_width_at_half_max) < 1e-5)
 * 
 *     def __richcmp__(self, other, int code):             # <<<<<<<<<<<<<<
 *         if code == 2:
 *             if other is None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pxd":14
 * cdef class FittedPeak(PeakBase):
 *     cdef:
 *         public double signal_to_noise             # <<<<<<<<<<<<<<
 *         public double full_width_at_half_max
 *         public double left_width
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_15signal_to_noise_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_15signal_to_noise_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_15signal_to_noise___get__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_15signal_to_noise___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->signal_to_noise); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.signal_to_noise.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_15signal_to_noise_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_15signal_to_noise_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_15signal_to_noise_2__set__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_15signal_to_noise_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 14, __pyx_L1_error)
  __pyx_v_self->signal_to_noise = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.signal_to_noise.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pxd":15
 *     cdef:
 *         public double signal_to_noise
 *         public double full_width_at_half_max             # <<<<<<<<<<<<<<
 *         public double left_width
 *         public double right_width
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_22full_width_at_half_max_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_22full_width_at_half_max_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_22full_width_at_half_max___get__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_22full_width_at_half_max___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->full_width_at_half_max); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.full_width_at_half_max.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_22full_width_at_half_max_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_22full_width_at_half_max_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_22full_width_at_half_max_2__set__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_22full_width_at_half_max_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 15, __pyx_L1_error)
  __pyx_v_self->full_width_at_half_max = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.full_width_at_half_max.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pxd":16
 *         public double signal_to_noise
 *         public double full_width_at_half_max
 *         public double left_width             # <<<<<<<<<<<<<<
 *         public double right_width
 *         public long peak_count
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10left_width_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10left_width_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10left_width___get__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10left_width___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->left_width); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.left_width.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10left_width_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10left_width_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10left_width_2__set__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10left_width_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L1_error)
  __pyx_v_self->left_width = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.left_width.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pxd":17
 *         public double full_width_at_half_max
 *         public double left_width
 *         public double right_width             # <<<<<<<<<<<<<<
 *         public long peak_count
 *         public long index
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_11right_width_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_11right_width_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_11right_width___get__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_11right_width___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->right_width); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.right_width.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_11right_width_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_11right_width_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_11right_width_2__set__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_11right_width_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_v_self->right_width = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.right_width.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pxd":18
 *         public double left_width
 *         public double right_width
 *         public long peak_count             # <<<<<<<<<<<<<<
 *         public long index
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10peak_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10peak_count_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10peak_count___get__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10peak_count___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_self->peak_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.peak_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10peak_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10peak_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10peak_count_2__set__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10peak_count_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_long(__pyx_v_value); if (unlikely((__pyx_t_1 == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 18, __pyx_L1_error)
  __pyx_v_self->peak_count = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.peak_count.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pxd":19
 *         public double right_width
 *         public long peak_count
 *         public long index             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _eq(self, FittedPeak other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_5index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_5index_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_5index___get__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_5index___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_self->index); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_5index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_5index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_5index_2__set__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_10FittedPeak_5index_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_long(__pyx_v_value); if (unlikely((__pyx_t_1 == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 19, __pyx_L1_error)
  __pyx_v_self->index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.FittedPeak.index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":159
 * 
 *     @staticmethod
 *     cdef PeakSet _create(tuple peaks):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PeakSet inst
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__create(PyObject *__pyx_v_peaks) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_inst = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":162
 *         cdef:
 *             PeakSet inst
 *         inst = PeakSet.__new__(PeakSet)             # <<<<<<<<<<<<<<
 *         inst.peaks = peaks
 *         return inst
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_14ms_peak_picker_2_c_8peak_set_PeakSet(((PyTypeObject *)__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_inst = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":163
 *             PeakSet inst
 *         inst = PeakSet.__new__(PeakSet)
 *         inst.peaks = peaks             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __Pyx_INCREF(__pyx_v_peaks);
  __Pyx_GIVEREF(__pyx_v_peaks);
  __Pyx_GOTREF(__pyx_v_inst->peaks);
  __Pyx_DECREF(__pyx_v_inst->peaks);
  __pyx_v_inst->peaks = __pyx_v_peaks;

  /* "ms_peak_picker/_c/peak_set.pyx":164
 *         inst = PeakSet.__new__(PeakSet)
 *         inst.peaks = peaks
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, peaks):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_inst));
  __pyx_r = __pyx_v_inst;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":159
 * 
 *     @staticmethod
 *     cdef PeakSet _create(tuple peaks):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PeakSet inst
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":166
 *         return inst
 * 
 *     def __init__(self, peaks):             # <<<<<<<<<<<<<<
 *         self.peaks = tuple(peaks)
 * 
 */

/* Python wrapper */
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_peaks = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaks,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaks)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 166, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_peaks = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 166, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet___init__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), __pyx_v_peaks);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet___init__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, PyObject *__pyx_v_peaks) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":167
 * 
 *     def __init__(self, peaks):
 *         self.peaks = tuple(peaks)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_v_peaks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->peaks);
  __Pyx_DECREF(__pyx_v_self->peaks);
  __pyx_v_self->peaks = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":166
 *         return inst
 * 
 *     def __init__(self, peaks):             # <<<<<<<<<<<<<<
 *         self.peaks = tuple(peaks)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":169
 *         self.peaks = tuple(peaks)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return PyTuple_GET_SIZE(self.peaks)
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_3__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_2__len__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_2__len__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":170
 * 
 *     def __len__(self):
 *         return PyTuple_GET_SIZE(self.peaks)             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t get_size(self):
 */
  __pyx_t_1 = __pyx_v_self->peaks;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_r = PyTuple_GET_SIZE(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":169
 *         self.peaks = tuple(peaks)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return PyTuple_GET_SIZE(self.peaks)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":172
 *         return PyTuple_GET_SIZE(self.peaks)
 * 
 *     cdef size_t get_size(self):             # <<<<<<<<<<<<<<
 *         return PyTuple_GET_SIZE(self.peaks)
 * 
 */

static size_t __pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_get_size(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self) {
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_size", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":173
 * 
 *     cdef size_t get_size(self):
 *         return PyTuple_GET_SIZE(self.peaks)             # <<<<<<<<<<<<<<
 * 
 *     cpdef reindex(self):
 */
  __pyx_t_1 = __pyx_v_self->peaks;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_r = PyTuple_GET_SIZE(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":172
 *         return PyTuple_GET_SIZE(self.peaks)
 * 
 *     cdef size_t get_size(self):             # <<<<<<<<<<<<<<
 *         return PyTuple_GET_SIZE(self.peaks)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":175
 *         return PyTuple_GET_SIZE(self.peaks)
 * 
 *     cpdef reindex(self):             # <<<<<<<<<<<<<<
 *         """Re-indexes the sequence of peaks, updating their
 *         :attr:`peak_count` and setting their :attr:`index` if
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_5reindex(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_reindex(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reindex", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reindex); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_5reindex)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_peak_picker/_c/peak_set.pyx":180
 *         it is missing.
 *         """
 *         self._index()             # <<<<<<<<<<<<<<
 * 
 *     cpdef size_t _index(self):
 */
  (void)(((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->_index(__pyx_v_self, 0));

  /* "ms_peak_picker/_c/peak_set.pyx":175
 *         return PyTuple_GET_SIZE(self.peaks)
 * 
 *     cpdef reindex(self):             # <<<<<<<<<<<<<<
 *         """Re-indexes the sequence of peaks, updating their
 *         :attr:`peak_count` and setting their :attr:`index` if
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.reindex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_5reindex(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_4reindex[] = "PeakSet.reindex(self)\nRe-indexes the sequence of peaks, updating their\n        :attr:`peak_count` and setting their :attr:`index` if\n        it is missing.\n        ";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_5reindex(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reindex (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_4reindex(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_4reindex(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reindex", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_reindex(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.reindex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":182
 *         self._index()
 * 
 *     cpdef size_t _index(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_7_index(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static size_t __pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__index(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_index", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_index_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_7_index)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 182, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_5 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 182, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_peak_picker/_c/peak_set.pyx":186
 *             size_t i, n
 *             FittedPeak peak
 *         self.peaks = tuple(sorted(self.peaks, key=operator.attrgetter('mz')))             # <<<<<<<<<<<<<<
 *         i = 0
 *         n = self.get_size()
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->peaks);
  __Pyx_GIVEREF(__pyx_v_self->peaks);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->peaks);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_operator); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_attrgetter); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_n_s_mz) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_n_s_mz);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_key, __pyx_t_3) < 0) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_sorted, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->peaks);
  __Pyx_DECREF(__pyx_v_self->peaks);
  __pyx_v_self->peaks = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":187
 *             FittedPeak peak
 *         self.peaks = tuple(sorted(self.peaks, key=operator.attrgetter('mz')))
 *         i = 0             # <<<<<<<<<<<<<<
 *         n = self.get_size()
 *         for i in range(n):
 */
  __pyx_v_i = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":188
 *         self.peaks = tuple(sorted(self.peaks, key=operator.attrgetter('mz')))
 *         i = 0
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             peak = self.getitem(i)
 */
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->get_size(__pyx_v_self);

  /* "ms_peak_picker/_c/peak_set.pyx":189
 *         i = 0
 *         n = self.get_size()
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             peak = self.getitem(i)
 *             peak.peak_count = i
 */
  __pyx_t_5 = __pyx_v_n;
  __pyx_t_7 = __pyx_t_5;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_i = __pyx_t_8;

    /* "ms_peak_picker/_c/peak_set.pyx":190
 *         n = self.get_size()
 *         for i in range(n):
 *             peak = self.getitem(i)             # <<<<<<<<<<<<<<
 *             peak.peak_count = i
 *             if peak.index == -1:
 */
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->getitem(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "ms_peak_picker/_c/peak_set.pyx":191
 *         for i in range(n):
 *             peak = self.getitem(i)
 *             peak.peak_count = i             # <<<<<<<<<<<<<<
 *             if peak.index == -1:
 *                 peak.index = i
 */
    __pyx_v_peak->peak_count = __pyx_v_i;

    /* "ms_peak_picker/_c/peak_set.pyx":192
 *             peak = self.getitem(i)
 *             peak.peak_count = i
 *             if peak.index == -1:             # <<<<<<<<<<<<<<
 *                 peak.index = i
 *         self.indexed = True
 */
    __pyx_t_9 = ((__pyx_v_peak->index == -1L) != 0);
    if (__pyx_t_9) {

      /* "ms_peak_picker/_c/peak_set.pyx":193
 *             peak.peak_count = i
 *             if peak.index == -1:
 *                 peak.index = i             # <<<<<<<<<<<<<<
 *         self.indexed = True
 *         return i
 */
      __pyx_v_peak->index = __pyx_v_i;

      /* "ms_peak_picker/_c/peak_set.pyx":192
 *             peak = self.getitem(i)
 *             peak.peak_count = i
 *             if peak.index == -1:             # <<<<<<<<<<<<<<
 *                 peak.index = i
 *         self.indexed = True
 */
    }
  }

  /* "ms_peak_picker/_c/peak_set.pyx":194
 *             if peak.index == -1:
 *                 peak.index = i
 *         self.indexed = True             # <<<<<<<<<<<<<<
 *         return i
 * 
 */
  __pyx_v_self->indexed = 1;

  /* "ms_peak_picker/_c/peak_set.pyx":195
 *                 peak.index = i
 *         self.indexed = True
 *         return i             # <<<<<<<<<<<<<<
 * 
 *     def has_peak(self, double mz, double tolerance=1e-5):
 */
  __pyx_r = __pyx_v_i;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":182
 *         self._index()
 * 
 *     cpdef size_t _index(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("ms_peak_picker._c.peak_set.PeakSet._index", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_7_index(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_6_index[] = "PeakSet._index(self) -> size_t";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_7_index(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_index (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_6_index(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_6_index(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_index", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__index(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet._index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":197
 *         return i
 * 
 *     def has_peak(self, double mz, double tolerance=1e-5):             # <<<<<<<<<<<<<<
 *         """Search the for the peak nearest to `mz` within
 *         `tolerance` error (in PPM)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_9has_peak(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_8has_peak[] = "PeakSet.has_peak(self, double mz, double tolerance=1e-5)\nSearch the for the peak nearest to `mz` within\n        `tolerance` error (in PPM)\n\n        Parameters\n        ----------\n        mz : float\n            The m/z to search for\n        tolerance : float, optional\n            The error tolerance to accept. Defaults to 1e-5 (10 ppm)\n\n        Returns\n        -------\n        FittedPeak\n            The peak, if found. `None` otherwise.\n        ";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_9has_peak(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_mz;
  double __pyx_v_tolerance;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("has_peak (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_tolerance,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tolerance);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "has_peak") < 0)) __PYX_ERR(0, 197, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_mz = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 197, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 197, __pyx_L3_error)
    } else {
      __pyx_v_tolerance = ((double)1e-5);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("has_peak", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 197, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.has_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_8has_peak(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), __pyx_v_mz, __pyx_v_tolerance);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_8has_peak(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_mz, double __pyx_v_tolerance) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("has_peak", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":213
 *             The peak, if found. `None` otherwise.
 *         """
 *         peak = self._has_peak(mz, tolerance)             # <<<<<<<<<<<<<<
 *         if peak is _null_peak:
 *             return None
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.tolerance = __pyx_v_tolerance;
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->_has_peak(__pyx_v_self, __pyx_v_mz, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":214
 *         """
 *         peak = self._has_peak(mz, tolerance)
 *         if peak is _null_peak:             # <<<<<<<<<<<<<<
 *             return None
 *         return peak
 */
  __pyx_t_3 = (__pyx_v_peak == __pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "ms_peak_picker/_c/peak_set.pyx":215
 *         peak = self._has_peak(mz, tolerance)
 *         if peak is _null_peak:
 *             return None             # <<<<<<<<<<<<<<
 *         return peak
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":214
 *         """
 *         peak = self._has_peak(mz, tolerance)
 *         if peak is _null_peak:             # <<<<<<<<<<<<<<
 *             return None
 *         return peak
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":216
 *         if peak is _null_peak:
 *             return None
 *         return peak             # <<<<<<<<<<<<<<
 * 
 *     def get_nearest_peak(self, double mz):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_peak));
  __pyx_r = ((PyObject *)__pyx_v_peak);
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":197
 *         return i
 * 
 *     def has_peak(self, double mz, double tolerance=1e-5):             # <<<<<<<<<<<<<<
 *         """Search the for the peak nearest to `mz` within
 *         `tolerance` error (in PPM)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.has_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":218
 *         return peak
 * 
 *     def get_nearest_peak(self, double mz):             # <<<<<<<<<<<<<<
 *         cdef:
 *             FittedPeak peak
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_11get_nearest_peak(PyObject *__pyx_v_self, PyObject *__pyx_arg_mz); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_10get_nearest_peak[] = "PeakSet.get_nearest_peak(self, double mz)";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_11get_nearest_peak(PyObject *__pyx_v_self, PyObject *__pyx_arg_mz) {
  double __pyx_v_mz;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_nearest_peak (wrapper)", 0);
  assert(__pyx_arg_mz); {
    __pyx_v_mz = __pyx_PyFloat_AsDouble(__pyx_arg_mz); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 218, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.get_nearest_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_10get_nearest_peak(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), ((double)__pyx_v_mz));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_10get_nearest_peak(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_mz) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  double __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_nearest_peak", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":222
 *             FittedPeak peak
 *             double err
 *         peak = self._get_nearest_peak(mz, &err)             # <<<<<<<<<<<<<<
 *         return peak, err
 * 
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->_get_nearest_peak(__pyx_v_self, __pyx_v_mz, (&__pyx_v_err))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":223
 *             double err
 *         peak = self._get_nearest_peak(mz, &err)
 *         return peak, err             # <<<<<<<<<<<<<<
 * 
 *     cdef FittedPeak _get_nearest_peak(self, double mz, double* errout):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_peak));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_peak));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_peak));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":218
 *         return peak
 * 
 *     def get_nearest_peak(self, double mz):             # <<<<<<<<<<<<<<
 *         cdef:
 *             FittedPeak peak
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.get_nearest_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":225
 *         return peak, err
 * 
 *     cdef FittedPeak _get_nearest_peak(self, double mz, double* errout):             # <<<<<<<<<<<<<<
 *         cdef:
 *             FittedPeak peak
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__get_nearest_peak(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_mz, double *__pyx_v_errout) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_nearest_peak", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":228
 *         cdef:
 *             FittedPeak peak
 *         peak = _binary_search_nearest_match(self.peaks, mz, 0, PyTuple_GET_SIZE(self.peaks), errout)             # <<<<<<<<<<<<<<
 *         return peak
 * 
 */
  __pyx_t_1 = __pyx_v_self->peaks;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_self->peaks;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set__binary_search_nearest_match(((PyObject*)__pyx_t_1), __pyx_v_mz, 0, PyTuple_GET_SIZE(__pyx_t_2), __pyx_v_errout)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":229
 *             FittedPeak peak
 *         peak = _binary_search_nearest_match(self.peaks, mz, 0, PyTuple_GET_SIZE(self.peaks), errout)
 *         return peak             # <<<<<<<<<<<<<<
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=1e-5):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_peak));
  __pyx_r = __pyx_v_peak;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":225
 *         return peak, err
 * 
 *     cdef FittedPeak _get_nearest_peak(self, double mz, double* errout):             # <<<<<<<<<<<<<<
 *         cdef:
 *             FittedPeak peak
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet._get_nearest_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":231
 *         return peak
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=1e-5):             # <<<<<<<<<<<<<<
 *         cdef:
 *             FittedPeak peak
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_mz, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak *__pyx_optional_args) {
  double __pyx_v_tolerance = ((double)1e-5);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_has_peak", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_tolerance = __pyx_optional_args->tolerance;
    }
  }

  /* "ms_peak_picker/_c/peak_set.pyx":234
 *         cdef:
 *             FittedPeak peak
 *         peak = binary_search_ppm_error(self.peaks, mz, tolerance)             # <<<<<<<<<<<<<<
 *         return peak
 * 
 */
  __pyx_t_1 = __pyx_v_self->peaks;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set_binary_search_ppm_error(((PyObject*)__pyx_t_1), __pyx_v_mz, __pyx_v_tolerance, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":235
 *             FittedPeak peak
 *         peak = binary_search_ppm_error(self.peaks, mz, tolerance)
 *         return peak             # <<<<<<<<<<<<<<
 * 
 *     cdef FittedPeak getitem(self, size_t i):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_peak));
  __pyx_r = __pyx_v_peak;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":231
 *         return peak
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=1e-5):             # <<<<<<<<<<<<<<
 *         cdef:
 *             FittedPeak peak
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet._has_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":237
 *         return peak
 * 
 *     cdef FittedPeak getitem(self, size_t i):             # <<<<<<<<<<<<<<
 *         return <FittedPeak>PyTuple_GET_ITEM(self.peaks, i)
 * 
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_getitem(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, size_t __pyx_v_i) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2;
  __Pyx_RefNannySetupContext("getitem", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":238
 * 
 *     cdef FittedPeak getitem(self, size_t i):
 *         return <FittedPeak>PyTuple_GET_ITEM(self.peaks, i)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = __pyx_v_self->peaks;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_v_i);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_INCREF(((PyObject *)((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2)));
  __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2);
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":237
 *         return peak
 * 
 *     cdef FittedPeak getitem(self, size_t i):             # <<<<<<<<<<<<<<
 *         return <FittedPeak>PyTuple_GET_ITEM(self.peaks, i)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":240
 *         return <FittedPeak>PyTuple_GET_ITEM(self.peaks, i)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<PeakSet %d Peaks>" % (len(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_13__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_13__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_12__repr__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_12__repr__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":241
 * 
 *     def __repr__(self):
 *         return "<PeakSet %d Peaks>" % (len(self))             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, item):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 241, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_PeakSet_d_Peaks, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":240
 *         return <FittedPeak>PyTuple_GET_ITEM(self.peaks, i)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<PeakSet %d Peaks>" % (len(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":243
 *         return "<PeakSet %d Peaks>" % (len(self))
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         if isinstance(item, slice):
 *             return PeakSet(self.peaks[item])
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_15__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_15__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_14__getitem__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_14__getitem__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":244
 * 
 *     def __getitem__(self, item):
 *         if isinstance(item, slice):             # <<<<<<<<<<<<<<
 *             return PeakSet(self.peaks[item])
 *         return self.peaks[item]
 */
  __pyx_t_1 = PySlice_Check(__pyx_v_item); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "ms_peak_picker/_c/peak_set.pyx":245
 *     def __getitem__(self, item):
 *         if isinstance(item, slice):
 *             return PeakSet(self.peaks[item])             # <<<<<<<<<<<<<<
 *         return self.peaks[item]
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(__pyx_v_self->peaks == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 245, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_self->peaks, __pyx_v_item); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet), __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":244
 * 
 *     def __getitem__(self, item):
 *         if isinstance(item, slice):             # <<<<<<<<<<<<<<
 *             return PeakSet(self.peaks[item])
 *         return self.peaks[item]
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":246
 *         if isinstance(item, slice):
 *             return PeakSet(self.peaks[item])
 *         return self.peaks[item]             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakSet clone(self):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->peaks == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 246, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_self->peaks, __pyx_v_item); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":243
 *         return "<PeakSet %d Peaks>" % (len(self))
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         if isinstance(item, slice):
 *             return PeakSet(self.peaks[item])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":248
 *         return self.peaks[item]
 * 
 *     cpdef PeakSet clone(self):             # <<<<<<<<<<<<<<
 *         """Creates a deep copy of the sequence of peaks
 * 
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_17clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_clone(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_inst = 0;
  PyObject *__pyx_v_p = NULL;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_clone); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_17clone)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 248, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet))))) __PYX_ERR(0, 248, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_peak_picker/_c/peak_set.pyx":255
 *         PeakSet
 *         """
 *         cdef PeakSet inst = PeakSet([p.clone() for p in self])             # <<<<<<<<<<<<<<
 *         inst.indexed = self.indexed
 *         return inst
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(((PyObject *)__pyx_v_self))) || PyTuple_CheckExact(((PyObject *)__pyx_v_self))) {
    __pyx_t_2 = ((PyObject *)__pyx_v_self); __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 255, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 255, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 255, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 255, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 255, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 255, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_p, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_p, __pyx_n_s_clone); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_inst = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":256
 *         """
 *         cdef PeakSet inst = PeakSet([p.clone() for p in self])
 *         inst.indexed = self.indexed             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_t_8 = __pyx_v_self->indexed;
  __pyx_v_inst->indexed = __pyx_t_8;

  /* "ms_peak_picker/_c/peak_set.pyx":257
 *         cdef PeakSet inst = PeakSet([p.clone() for p in self])
 *         inst.indexed = self.indexed
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakSet copy(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_inst));
  __pyx_r = __pyx_v_inst;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":248
 *         return self.peaks[item]
 * 
 *     cpdef PeakSet clone(self):             # <<<<<<<<<<<<<<
 *         """Creates a deep copy of the sequence of peaks
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XDECREF(__pyx_v_p);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_17clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_16clone[] = "PeakSet.clone(self) -> PeakSet\nCreates a deep copy of the sequence of peaks\n\n        Returns\n        -------\n        PeakSet\n        ";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_17clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_16clone(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_16clone(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_clone(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":259
 *         return inst
 * 
 *     cpdef PeakSet copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone()
 * 
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_19copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_copy(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_19copy)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 259, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet))))) __PYX_ERR(0, 259, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_peak_picker/_c/peak_set.pyx":260
 * 
 *     cpdef PeakSet copy(self):
 *         return self.clone()             # <<<<<<<<<<<<<<
 * 
 *     def between(self, double m1, double m2):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->clone(__pyx_v_self, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":259
 *         return inst
 * 
 *     cpdef PeakSet copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_19copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_18copy[] = "PeakSet.copy(self) -> PeakSet";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_19copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_18copy(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_18copy(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":262
 *         return self.clone()
 * 
 *     def between(self, double m1, double m2):             # <<<<<<<<<<<<<<
 *         """Retrieve a :class:`PeakSet` containing all the peaks
 *         in `self` whose m/z is between `m1` and `m2`.
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_21between(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_20between[] = "PeakSet.between(self, double m1, double m2)\nRetrieve a :class:`PeakSet` containing all the peaks\n        in `self` whose m/z is between `m1` and `m2`.\n\n        These peaks are not copied.\n\n        Parameters\n        ----------\n        m1 : float\n            The lower m/z limit\n        m2 : float\n            The upper m/z limit\n\n        Returns\n        -------\n        PeakSet\n        ";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_21between(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_m1;
  double __pyx_v_m2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("between (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_m1,&__pyx_n_s_m2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("between", 1, 2, 2, 1); __PYX_ERR(0, 262, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "between") < 0)) __PYX_ERR(0, 262, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_m1 = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_m1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 262, __pyx_L3_error)
    __pyx_v_m2 = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_m2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 262, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("between", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 262, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.between", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_20between(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), __pyx_v_m1, __pyx_v_m2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_20between(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_m1, double __pyx_v_m2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("between", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":279
 *         PeakSet
 *         """
 *         return self._between(m1, m2)             # <<<<<<<<<<<<<<
 * 
 *     cdef PeakSet _between(self, double m1, double m2):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->_between(__pyx_v_self, __pyx_v_m1, __pyx_v_m2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":262
 *         return self.clone()
 * 
 *     def between(self, double m1, double m2):             # <<<<<<<<<<<<<<
 *         """Retrieve a :class:`PeakSet` containing all the peaks
 *         in `self` whose m/z is between `m1` and `m2`.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.between", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":281
 *         return self._between(m1, m2)
 * 
 *     cdef PeakSet _between(self, double m1, double m2):             # <<<<<<<<<<<<<<
 *         cdef:
 *             FittedPeak p1
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__between(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_m1, double __pyx_v_m2) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_p1 = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_p2 = 0;
  double __pyx_v_err;
  PyObject *__pyx_v_sliced = 0;
  size_t __pyx_v_start;
  size_t __pyx_v_end;
  size_t __pyx_v_n;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  long __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_between", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":288
 *             tuple sliced
 *             size_t start, end, n
 *         p1 = self._get_nearest_peak(m1, &err)             # <<<<<<<<<<<<<<
 *         p2 = self._get_nearest_peak(m2, &err)
 *         start = p1.peak_count
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->_get_nearest_peak(__pyx_v_self, __pyx_v_m1, (&__pyx_v_err))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_p1 = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":289
 *             size_t start, end, n
 *         p1 = self._get_nearest_peak(m1, &err)
 *         p2 = self._get_nearest_peak(m2, &err)             # <<<<<<<<<<<<<<
 *         start = p1.peak_count
 *         end = p2.peak_count + 1
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->_get_nearest_peak(__pyx_v_self, __pyx_v_m2, (&__pyx_v_err))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_p2 = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":290
 *         p1 = self._get_nearest_peak(m1, &err)
 *         p2 = self._get_nearest_peak(m2, &err)
 *         start = p1.peak_count             # <<<<<<<<<<<<<<
 *         end = p2.peak_count + 1
 *         n = self.get_size()
 */
  __pyx_t_2 = __pyx_v_p1->peak_count;
  __pyx_v_start = __pyx_t_2;

  /* "ms_peak_picker/_c/peak_set.pyx":291
 *         p2 = self._get_nearest_peak(m2, &err)
 *         start = p1.peak_count
 *         end = p2.peak_count + 1             # <<<<<<<<<<<<<<
 *         n = self.get_size()
 *         if p1.mz < m1 and start + 1 < n:
 */
  __pyx_v_end = (__pyx_v_p2->peak_count + 1);

  /* "ms_peak_picker/_c/peak_set.pyx":292
 *         start = p1.peak_count
 *         end = p2.peak_count + 1
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         if p1.mz < m1 and start + 1 < n:
 *             start += 1
 */
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->get_size(__pyx_v_self);

  /* "ms_peak_picker/_c/peak_set.pyx":293
 *         end = p2.peak_count + 1
 *         n = self.get_size()
 *         if p1.mz < m1 and start + 1 < n:             # <<<<<<<<<<<<<<
 *             start += 1
 *         if p2.mz > m2 and end > 0:
 */
  __pyx_t_4 = ((__pyx_v_p1->__pyx_base.mz < __pyx_v_m1) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = (((__pyx_v_start + 1) < __pyx_v_n) != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "ms_peak_picker/_c/peak_set.pyx":294
 *         n = self.get_size()
 *         if p1.mz < m1 and start + 1 < n:
 *             start += 1             # <<<<<<<<<<<<<<
 *         if p2.mz > m2 and end > 0:
 *             end -= 1
 */
    __pyx_v_start = (__pyx_v_start + 1);

    /* "ms_peak_picker/_c/peak_set.pyx":293
 *         end = p2.peak_count + 1
 *         n = self.get_size()
 *         if p1.mz < m1 and start + 1 < n:             # <<<<<<<<<<<<<<
 *             start += 1
 *         if p2.mz > m2 and end > 0:
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":295
 *         if p1.mz < m1 and start + 1 < n:
 *             start += 1
 *         if p2.mz > m2 and end > 0:             # <<<<<<<<<<<<<<
 *             end -= 1
 *         if p2.mz < m1 or p1.mz > m2:
 */
  __pyx_t_4 = ((__pyx_v_p2->__pyx_base.mz > __pyx_v_m2) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_v_end > 0) != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_3) {

    /* "ms_peak_picker/_c/peak_set.pyx":296
 *             start += 1
 *         if p2.mz > m2 and end > 0:
 *             end -= 1             # <<<<<<<<<<<<<<
 *         if p2.mz < m1 or p1.mz > m2:
 *             end = start = 0
 */
    __pyx_v_end = (__pyx_v_end - 1);

    /* "ms_peak_picker/_c/peak_set.pyx":295
 *         if p1.mz < m1 and start + 1 < n:
 *             start += 1
 *         if p2.mz > m2 and end > 0:             # <<<<<<<<<<<<<<
 *             end -= 1
 *         if p2.mz < m1 or p1.mz > m2:
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":297
 *         if p2.mz > m2 and end > 0:
 *             end -= 1
 *         if p2.mz < m1 or p1.mz > m2:             # <<<<<<<<<<<<<<
 *             end = start = 0
 *         sliced = <tuple>PyTuple_GetSlice(self.peaks, start, end)
 */
  __pyx_t_4 = ((__pyx_v_p2->__pyx_base.mz < __pyx_v_m1) != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_v_p1->__pyx_base.mz > __pyx_v_m2) != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L10_bool_binop_done:;
  if (__pyx_t_3) {

    /* "ms_peak_picker/_c/peak_set.pyx":298
 *             end -= 1
 *         if p2.mz < m1 or p1.mz > m2:
 *             end = start = 0             # <<<<<<<<<<<<<<
 *         sliced = <tuple>PyTuple_GetSlice(self.peaks, start, end)
 *         return PeakSet._create(sliced)
 */
    __pyx_v_end = 0;
    __pyx_v_start = 0;

    /* "ms_peak_picker/_c/peak_set.pyx":297
 *         if p2.mz > m2 and end > 0:
 *             end -= 1
 *         if p2.mz < m1 or p1.mz > m2:             # <<<<<<<<<<<<<<
 *             end = start = 0
 *         sliced = <tuple>PyTuple_GetSlice(self.peaks, start, end)
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":299
 *         if p2.mz < m1 or p1.mz > m2:
 *             end = start = 0
 *         sliced = <tuple>PyTuple_GetSlice(self.peaks, start, end)             # <<<<<<<<<<<<<<
 *         return PeakSet._create(sliced)
 * 
 */
  __pyx_t_1 = __pyx_v_self->peaks;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_GetSlice(__pyx_t_1, __pyx_v_start, __pyx_v_end); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_t_5;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_sliced = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":300
 *             end = start = 0
 *         sliced = <tuple>PyTuple_GetSlice(self.peaks, start, end)
 *         return PeakSet._create(sliced)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _between_bounds(self, double m1, double m2, size_t* startp, size_t* endp):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__create(__pyx_v_sliced)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":281
 *         return self._between(m1, m2)
 * 
 *     cdef PeakSet _between(self, double m1, double m2):             # <<<<<<<<<<<<<<
 *         cdef:
 *             FittedPeak p1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet._between", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_p1);
  __Pyx_XDECREF((PyObject *)__pyx_v_p2);
  __Pyx_XDECREF(__pyx_v_sliced);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":302
 *         return PeakSet._create(sliced)
 * 
 *     cdef int _between_bounds(self, double m1, double m2, size_t* startp, size_t* endp):             # <<<<<<<<<<<<<<
 *         cdef:
 *             FittedPeak p1
 */

static int __pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__between_bounds(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_m1, double __pyx_v_m2, size_t *__pyx_v_startp, size_t *__pyx_v_endp) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_p1 = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_p2 = 0;
  double __pyx_v_err;
  size_t __pyx_v_start;
  size_t __pyx_v_end;
  size_t __pyx_v_n;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  long __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_between_bounds", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":309
 *             size_t start, end, n
 * 
 *         p1 = self._get_nearest_peak(m1, &err)             # <<<<<<<<<<<<<<
 *         p2 = self._get_nearest_peak(m2, &err)
 *         start = p1.peak_count
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->_get_nearest_peak(__pyx_v_self, __pyx_v_m1, (&__pyx_v_err))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_p1 = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":310
 * 
 *         p1 = self._get_nearest_peak(m1, &err)
 *         p2 = self._get_nearest_peak(m2, &err)             # <<<<<<<<<<<<<<
 *         start = p1.peak_count
 *         end = p2.peak_count + 1
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->_get_nearest_peak(__pyx_v_self, __pyx_v_m2, (&__pyx_v_err))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_p2 = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":311
 *         p1 = self._get_nearest_peak(m1, &err)
 *         p2 = self._get_nearest_peak(m2, &err)
 *         start = p1.peak_count             # <<<<<<<<<<<<<<
 *         end = p2.peak_count + 1
 *         n = self.get_size()
 */
  __pyx_t_2 = __pyx_v_p1->peak_count;
  __pyx_v_start = __pyx_t_2;

  /* "ms_peak_picker/_c/peak_set.pyx":312
 *         p2 = self._get_nearest_peak(m2, &err)
 *         start = p1.peak_count
 *         end = p2.peak_count + 1             # <<<<<<<<<<<<<<
 *         n = self.get_size()
 *         if p1.mz < m1 and start + 1 < n:
 */
  __pyx_v_end = (__pyx_v_p2->peak_count + 1);

  /* "ms_peak_picker/_c/peak_set.pyx":313
 *         start = p1.peak_count
 *         end = p2.peak_count + 1
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         if p1.mz < m1 and start + 1 < n:
 *             start += 1
 */
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->get_size(__pyx_v_self);

  /* "ms_peak_picker/_c/peak_set.pyx":314
 *         end = p2.peak_count + 1
 *         n = self.get_size()
 *         if p1.mz < m1 and start + 1 < n:             # <<<<<<<<<<<<<<
 *             start += 1
 *         if p2.mz > m2 and end > 0:
 */
  __pyx_t_4 = ((__pyx_v_p1->__pyx_base.mz < __pyx_v_m1) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = (((__pyx_v_start + 1) < __pyx_v_n) != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "ms_peak_picker/_c/peak_set.pyx":315
 *         n = self.get_size()
 *         if p1.mz < m1 and start + 1 < n:
 *             start += 1             # <<<<<<<<<<<<<<
 *         if p2.mz > m2 and end > 0:
 *             end -= 1
 */
    __pyx_v_start = (__pyx_v_start + 1);

    /* "ms_peak_picker/_c/peak_set.pyx":314
 *         end = p2.peak_count + 1
 *         n = self.get_size()
 *         if p1.mz < m1 and start + 1 < n:             # <<<<<<<<<<<<<<
 *             start += 1
 *         if p2.mz > m2 and end > 0:
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":316
 *         if p1.mz < m1 and start + 1 < n:
 *             start += 1
 *         if p2.mz > m2 and end > 0:             # <<<<<<<<<<<<<<
 *             end -= 1
 *         if p2.mz < m1 or p1.mz > m2:
 */
  __pyx_t_4 = ((__pyx_v_p2->__pyx_base.mz > __pyx_v_m2) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_v_end > 0) != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_3) {

    /* "ms_peak_picker/_c/peak_set.pyx":317
 *             start += 1
 *         if p2.mz > m2 and end > 0:
 *             end -= 1             # <<<<<<<<<<<<<<
 *         if p2.mz < m1 or p1.mz > m2:
 *             end = start = 0
 */
    __pyx_v_end = (__pyx_v_end - 1);

    /* "ms_peak_picker/_c/peak_set.pyx":316
 *         if p1.mz < m1 and start + 1 < n:
 *             start += 1
 *         if p2.mz > m2 and end > 0:             # <<<<<<<<<<<<<<
 *             end -= 1
 *         if p2.mz < m1 or p1.mz > m2:
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":318
 *         if p2.mz > m2 and end > 0:
 *             end -= 1
 *         if p2.mz < m1 or p1.mz > m2:             # <<<<<<<<<<<<<<
 *             end = start = 0
 *         startp[0] = start
 */
  __pyx_t_4 = ((__pyx_v_p2->__pyx_base.mz < __pyx_v_m1) != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_v_p1->__pyx_base.mz > __pyx_v_m2) != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L10_bool_binop_done:;
  if (__pyx_t_3) {

    /* "ms_peak_picker/_c/peak_set.pyx":319
 *             end -= 1
 *         if p2.mz < m1 or p1.mz > m2:
 *             end = start = 0             # <<<<<<<<<<<<<<
 *         startp[0] = start
 *         endp[0] = end
 */
    __pyx_v_end = 0;
    __pyx_v_start = 0;

    /* "ms_peak_picker/_c/peak_set.pyx":318
 *         if p2.mz > m2 and end > 0:
 *             end -= 1
 *         if p2.mz < m1 or p1.mz > m2:             # <<<<<<<<<<<<<<
 *             end = start = 0
 *         startp[0] = start
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":320
 *         if p2.mz < m1 or p1.mz > m2:
 *             end = start = 0
 *         startp[0] = start             # <<<<<<<<<<<<<<
 *         endp[0] = end
 *         return 0
 */
  (__pyx_v_startp[0]) = __pyx_v_start;

  /* "ms_peak_picker/_c/peak_set.pyx":321
 *             end = start = 0
 *         startp[0] = start
 *         endp[0] = end             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  (__pyx_v_endp[0]) = __pyx_v_end;

  /* "ms_peak_picker/_c/peak_set.pyx":322
 *         startp[0] = start
 *         endp[0] = end
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=2e-5):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":302
 *         return PeakSet._create(sliced)
 * 
 *     cdef int _between_bounds(self, double m1, double m2, size_t* startp, size_t* endp):             # <<<<<<<<<<<<<<
 *         cdef:
 *             FittedPeak p1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("ms_peak_picker._c.peak_set.PeakSet._between_bounds", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_p1);
  __Pyx_XDECREF((PyObject *)__pyx_v_p2);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":324
 *         return 0
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         """Find all peaks within `error_tolerance` ppm of `mz` m/z.
 * 
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_23all_peaks_for(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_mz, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for *__pyx_optional_args) {
  double __pyx_v_error_tolerance = ((double)2e-5);
  double __pyx_v_lo;
  double __pyx_v_hi;
  size_t __pyx_v_start;
  size_t __pyx_v_end;
  PyObject *__pyx_v_sliced = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("all_peaks_for", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_all_peaks_for); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_23all_peaks_for)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 324, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 324, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 324, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 324, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 324, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 324, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (!(likely(PyTuple_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 324, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_peak_picker/_c/peak_set.pyx":343
 *             tuple sliced
 * 
 *         lo = mz - (mz * error_tolerance)             # <<<<<<<<<<<<<<
 *         hi = mz + (mz * error_tolerance)
 *         self._between_bounds(lo, hi, &start, &end)
 */
  __pyx_v_lo = (__pyx_v_mz - (__pyx_v_mz * __pyx_v_error_tolerance));

  /* "ms_peak_picker/_c/peak_set.pyx":344
 * 
 *         lo = mz - (mz * error_tolerance)
 *         hi = mz + (mz * error_tolerance)             # <<<<<<<<<<<<<<
 *         self._between_bounds(lo, hi, &start, &end)
 *         sliced = <tuple>PyTuple_GetSlice(self.peaks, start, end)
 */
  __pyx_v_hi = (__pyx_v_mz + (__pyx_v_mz * __pyx_v_error_tolerance));

  /* "ms_peak_picker/_c/peak_set.pyx":345
 *         lo = mz - (mz * error_tolerance)
 *         hi = mz + (mz * error_tolerance)
 *         self._between_bounds(lo, hi, &start, &end)             # <<<<<<<<<<<<<<
 *         sliced = <tuple>PyTuple_GetSlice(self.peaks, start, end)
 *         return sliced
 */
  (void)(((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->_between_bounds(__pyx_v_self, __pyx_v_lo, __pyx_v_hi, (&__pyx_v_start), (&__pyx_v_end)));

  /* "ms_peak_picker/_c/peak_set.pyx":346
 *         hi = mz + (mz * error_tolerance)
 *         self._between_bounds(lo, hi, &start, &end)
 *         sliced = <tuple>PyTuple_GetSlice(self.peaks, start, end)             # <<<<<<<<<<<<<<
 *         return sliced
 * 
 */
  __pyx_t_1 = __pyx_v_self->peaks;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_GetSlice(__pyx_t_1, __pyx_v_start, __pyx_v_end); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_sliced = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":347
 *         self._between_bounds(lo, hi, &start, &end)
 *         sliced = <tuple>PyTuple_GetSlice(self.peaks, start, end)
 *         return sliced             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_sliced);
  __pyx_r = __pyx_v_sliced;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":324
 *         return 0
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         """Find all peaks within `error_tolerance` ppm of `mz` m/z.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.all_peaks_for", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sliced);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_23all_peaks_for(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_22all_peaks_for[] = "PeakSet.all_peaks_for(self, double mz, double error_tolerance=2e-5) -> tuple\nFind all peaks within `error_tolerance` ppm of `mz` m/z.\n\n        Parameters\n        ----------\n        mz : float\n            The query m/z\n        error_tolerance : float, optional\n            The parts-per-million error tolerance (the default is 2e-5)\n\n        Returns\n        -------\n        tuple\n        ";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_23all_peaks_for(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_mz;
  double __pyx_v_error_tolerance;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("all_peaks_for (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_error_tolerance,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "all_peaks_for") < 0)) __PYX_ERR(0, 324, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_mz = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 324, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 324, __pyx_L3_error)
    } else {
      __pyx_v_error_tolerance = ((double)2e-5);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("all_peaks_for", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 324, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.all_peaks_for", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_22all_peaks_for(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), __pyx_v_mz, __pyx_v_error_tolerance);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_22all_peaks_for(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("all_peaks_for", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;
  __pyx_t_1 = __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet->all_peaks_for(__pyx_v_self, __pyx_v_mz, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.all_peaks_for", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":349
 *         return sliced
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.peaks
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_25__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_24__getstate__[] = "PeakSet.__getstate__(self)";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_25__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_24__getstate__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_24__getstate__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":350
 * 
 *     def __getstate__(self):
 *         return self.peaks             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->peaks);
  __pyx_r = __pyx_v_self->peaks;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":349
 *         return sliced
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.peaks
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":352
 *         return self.peaks
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.peaks = state
 *         self.indexed = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_27__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_26__setstate__[] = "PeakSet.__setstate__(self, state)";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_27__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_26__setstate__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), ((PyObject *)__pyx_v_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_26__setstate__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":353
 * 
 *     def __setstate__(self, state):
 *         self.peaks = state             # <<<<<<<<<<<<<<
 *         self.indexed = False
 *         self.reindex()
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v_state))||((__pyx_v_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v_state)->tp_name), 0))) __PYX_ERR(0, 353, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_state;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->peaks);
  __Pyx_DECREF(__pyx_v_self->peaks);
  __pyx_v_self->peaks = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":354
 *     def __setstate__(self, state):
 *         self.peaks = state
 *         self.indexed = False             # <<<<<<<<<<<<<<
 *         self.reindex()
 * 
 */
  __pyx_v_self->indexed = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":355
 *         self.peaks = state
 *         self.indexed = False
 *         self.reindex()             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->reindex(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":352
 *         return self.peaks
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.peaks = state
 *         self.indexed = False
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":357
 *         self.reindex()
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return PeakSet, (tuple(),), self.__getstate__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_29__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_28__reduce__[] = "PeakSet.__reduce__(self)";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_29__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_28__reduce__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_28__reduce__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":358
 * 
 *     def __reduce__(self):
 *         return PeakSet, (tuple(),), self.__getstate__()             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _eq(self, PeakSet other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)(&PyTuple_Type))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getstate); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet));
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet));
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":357
 *         self.reindex()
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return PeakSet, (tuple(),), self.__getstate__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":360
 *         return PeakSet, (tuple(),), self.__getstate__()
 * 
 *     cpdef bint _eq(self, PeakSet other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_31_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__eq(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_other, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_p1 = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_p2 = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  size_t __pyx_t_6;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_eq", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_eq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 360, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_31_eq)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_other)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 360, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 360, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_peak_picker/_c/peak_set.pyx":365
 *             FittedPeak p1, p2
 * 
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         if n != other.get_size():
 *             return False
 */
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->get_size(__pyx_v_self);

  /* "ms_peak_picker/_c/peak_set.pyx":366
 * 
 *         n = self.get_size()
 *         if n != other.get_size():             # <<<<<<<<<<<<<<
 *             return False
 *         for i in range(n):
 */
  __pyx_t_5 = ((__pyx_v_n != ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_other->__pyx_vtab)->get_size(__pyx_v_other)) != 0);
  if (__pyx_t_5) {

    /* "ms_peak_picker/_c/peak_set.pyx":367
 *         n = self.get_size()
 *         if n != other.get_size():
 *             return False             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             p1 = self.getitem(i)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":366
 * 
 *         n = self.get_size()
 *         if n != other.get_size():             # <<<<<<<<<<<<<<
 *             return False
 *         for i in range(n):
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":368
 *         if n != other.get_size():
 *             return False
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             p1 = self.getitem(i)
 *             p2 = other.getitem(i)
 */
  __pyx_t_6 = __pyx_v_n;
  __pyx_t_7 = __pyx_t_6;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_i = __pyx_t_8;

    /* "ms_peak_picker/_c/peak_set.pyx":369
 *             return False
 *         for i in range(n):
 *             p1 = self.getitem(i)             # <<<<<<<<<<<<<<
 *             p2 = other.getitem(i)
 *             if not p1._eq(p2):
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->getitem(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_p1, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_peak_picker/_c/peak_set.pyx":370
 *         for i in range(n):
 *             p1 = self.getitem(i)
 *             p2 = other.getitem(i)             # <<<<<<<<<<<<<<
 *             if not p1._eq(p2):
 *                 return False
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_other->__pyx_vtab)->getitem(__pyx_v_other, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_p2, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_peak_picker/_c/peak_set.pyx":371
 *             p1 = self.getitem(i)
 *             p2 = other.getitem(i)
 *             if not p1._eq(p2):             # <<<<<<<<<<<<<<
 *                 return False
 *         return True
 */
    __pyx_t_5 = ((!(((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_p1->__pyx_base.__pyx_vtab)->_eq(__pyx_v_p1, __pyx_v_p2, 0) != 0)) != 0);
    if (__pyx_t_5) {

      /* "ms_peak_picker/_c/peak_set.pyx":372
 *             p2 = other.getitem(i)
 *             if not p1._eq(p2):
 *                 return False             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":371
 *             p1 = self.getitem(i)
 *             p2 = other.getitem(i)
 *             if not p1._eq(p2):             # <<<<<<<<<<<<<<
 *                 return False
 *         return True
 */
    }
  }

  /* "ms_peak_picker/_c/peak_set.pyx":373
 *             if not p1._eq(p2):
 *                 return False
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, object other, int code):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":360
 *         return PeakSet, (tuple(),), self.__getstate__()
 * 
 *     cpdef bint _eq(self, PeakSet other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("ms_peak_picker._c.peak_set.PeakSet._eq", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_p1);
  __Pyx_XDECREF((PyObject *)__pyx_v_p2);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_31_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_30_eq[] = "PeakSet._eq(self, PeakSet other) -> bool";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_31_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_eq (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "other", 0))) __PYX_ERR(0, 360, __pyx_L1_error)
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_30_eq(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_30_eq(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_eq", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__eq(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet._eq", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":375
 *         return True
 * 
 *     def __richcmp__(self, object other, int code):             # <<<<<<<<<<<<<<
 *         cdef:
 *             tuple other_tuple
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_33__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_33__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_32__richcmp__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((int)__pyx_v_code));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_32__richcmp__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {
  PyObject *__pyx_v_other_tuple = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":379
 *             tuple other_tuple
 * 
 *         if other is None:             # <<<<<<<<<<<<<<
 *             if code == 2:
 *                 return False
 */
  __pyx_t_1 = (__pyx_v_other == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "ms_peak_picker/_c/peak_set.pyx":380
 * 
 *         if other is None:
 *             if code == 2:             # <<<<<<<<<<<<<<
 *                 return False
 *             elif code == 3:
 */
    switch (__pyx_v_code) {
      case 2:

      /* "ms_peak_picker/_c/peak_set.pyx":381
 *         if other is None:
 *             if code == 2:
 *                 return False             # <<<<<<<<<<<<<<
 *             elif code == 3:
 *                 return True
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":380
 * 
 *         if other is None:
 *             if code == 2:             # <<<<<<<<<<<<<<
 *                 return False
 *             elif code == 3:
 */
      break;
      case 3:

      /* "ms_peak_picker/_c/peak_set.pyx":383
 *                 return False
 *             elif code == 3:
 *                 return True             # <<<<<<<<<<<<<<
 *         if isinstance(other, PeakSet):
 *             if code == 2:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":382
 *             if code == 2:
 *                 return False
 *             elif code == 3:             # <<<<<<<<<<<<<<
 *                 return True
 *         if isinstance(other, PeakSet):
 */
      break;
      default: break;
    }

    /* "ms_peak_picker/_c/peak_set.pyx":379
 *             tuple other_tuple
 * 
 *         if other is None:             # <<<<<<<<<<<<<<
 *             if code == 2:
 *                 return False
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":384
 *             elif code == 3:
 *                 return True
 *         if isinstance(other, PeakSet):             # <<<<<<<<<<<<<<
 *             if code == 2:
 *                 return self._eq(<PeakSet>other)
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":385
 *                 return True
 *         if isinstance(other, PeakSet):
 *             if code == 2:             # <<<<<<<<<<<<<<
 *                 return self._eq(<PeakSet>other)
 *             elif code == 3:
 */
    switch (__pyx_v_code) {
      case 2:

      /* "ms_peak_picker/_c/peak_set.pyx":386
 *         if isinstance(other, PeakSet):
 *             if code == 2:
 *                 return self._eq(<PeakSet>other)             # <<<<<<<<<<<<<<
 *             elif code == 3:
 *                 return not self._eq(<PeakSet>other)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->_eq(__pyx_v_self, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_other), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 386, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":385
 *                 return True
 *         if isinstance(other, PeakSet):
 *             if code == 2:             # <<<<<<<<<<<<<<
 *                 return self._eq(<PeakSet>other)
 *             elif code == 3:
 */
      break;
      case 3:

      /* "ms_peak_picker/_c/peak_set.pyx":388
 *                 return self._eq(<PeakSet>other)
 *             elif code == 3:
 *                 return not self._eq(<PeakSet>other)             # <<<<<<<<<<<<<<
 *             else:
 *                 return NotImplemented
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyBool_FromLong((!(((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->__pyx_vtab)->_eq(__pyx_v_self, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_other), 0) != 0))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 388, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":387
 *             if code == 2:
 *                 return self._eq(<PeakSet>other)
 *             elif code == 3:             # <<<<<<<<<<<<<<
 *                 return not self._eq(<PeakSet>other)
 *             else:
 */
      break;
      default:

      /* "ms_peak_picker/_c/peak_set.pyx":390
 *                 return not self._eq(<PeakSet>other)
 *             else:
 *                 return NotImplemented             # <<<<<<<<<<<<<<
 *         else:
 *             try:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_builtin_NotImplemented);
      __pyx_r = __pyx_builtin_NotImplemented;
      goto __pyx_L0;
      break;
    }

    /* "ms_peak_picker/_c/peak_set.pyx":384
 *             elif code == 3:
 *                 return True
 *         if isinstance(other, PeakSet):             # <<<<<<<<<<<<<<
 *             if code == 2:
 *                 return self._eq(<PeakSet>other)
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":392
 *                 return NotImplemented
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 other_tuple = tuple(other)
 *                 if code == 2:
 */
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      /*try:*/ {

        /* "ms_peak_picker/_c/peak_set.pyx":393
 *         else:
 *             try:
 *                 other_tuple = tuple(other)             # <<<<<<<<<<<<<<
 *                 if code == 2:
 *                     return self.peaks == other_tuple
 */
        __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_v_other); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 393, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_v_other_tuple = ((PyObject*)__pyx_t_3);
        __pyx_t_3 = 0;

        /* "ms_peak_picker/_c/peak_set.pyx":394
 *             try:
 *                 other_tuple = tuple(other)
 *                 if code == 2:             # <<<<<<<<<<<<<<
 *                     return self.peaks == other_tuple
 *                 elif code == 3:
 */
        switch (__pyx_v_code) {
          case 2:

          /* "ms_peak_picker/_c/peak_set.pyx":395
 *                 other_tuple = tuple(other)
 *                 if code == 2:
 *                     return self.peaks == other_tuple             # <<<<<<<<<<<<<<
 *                 elif code == 3:
 *                     return self.peaks != other_tuple
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_3 = PyObject_RichCompare(__pyx_v_self->peaks, __pyx_v_other_tuple, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 395, __pyx_L5_error)
          __pyx_r = __pyx_t_3;
          __pyx_t_3 = 0;
          goto __pyx_L9_try_return;

          /* "ms_peak_picker/_c/peak_set.pyx":394
 *             try:
 *                 other_tuple = tuple(other)
 *                 if code == 2:             # <<<<<<<<<<<<<<
 *                     return self.peaks == other_tuple
 *                 elif code == 3:
 */
          break;
          case 3:

          /* "ms_peak_picker/_c/peak_set.pyx":397
 *                     return self.peaks == other_tuple
 *                 elif code == 3:
 *                     return self.peaks != other_tuple             # <<<<<<<<<<<<<<
 *                 else:
 *                     return NotImplemented
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_3 = PyObject_RichCompare(__pyx_v_self->peaks, __pyx_v_other_tuple, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 397, __pyx_L5_error)
          __pyx_r = __pyx_t_3;
          __pyx_t_3 = 0;
          goto __pyx_L9_try_return;

          /* "ms_peak_picker/_c/peak_set.pyx":396
 *                 if code == 2:
 *                     return self.peaks == other_tuple
 *                 elif code == 3:             # <<<<<<<<<<<<<<
 *                     return self.peaks != other_tuple
 *                 else:
 */
          break;
          default:

          /* "ms_peak_picker/_c/peak_set.pyx":399
 *                     return self.peaks != other_tuple
 *                 else:
 *                     return NotImplemented             # <<<<<<<<<<<<<<
 *             except Exception:
 *                 if code == 2:
 */
          __Pyx_XDECREF(__pyx_r);
          __Pyx_INCREF(__pyx_builtin_NotImplemented);
          __pyx_r = __pyx_builtin_NotImplemented;
          goto __pyx_L9_try_return;
          break;
        }

        /* "ms_peak_picker/_c/peak_set.pyx":392
 *                 return NotImplemented
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 other_tuple = tuple(other)
 *                 if code == 2:
 */
      }
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "ms_peak_picker/_c/peak_set.pyx":400
 *                 else:
 *                     return NotImplemented
 *             except Exception:             # <<<<<<<<<<<<<<
 *                 if code == 2:
 *                     return False
 */
      __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_7) {
        __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(0, 400, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_9);

        /* "ms_peak_picker/_c/peak_set.pyx":401
 *                     return NotImplemented
 *             except Exception:
 *                 if code == 2:             # <<<<<<<<<<<<<<
 *                     return False
 *                 elif code == 3:
 */
        switch (__pyx_v_code) {
          case 2:

          /* "ms_peak_picker/_c/peak_set.pyx":402
 *             except Exception:
 *                 if code == 2:
 *                     return False             # <<<<<<<<<<<<<<
 *                 elif code == 3:
 *                     return True
 */
          __Pyx_XDECREF(__pyx_r);
          __Pyx_INCREF(Py_False);
          __pyx_r = Py_False;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          goto __pyx_L8_except_return;

          /* "ms_peak_picker/_c/peak_set.pyx":401
 *                     return NotImplemented
 *             except Exception:
 *                 if code == 2:             # <<<<<<<<<<<<<<
 *                     return False
 *                 elif code == 3:
 */
          break;
          case 3:

          /* "ms_peak_picker/_c/peak_set.pyx":404
 *                     return False
 *                 elif code == 3:
 *                     return True             # <<<<<<<<<<<<<<
 *                 else:
 *                     return NotImplemented
 */
          __Pyx_XDECREF(__pyx_r);
          __Pyx_INCREF(Py_True);
          __pyx_r = Py_True;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          goto __pyx_L8_except_return;

          /* "ms_peak_picker/_c/peak_set.pyx":403
 *                 if code == 2:
 *                     return False
 *                 elif code == 3:             # <<<<<<<<<<<<<<
 *                     return True
 *                 else:
 */
          break;
          default:

          /* "ms_peak_picker/_c/peak_set.pyx":406
 *                     return True
 *                 else:
 *                     return NotImplemented             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __Pyx_XDECREF(__pyx_r);
          __Pyx_INCREF(__pyx_builtin_NotImplemented);
          __pyx_r = __pyx_builtin_NotImplemented;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          goto __pyx_L8_except_return;
          break;
        }
      }
      goto __pyx_L7_except_error;
      __pyx_L7_except_error:;

      /* "ms_peak_picker/_c/peak_set.pyx":392
 *                 return NotImplemented
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 other_tuple = tuple(other)
 *                 if code == 2:
 */
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L1_error;
      __pyx_L9_try_return:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L0;
      __pyx_L8_except_return:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L0;
    }
  }

  /* "ms_peak_picker/_c/peak_set.pyx":375
 *         return True
 * 
 *     def __richcmp__(self, object other, int code):             # <<<<<<<<<<<<<<
 *         cdef:
 *             tuple other_tuple
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_other_tuple);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pxd":31
 * cdef class PeakSet(object):
 *     cdef:
 *         public tuple peaks             # <<<<<<<<<<<<<<
 *         public bint indexed
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks___get__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->peaks);
  __pyx_r = __pyx_v_self->peaks;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks_2__set__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyTuple_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 31, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->peaks);
  __Pyx_DECREF(__pyx_v_self->peaks);
  __pyx_v_self->peaks = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.peaks.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks_4__del__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks_4__del__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->peaks);
  __Pyx_DECREF(__pyx_v_self->peaks);
  __pyx_v_self->peaks = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pxd":32
 *     cdef:
 *         public tuple peaks
 *         public bint indexed             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_7indexed_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_7indexed_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_7indexed___get__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_7indexed___get__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->indexed); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.indexed.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_7indexed_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_7indexed_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_7indexed_2__set__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_7PeakSet_7indexed_2__set__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 32, __pyx_L1_error)
  __pyx_v_self->indexed = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSet.indexed.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":413
 * 
 * 
 * cpdef FittedPeak binary_search_ppm_error(tuple array, double value, double tolerance):             # <<<<<<<<<<<<<<
 *     return _binary_search_ppm_error(array, value, 0, len(array), tolerance)
 * 
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_1binary_search_ppm_error(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set_binary_search_ppm_error(PyObject *__pyx_v_array, double __pyx_v_value, double __pyx_v_tolerance, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("binary_search_ppm_error", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":414
 * 
 * cpdef FittedPeak binary_search_ppm_error(tuple array, double value, double tolerance):
 *     return _binary_search_ppm_error(array, value, 0, len(array), tolerance)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  if (unlikely(__pyx_v_array == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 414, __pyx_L1_error)
  }
  __pyx_t_1 = PyTuple_GET_SIZE(__pyx_v_array); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 414, __pyx_L1_error)
  __pyx_t_2 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set__binary_search_ppm_error(__pyx_v_array, __pyx_v_value, 0, __pyx_t_1, __pyx_v_tolerance)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":413
 * 
 * 
 * cpdef FittedPeak binary_search_ppm_error(tuple array, double value, double tolerance):             # <<<<<<<<<<<<<<
 *     return _binary_search_ppm_error(array, value, 0, len(array), tolerance)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.binary_search_ppm_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_1binary_search_ppm_error(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_binary_search_ppm_error[] = "binary_search_ppm_error(tuple array, double value, double tolerance) -> FittedPeak";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_1binary_search_ppm_error(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_array = 0;
  double __pyx_v_value;
  double __pyx_v_tolerance;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("binary_search_ppm_error (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_array,&__pyx_n_s_value,&__pyx_n_s_tolerance,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_array)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("binary_search_ppm_error", 1, 3, 3, 1); __PYX_ERR(0, 413, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tolerance)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("binary_search_ppm_error", 1, 3, 3, 2); __PYX_ERR(0, 413, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "binary_search_ppm_error") < 0)) __PYX_ERR(0, 413, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_array = ((PyObject*)values[0]);
    __pyx_v_value = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 413, __pyx_L3_error)
    __pyx_v_tolerance = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 413, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("binary_search_ppm_error", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 413, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.binary_search_ppm_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_array), (&PyTuple_Type), 1, "array", 1))) __PYX_ERR(0, 413, __pyx_L1_error)
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_binary_search_ppm_error(__pyx_self, __pyx_v_array, __pyx_v_value, __pyx_v_tolerance);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_binary_search_ppm_error(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_array, double __pyx_v_value, double __pyx_v_tolerance) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("binary_search_ppm_error", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set_binary_search_ppm_error(__pyx_v_array, __pyx_v_value, __pyx_v_tolerance, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.binary_search_ppm_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":417
 * 
 * 
 * cdef FittedPeak _binary_search_ppm_error(tuple array, double value, size_t lo, size_t hi, double tolerance):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t mid, lower_edge
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set__binary_search_ppm_error(PyObject *__pyx_v_array, double __pyx_v_value, size_t __pyx_v_lo, size_t __pyx_v_hi, double __pyx_v_tolerance) {
  size_t __pyx_v_mid;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_target = 0;
  double __pyx_v_target_value;
  double __pyx_v_error;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3;
  double __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_search_ppm_error", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":422
 *         FittedPeak target
 *         double target_value, error
 *     if (hi - lo) < 5:             # <<<<<<<<<<<<<<
 *         return _sweep_solution_ppm_error(array, value, lo, hi, tolerance)
 *     else:
 */
  __pyx_t_1 = (((__pyx_v_hi - __pyx_v_lo) < 5) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":423
 *         double target_value, error
 *     if (hi - lo) < 5:
 *         return _sweep_solution_ppm_error(array, value, lo, hi, tolerance)             # <<<<<<<<<<<<<<
 *     else:
 *         mid = (hi + lo) // 2
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __pyx_t_2 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set__sweep_solution_ppm_error(__pyx_v_array, __pyx_v_value, __pyx_v_lo, __pyx_v_hi, __pyx_v_tolerance)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":422
 *         FittedPeak target
 *         double target_value, error
 *     if (hi - lo) < 5:             # <<<<<<<<<<<<<<
 *         return _sweep_solution_ppm_error(array, value, lo, hi, tolerance)
 *     else:
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":425
 *         return _sweep_solution_ppm_error(array, value, lo, hi, tolerance)
 *     else:
 *         mid = (hi + lo) // 2             # <<<<<<<<<<<<<<
 *         target = <FittedPeak>PyTuple_GetItem(array, mid)
 *         target_value = target.mz
 */
  /*else*/ {
    __pyx_v_mid = ((__pyx_v_hi + __pyx_v_lo) / 2);

    /* "ms_peak_picker/_c/peak_set.pyx":426
 *     else:
 *         mid = (hi + lo) // 2
 *         target = <FittedPeak>PyTuple_GetItem(array, mid)             # <<<<<<<<<<<<<<
 *         target_value = target.mz
 *         error = ppm_error(value, target_value)
 */
    __pyx_t_3 = PyTuple_GetItem(__pyx_v_array, __pyx_v_mid); if (unlikely(__pyx_t_3 == ((PyObject *)NULL))) __PYX_ERR(0, 426, __pyx_L1_error)
    __pyx_t_2 = ((PyObject *)__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_target = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "ms_peak_picker/_c/peak_set.pyx":427
 *         mid = (hi + lo) // 2
 *         target = <FittedPeak>PyTuple_GetItem(array, mid)
 *         target_value = target.mz             # <<<<<<<<<<<<<<
 *         error = ppm_error(value, target_value)
 *         if abs(error) <= tolerance:
 */
    __pyx_t_4 = __pyx_v_target->__pyx_base.mz;
    __pyx_v_target_value = __pyx_t_4;

    /* "ms_peak_picker/_c/peak_set.pyx":428
 *         target = <FittedPeak>PyTuple_GetItem(array, mid)
 *         target_value = target.mz
 *         error = ppm_error(value, target_value)             # <<<<<<<<<<<<<<
 *         if abs(error) <= tolerance:
 *             return _sweep_solution_ppm_error(array, value, max(mid - (mid if mid < 5 else 5), lo), min(mid + 5, hi), tolerance)
 */
    __pyx_v_error = __pyx_f_14ms_peak_picker_2_c_8peak_set_ppm_error(__pyx_v_value, __pyx_v_target_value);

    /* "ms_peak_picker/_c/peak_set.pyx":429
 *         target_value = target.mz
 *         error = ppm_error(value, target_value)
 *         if abs(error) <= tolerance:             # <<<<<<<<<<<<<<
 *             return _sweep_solution_ppm_error(array, value, max(mid - (mid if mid < 5 else 5), lo), min(mid + 5, hi), tolerance)
 *         elif target_value > value:
 */
    __pyx_t_1 = ((fabs(__pyx_v_error) <= __pyx_v_tolerance) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":430
 *         error = ppm_error(value, target_value)
 *         if abs(error) <= tolerance:
 *             return _sweep_solution_ppm_error(array, value, max(mid - (mid if mid < 5 else 5), lo), min(mid + 5, hi), tolerance)             # <<<<<<<<<<<<<<
 *         elif target_value > value:
 *             return _binary_search_ppm_error(array, value, lo, mid, tolerance)
 */
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_t_5 = __pyx_v_lo;
      if (((__pyx_v_mid < 5) != 0)) {
        __pyx_t_6 = __pyx_v_mid;
      } else {
        __pyx_t_6 = 5;
      }
      __pyx_t_7 = (__pyx_v_mid - __pyx_t_6);
      if (((__pyx_t_5 > __pyx_t_7) != 0)) {
        __pyx_t_6 = __pyx_t_5;
      } else {
        __pyx_t_6 = __pyx_t_7;
      }
      __pyx_t_5 = __pyx_v_hi;
      __pyx_t_7 = (__pyx_v_mid + 5);
      if (((__pyx_t_5 < __pyx_t_7) != 0)) {
        __pyx_t_8 = __pyx_t_5;
      } else {
        __pyx_t_8 = __pyx_t_7;
      }
      __pyx_t_2 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set__sweep_solution_ppm_error(__pyx_v_array, __pyx_v_value, __pyx_t_6, __pyx_t_8, __pyx_v_tolerance)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 430, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2);
      __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":429
 *         target_value = target.mz
 *         error = ppm_error(value, target_value)
 *         if abs(error) <= tolerance:             # <<<<<<<<<<<<<<
 *             return _sweep_solution_ppm_error(array, value, max(mid - (mid if mid < 5 else 5), lo), min(mid + 5, hi), tolerance)
 *         elif target_value > value:
 */
    }

    /* "ms_peak_picker/_c/peak_set.pyx":431
 *         if abs(error) <= tolerance:
 *             return _sweep_solution_ppm_error(array, value, max(mid - (mid if mid < 5 else 5), lo), min(mid + 5, hi), tolerance)
 *         elif target_value > value:             # <<<<<<<<<<<<<<
 *             return _binary_search_ppm_error(array, value, lo, mid, tolerance)
 *         elif target_value < value:
 */
    __pyx_t_1 = ((__pyx_v_target_value > __pyx_v_value) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":432
 *             return _sweep_solution_ppm_error(array, value, max(mid - (mid if mid < 5 else 5), lo), min(mid + 5, hi), tolerance)
 *         elif target_value > value:
 *             return _binary_search_ppm_error(array, value, lo, mid, tolerance)             # <<<<<<<<<<<<<<
 *         elif target_value < value:
 *             return _binary_search_ppm_error(array, value, mid, hi, tolerance)
 */
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_t_2 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set__binary_search_ppm_error(__pyx_v_array, __pyx_v_value, __pyx_v_lo, __pyx_v_mid, __pyx_v_tolerance)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2);
      __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":431
 *         if abs(error) <= tolerance:
 *             return _sweep_solution_ppm_error(array, value, max(mid - (mid if mid < 5 else 5), lo), min(mid + 5, hi), tolerance)
 *         elif target_value > value:             # <<<<<<<<<<<<<<
 *             return _binary_search_ppm_error(array, value, lo, mid, tolerance)
 *         elif target_value < value:
 */
    }

    /* "ms_peak_picker/_c/peak_set.pyx":433
 *         elif target_value > value:
 *             return _binary_search_ppm_error(array, value, lo, mid, tolerance)
 *         elif target_value < value:             # <<<<<<<<<<<<<<
 *             return _binary_search_ppm_error(array, value, mid, hi, tolerance)
 *     return _null_peak
 */
    __pyx_t_1 = ((__pyx_v_target_value < __pyx_v_value) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":434
 *             return _binary_search_ppm_error(array, value, lo, mid, tolerance)
 *         elif target_value < value:
 *             return _binary_search_ppm_error(array, value, mid, hi, tolerance)             # <<<<<<<<<<<<<<
 *     return _null_peak
 * 
 */
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_t_2 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set__binary_search_ppm_error(__pyx_v_array, __pyx_v_value, __pyx_v_mid, __pyx_v_hi, __pyx_v_tolerance)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 434, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2);
      __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":433
 *         elif target_value > value:
 *             return _binary_search_ppm_error(array, value, lo, mid, tolerance)
 *         elif target_value < value:             # <<<<<<<<<<<<<<
 *             return _binary_search_ppm_error(array, value, mid, hi, tolerance)
 *     return _null_peak
 */
    }
  }

  /* "ms_peak_picker/_c/peak_set.pyx":435
 *         elif target_value < value:
 *             return _binary_search_ppm_error(array, value, mid, hi, tolerance)
 *     return _null_peak             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak));
  __pyx_r = __pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":417
 * 
 * 
 * cdef FittedPeak _binary_search_ppm_error(tuple array, double value, size_t lo, size_t hi, double tolerance):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t mid, lower_edge
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set._binary_search_ppm_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_target);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":438
 * 
 * 
 * cdef FittedPeak _sweep_solution_ppm_error(tuple array, double value, size_t lo, size_t hi, double tolerance):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double best_error, error, abs_error
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set__sweep_solution_ppm_error(PyObject *__pyx_v_array, double __pyx_v_value, size_t __pyx_v_lo, size_t __pyx_v_hi, double __pyx_v_tolerance) {
  double __pyx_v_best_error;
  double __pyx_v_error;
  double __pyx_v_abs_error;
  size_t __pyx_v_i;
  size_t __pyx_v_best_index;
  long __pyx_v_best_intensity;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_target = NULL;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_sweep_solution_ppm_error", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":443
 *         size_t i, limit
 * 
 *     best_index = -1             # <<<<<<<<<<<<<<
 *     best_intensity = 0
 *     best_error = 1000000000000000
 */
  __pyx_v_best_index = -1L;

  /* "ms_peak_picker/_c/peak_set.pyx":444
 * 
 *     best_index = -1
 *     best_intensity = 0             # <<<<<<<<<<<<<<
 *     best_error = 1000000000000000
 *     for i in range(hi - lo):
 */
  __pyx_v_best_intensity = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":445
 *     best_index = -1
 *     best_intensity = 0
 *     best_error = 1000000000000000             # <<<<<<<<<<<<<<
 *     for i in range(hi - lo):
 *         target = <FittedPeak>PyTuple_GetItem(array, lo + i)
 */
  __pyx_v_best_error = 1000000000000000.0;

  /* "ms_peak_picker/_c/peak_set.pyx":446
 *     best_intensity = 0
 *     best_error = 1000000000000000
 *     for i in range(hi - lo):             # <<<<<<<<<<<<<<
 *         target = <FittedPeak>PyTuple_GetItem(array, lo + i)
 *         error = ppm_error(value, target.mz)
 */
  __pyx_t_1 = (__pyx_v_hi - __pyx_v_lo);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "ms_peak_picker/_c/peak_set.pyx":447
 *     best_error = 1000000000000000
 *     for i in range(hi - lo):
 *         target = <FittedPeak>PyTuple_GetItem(array, lo + i)             # <<<<<<<<<<<<<<
 *         error = ppm_error(value, target.mz)
 *         abs_error = abs(error)
 */
    __pyx_t_4 = PyTuple_GetItem(__pyx_v_array, (__pyx_v_lo + __pyx_v_i)); if (unlikely(__pyx_t_4 == ((PyObject *)NULL))) __PYX_ERR(0, 447, __pyx_L1_error)
    __pyx_t_5 = ((PyObject *)__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_target, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "ms_peak_picker/_c/peak_set.pyx":448
 *     for i in range(hi - lo):
 *         target = <FittedPeak>PyTuple_GetItem(array, lo + i)
 *         error = ppm_error(value, target.mz)             # <<<<<<<<<<<<<<
 *         abs_error = abs(error)
 *         if abs_error < tolerance and (abs_error < (best_error * 1.1)) and (target.intensity > best_intensity):
 */
    __pyx_v_error = __pyx_f_14ms_peak_picker_2_c_8peak_set_ppm_error(__pyx_v_value, __pyx_v_target->__pyx_base.mz);

    /* "ms_peak_picker/_c/peak_set.pyx":449
 *         target = <FittedPeak>PyTuple_GetItem(array, lo + i)
 *         error = ppm_error(value, target.mz)
 *         abs_error = abs(error)             # <<<<<<<<<<<<<<
 *         if abs_error < tolerance and (abs_error < (best_error * 1.1)) and (target.intensity > best_intensity):
 *             best_index = lo + i
 */
    __pyx_v_abs_error = fabs(__pyx_v_error);

    /* "ms_peak_picker/_c/peak_set.pyx":450
 *         error = ppm_error(value, target.mz)
 *         abs_error = abs(error)
 *         if abs_error < tolerance and (abs_error < (best_error * 1.1)) and (target.intensity > best_intensity):             # <<<<<<<<<<<<<<
 *             best_index = lo + i
 *             best_error = abs_error
 */
    __pyx_t_7 = ((__pyx_v_abs_error < __pyx_v_tolerance) != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_7 = ((__pyx_v_abs_error < (__pyx_v_best_error * 1.1)) != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_7 = ((__pyx_v_target->__pyx_base.intensity > __pyx_v_best_intensity) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_6) {

      /* "ms_peak_picker/_c/peak_set.pyx":451
 *         abs_error = abs(error)
 *         if abs_error < tolerance and (abs_error < (best_error * 1.1)) and (target.intensity > best_intensity):
 *             best_index = lo + i             # <<<<<<<<<<<<<<
 *             best_error = abs_error
 *     if best_index == -1:
 */
      __pyx_v_best_index = (__pyx_v_lo + __pyx_v_i);

      /* "ms_peak_picker/_c/peak_set.pyx":452
 *         if abs_error < tolerance and (abs_error < (best_error * 1.1)) and (target.intensity > best_intensity):
 *             best_index = lo + i
 *             best_error = abs_error             # <<<<<<<<<<<<<<
 *     if best_index == -1:
 *         return _null_peak
 */
      __pyx_v_best_error = __pyx_v_abs_error;

      /* "ms_peak_picker/_c/peak_set.pyx":450
 *         error = ppm_error(value, target.mz)
 *         abs_error = abs(error)
 *         if abs_error < tolerance and (abs_error < (best_error * 1.1)) and (target.intensity > best_intensity):             # <<<<<<<<<<<<<<
 *             best_index = lo + i
 *             best_error = abs_error
 */
    }
  }

  /* "ms_peak_picker/_c/peak_set.pyx":453
 *             best_index = lo + i
 *             best_error = abs_error
 *     if best_index == -1:             # <<<<<<<<<<<<<<
 *         return _null_peak
 *     else:
 */
  __pyx_t_6 = ((__pyx_v_best_index == -1L) != 0);
  if (__pyx_t_6) {

    /* "ms_peak_picker/_c/peak_set.pyx":454
 *             best_error = abs_error
 *     if best_index == -1:
 *         return _null_peak             # <<<<<<<<<<<<<<
 *     else:
 *         return <FittedPeak>PyTuple_GetItem(array, best_index)
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak));
    __pyx_r = __pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":453
 *             best_index = lo + i
 *             best_error = abs_error
 *     if best_index == -1:             # <<<<<<<<<<<<<<
 *         return _null_peak
 *     else:
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":456
 *         return _null_peak
 *     else:
 *         return <FittedPeak>PyTuple_GetItem(array, best_index)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __pyx_t_4 = PyTuple_GetItem(__pyx_v_array, __pyx_v_best_index); if (unlikely(__pyx_t_4 == ((PyObject *)NULL))) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_INCREF(((PyObject *)((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_4)));
    __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_4);
    goto __pyx_L0;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":438
 * 
 * 
 * cdef FittedPeak _sweep_solution_ppm_error(tuple array, double value, size_t lo, size_t hi, double tolerance):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double best_error, error, abs_error
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set._sweep_solution_ppm_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_target);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":459
 * 
 * 
 * cdef FittedPeak _sweep_nearest_match(tuple array, double value, size_t lo, size_t hi, double* errout):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set__sweep_nearest_match(PyObject *__pyx_v_array, double __pyx_v_value, size_t __pyx_v_lo, size_t __pyx_v_hi, double *__pyx_v_errout) {
  size_t __pyx_v_i;
  size_t __pyx_v_best_index;
  double __pyx_v_best_error;
  double __pyx_v_err;
  double __pyx_v_v;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_sweep_nearest_match", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":466
 *         double v
 * 
 *     best_error = float('inf')             # <<<<<<<<<<<<<<
 *     best_index = -1
 *     for i in range(hi - lo):
 */
  __pyx_t_1 = __Pyx_PyObject_AsDouble(__pyx_n_s_inf); if (unlikely(__pyx_t_1 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 466, __pyx_L1_error)
  __pyx_v_best_error = __pyx_t_1;

  /* "ms_peak_picker/_c/peak_set.pyx":467
 * 
 *     best_error = float('inf')
 *     best_index = -1             # <<<<<<<<<<<<<<
 *     for i in range(hi - lo):
 *         i += lo
 */
  __pyx_v_best_index = -1L;

  /* "ms_peak_picker/_c/peak_set.pyx":468
 *     best_error = float('inf')
 *     best_index = -1
 *     for i in range(hi - lo):             # <<<<<<<<<<<<<<
 *         i += lo
 *         v = array[i].mz
 */
  __pyx_t_2 = (__pyx_v_hi - __pyx_v_lo);
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "ms_peak_picker/_c/peak_set.pyx":469
 *     best_index = -1
 *     for i in range(hi - lo):
 *         i += lo             # <<<<<<<<<<<<<<
 *         v = array[i].mz
 *         err = abs(v - value)
 */
    __pyx_v_i = (__pyx_v_i + __pyx_v_lo);

    /* "ms_peak_picker/_c/peak_set.pyx":470
 *     for i in range(hi - lo):
 *         i += lo
 *         v = array[i].mz             # <<<<<<<<<<<<<<
 *         err = abs(v - value)
 *         if err < best_error:
 */
    if (unlikely(__pyx_v_array == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 470, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v_array, __pyx_v_i, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_mz); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_t_6); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_v = __pyx_t_1;

    /* "ms_peak_picker/_c/peak_set.pyx":471
 *         i += lo
 *         v = array[i].mz
 *         err = abs(v - value)             # <<<<<<<<<<<<<<
 *         if err < best_error:
 *             best_error = err
 */
    __pyx_v_err = fabs((__pyx_v_v - __pyx_v_value));

    /* "ms_peak_picker/_c/peak_set.pyx":472
 *         v = array[i].mz
 *         err = abs(v - value)
 *         if err < best_error:             # <<<<<<<<<<<<<<
 *             best_error = err
 *             best_index = i
 */
    __pyx_t_7 = ((__pyx_v_err < __pyx_v_best_error) != 0);
    if (__pyx_t_7) {

      /* "ms_peak_picker/_c/peak_set.pyx":473
 *         err = abs(v - value)
 *         if err < best_error:
 *             best_error = err             # <<<<<<<<<<<<<<
 *             best_index = i
 *     errout[0] = best_error
 */
      __pyx_v_best_error = __pyx_v_err;

      /* "ms_peak_picker/_c/peak_set.pyx":474
 *         if err < best_error:
 *             best_error = err
 *             best_index = i             # <<<<<<<<<<<<<<
 *     errout[0] = best_error
 *     return array[best_index]
 */
      __pyx_v_best_index = __pyx_v_i;

      /* "ms_peak_picker/_c/peak_set.pyx":472
 *         v = array[i].mz
 *         err = abs(v - value)
 *         if err < best_error:             # <<<<<<<<<<<<<<
 *             best_error = err
 *             best_index = i
 */
    }
  }

  /* "ms_peak_picker/_c/peak_set.pyx":475
 *             best_error = err
 *             best_index = i
 *     errout[0] = best_error             # <<<<<<<<<<<<<<
 *     return array[best_index]
 * 
 */
  (__pyx_v_errout[0]) = __pyx_v_best_error;

  /* "ms_peak_picker/_c/peak_set.pyx":476
 *             best_index = i
 *     errout[0] = best_error
 *     return array[best_index]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  if (unlikely(__pyx_v_array == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 476, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v_array, __pyx_v_best_index, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak))))) __PYX_ERR(0, 476, __pyx_L1_error)
  __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_6);
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":459
 * 
 * 
 * cdef FittedPeak _sweep_nearest_match(tuple array, double value, size_t lo, size_t hi, double* errout):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set._sweep_nearest_match", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":479
 * 
 * 
 * cdef FittedPeak _binary_search_nearest_match(tuple array, double value, size_t lo, size_t hi, double* errout):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t mid
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set__binary_search_nearest_match(PyObject *__pyx_v_array, double __pyx_v_value, size_t __pyx_v_lo, size_t __pyx_v_hi, double *__pyx_v_errout) {
  size_t __pyx_v_mid;
  double __pyx_v_v;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_search_nearest_match", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":485
 *         double err
 * 
 *     if (hi - lo) < 5:             # <<<<<<<<<<<<<<
 *         return _sweep_nearest_match(array, value, lo, hi, errout)
 *     else:
 */
  __pyx_t_1 = (((__pyx_v_hi - __pyx_v_lo) < 5) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":486
 * 
 *     if (hi - lo) < 5:
 *         return _sweep_nearest_match(array, value, lo, hi, errout)             # <<<<<<<<<<<<<<
 *     else:
 *         mid = (hi + lo) // 2
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __pyx_t_2 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set__sweep_nearest_match(__pyx_v_array, __pyx_v_value, __pyx_v_lo, __pyx_v_hi, __pyx_v_errout)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":485
 *         double err
 * 
 *     if (hi - lo) < 5:             # <<<<<<<<<<<<<<
 *         return _sweep_nearest_match(array, value, lo, hi, errout)
 *     else:
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":488
 *         return _sweep_nearest_match(array, value, lo, hi, errout)
 *     else:
 *         mid = (hi + lo) // 2             # <<<<<<<<<<<<<<
 *         v = array[mid].mz
 *         if abs(v - value) < 1.:
 */
  /*else*/ {
    __pyx_v_mid = ((__pyx_v_hi + __pyx_v_lo) / 2);

    /* "ms_peak_picker/_c/peak_set.pyx":489
 *     else:
 *         mid = (hi + lo) // 2
 *         v = array[mid].mz             # <<<<<<<<<<<<<<
 *         if abs(v - value) < 1.:
 *             return _sweep_nearest_match(array, value, lo, hi, errout)
 */
    if (unlikely(__pyx_v_array == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 489, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_GetItemInt_Tuple(__pyx_v_array, __pyx_v_mid, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_v = __pyx_t_4;

    /* "ms_peak_picker/_c/peak_set.pyx":490
 *         mid = (hi + lo) // 2
 *         v = array[mid].mz
 *         if abs(v - value) < 1.:             # <<<<<<<<<<<<<<
 *             return _sweep_nearest_match(array, value, lo, hi, errout)
 *         elif v > value:
 */
    __pyx_t_1 = ((fabs((__pyx_v_v - __pyx_v_value)) < 1.) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":491
 *         v = array[mid].mz
 *         if abs(v - value) < 1.:
 *             return _sweep_nearest_match(array, value, lo, hi, errout)             # <<<<<<<<<<<<<<
 *         elif v > value:
 *             return _binary_search_nearest_match(array, value, lo, mid, errout)
 */
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_t_3 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set__sweep_nearest_match(__pyx_v_array, __pyx_v_value, __pyx_v_lo, __pyx_v_hi, __pyx_v_errout)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 491, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_3);
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":490
 *         mid = (hi + lo) // 2
 *         v = array[mid].mz
 *         if abs(v - value) < 1.:             # <<<<<<<<<<<<<<
 *             return _sweep_nearest_match(array, value, lo, hi, errout)
 *         elif v > value:
 */
    }

    /* "ms_peak_picker/_c/peak_set.pyx":492
 *         if abs(v - value) < 1.:
 *             return _sweep_nearest_match(array, value, lo, hi, errout)
 *         elif v > value:             # <<<<<<<<<<<<<<
 *             return _binary_search_nearest_match(array, value, lo, mid, errout)
 *         else:
 */
    __pyx_t_1 = ((__pyx_v_v > __pyx_v_value) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":493
 *             return _sweep_nearest_match(array, value, lo, hi, errout)
 *         elif v > value:
 *             return _binary_search_nearest_match(array, value, lo, mid, errout)             # <<<<<<<<<<<<<<
 *         else:
 *             return _binary_search_nearest_match(array, value, mid, hi, errout)
 */
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_t_3 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set__binary_search_nearest_match(__pyx_v_array, __pyx_v_value, __pyx_v_lo, __pyx_v_mid, __pyx_v_errout)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_3);
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":492
 *         if abs(v - value) < 1.:
 *             return _sweep_nearest_match(array, value, lo, hi, errout)
 *         elif v > value:             # <<<<<<<<<<<<<<
 *             return _binary_search_nearest_match(array, value, lo, mid, errout)
 *         else:
 */
    }

    /* "ms_peak_picker/_c/peak_set.pyx":495
 *             return _binary_search_nearest_match(array, value, lo, mid, errout)
 *         else:
 *             return _binary_search_nearest_match(array, value, mid, hi, errout)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_t_3 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set__binary_search_nearest_match(__pyx_v_array, __pyx_v_value, __pyx_v_mid, __pyx_v_hi, __pyx_v_errout)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 495, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_3);
      __pyx_t_3 = 0;
      goto __pyx_L0;
    }
  }

  /* "ms_peak_picker/_c/peak_set.pyx":479
 * 
 * 
 * cdef FittedPeak _binary_search_nearest_match(tuple array, double value, size_t lo, size_t hi, double* errout):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t mid
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set._binary_search_nearest_match", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":499
 * 
 * @cython.cdivision(True)
 * cdef size_t double_binary_search_ppm(double* array, double value, double tolerance, size_t n):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t lo, hi, mid
 */

static size_t __pyx_f_14ms_peak_picker_2_c_8peak_set_double_binary_search_ppm(double *__pyx_v_array, double __pyx_v_value, double __pyx_v_tolerance, size_t __pyx_v_n) {
  size_t __pyx_v_lo;
  size_t __pyx_v_hi;
  size_t __pyx_v_mid;
  size_t __pyx_v_i;
  size_t __pyx_v_best_ix;
  double __pyx_v_x;
  double __pyx_v_err;
  double __pyx_v_abs_err;
  double __pyx_v_best_error;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("double_binary_search_ppm", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":504
 *         size_t i, best_ix
 *         double x, err, best_err, abs_err
 *     lo = 0             # <<<<<<<<<<<<<<
 *     hi = n
 *     while hi != lo:
 */
  __pyx_v_lo = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":505
 *         double x, err, best_err, abs_err
 *     lo = 0
 *     hi = n             # <<<<<<<<<<<<<<
 *     while hi != lo:
 *         mid = (hi + lo) // 2
 */
  __pyx_v_hi = __pyx_v_n;

  /* "ms_peak_picker/_c/peak_set.pyx":506
 *     lo = 0
 *     hi = n
 *     while hi != lo:             # <<<<<<<<<<<<<<
 *         mid = (hi + lo) // 2
 *         x = array[mid]
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_hi != __pyx_v_lo) != 0);
    if (!__pyx_t_1) break;

    /* "ms_peak_picker/_c/peak_set.pyx":507
 *     hi = n
 *     while hi != lo:
 *         mid = (hi + lo) // 2             # <<<<<<<<<<<<<<
 *         x = array[mid]
 *         err = (x - value) / value
 */
    __pyx_v_mid = ((__pyx_v_hi + __pyx_v_lo) / 2);

    /* "ms_peak_picker/_c/peak_set.pyx":508
 *     while hi != lo:
 *         mid = (hi + lo) // 2
 *         x = array[mid]             # <<<<<<<<<<<<<<
 *         err = (x - value) / value
 *         abs_err = abs(err)
 */
    __pyx_v_x = (__pyx_v_array[__pyx_v_mid]);

    /* "ms_peak_picker/_c/peak_set.pyx":509
 *         mid = (hi + lo) // 2
 *         x = array[mid]
 *         err = (x - value) / value             # <<<<<<<<<<<<<<
 *         abs_err = abs(err)
 *         if abs_err < tolerance:
 */
    __pyx_v_err = ((__pyx_v_x - __pyx_v_value) / __pyx_v_value);

    /* "ms_peak_picker/_c/peak_set.pyx":510
 *         x = array[mid]
 *         err = (x - value) / value
 *         abs_err = abs(err)             # <<<<<<<<<<<<<<
 *         if abs_err < tolerance:
 *             i = mid
 */
    __pyx_v_abs_err = fabs(__pyx_v_err);

    /* "ms_peak_picker/_c/peak_set.pyx":511
 *         err = (x - value) / value
 *         abs_err = abs(err)
 *         if abs_err < tolerance:             # <<<<<<<<<<<<<<
 *             i = mid
 *             best_error = abs_err
 */
    __pyx_t_1 = ((__pyx_v_abs_err < __pyx_v_tolerance) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":512
 *         abs_err = abs(err)
 *         if abs_err < tolerance:
 *             i = mid             # <<<<<<<<<<<<<<
 *             best_error = abs_err
 *             best_ix = mid
 */
      __pyx_v_i = __pyx_v_mid;

      /* "ms_peak_picker/_c/peak_set.pyx":513
 *         if abs_err < tolerance:
 *             i = mid
 *             best_error = abs_err             # <<<<<<<<<<<<<<
 *             best_ix = mid
 *             while i > 0:
 */
      __pyx_v_best_error = __pyx_v_abs_err;

      /* "ms_peak_picker/_c/peak_set.pyx":514
 *             i = mid
 *             best_error = abs_err
 *             best_ix = mid             # <<<<<<<<<<<<<<
 *             while i > 0:
 *                 i -= 1
 */
      __pyx_v_best_ix = __pyx_v_mid;

      /* "ms_peak_picker/_c/peak_set.pyx":515
 *             best_error = abs_err
 *             best_ix = mid
 *             while i > 0:             # <<<<<<<<<<<<<<
 *                 i -= 1
 *                 x = array[i]
 */
      while (1) {
        __pyx_t_1 = ((__pyx_v_i > 0) != 0);
        if (!__pyx_t_1) break;

        /* "ms_peak_picker/_c/peak_set.pyx":516
 *             best_ix = mid
 *             while i > 0:
 *                 i -= 1             # <<<<<<<<<<<<<<
 *                 x = array[i]
 *                 err = (x - value) / value
 */
        __pyx_v_i = (__pyx_v_i - 1);

        /* "ms_peak_picker/_c/peak_set.pyx":517
 *             while i > 0:
 *                 i -= 1
 *                 x = array[i]             # <<<<<<<<<<<<<<
 *                 err = (x - value) / value
 *                 abs_err = abs(err)
 */
        __pyx_v_x = (__pyx_v_array[__pyx_v_i]);

        /* "ms_peak_picker/_c/peak_set.pyx":518
 *                 i -= 1
 *                 x = array[i]
 *                 err = (x - value) / value             # <<<<<<<<<<<<<<
 *                 abs_err = abs(err)
 *                 if abs_err > tolerance:
 */
        __pyx_v_err = ((__pyx_v_x - __pyx_v_value) / __pyx_v_value);

        /* "ms_peak_picker/_c/peak_set.pyx":519
 *                 x = array[i]
 *                 err = (x - value) / value
 *                 abs_err = abs(err)             # <<<<<<<<<<<<<<
 *                 if abs_err > tolerance:
 *                     break
 */
        __pyx_v_abs_err = fabs(__pyx_v_err);

        /* "ms_peak_picker/_c/peak_set.pyx":520
 *                 err = (x - value) / value
 *                 abs_err = abs(err)
 *                 if abs_err > tolerance:             # <<<<<<<<<<<<<<
 *                     break
 *                 elif abs_err < best_error:
 */
        __pyx_t_1 = ((__pyx_v_abs_err > __pyx_v_tolerance) != 0);
        if (__pyx_t_1) {

          /* "ms_peak_picker/_c/peak_set.pyx":521
 *                 abs_err = abs(err)
 *                 if abs_err > tolerance:
 *                     break             # <<<<<<<<<<<<<<
 *                 elif abs_err < best_error:
 *                     best_error = abs_err
 */
          goto __pyx_L7_break;

          /* "ms_peak_picker/_c/peak_set.pyx":520
 *                 err = (x - value) / value
 *                 abs_err = abs(err)
 *                 if abs_err > tolerance:             # <<<<<<<<<<<<<<
 *                     break
 *                 elif abs_err < best_error:
 */
        }

        /* "ms_peak_picker/_c/peak_set.pyx":522
 *                 if abs_err > tolerance:
 *                     break
 *                 elif abs_err < best_error:             # <<<<<<<<<<<<<<
 *                     best_error = abs_err
 *                     best_ix = i
 */
        __pyx_t_1 = ((__pyx_v_abs_err < __pyx_v_best_error) != 0);
        if (__pyx_t_1) {

          /* "ms_peak_picker/_c/peak_set.pyx":523
 *                     break
 *                 elif abs_err < best_error:
 *                     best_error = abs_err             # <<<<<<<<<<<<<<
 *                     best_ix = i
 *             i = mid
 */
          __pyx_v_best_error = __pyx_v_abs_err;

          /* "ms_peak_picker/_c/peak_set.pyx":524
 *                 elif abs_err < best_error:
 *                     best_error = abs_err
 *                     best_ix = i             # <<<<<<<<<<<<<<
 *             i = mid
 *             while i < n - 1:
 */
          __pyx_v_best_ix = __pyx_v_i;

          /* "ms_peak_picker/_c/peak_set.pyx":522
 *                 if abs_err > tolerance:
 *                     break
 *                 elif abs_err < best_error:             # <<<<<<<<<<<<<<
 *                     best_error = abs_err
 *                     best_ix = i
 */
        }
      }
      __pyx_L7_break:;

      /* "ms_peak_picker/_c/peak_set.pyx":525
 *                     best_error = abs_err
 *                     best_ix = i
 *             i = mid             # <<<<<<<<<<<<<<
 *             while i < n - 1:
 *                 i += 1
 */
      __pyx_v_i = __pyx_v_mid;

      /* "ms_peak_picker/_c/peak_set.pyx":526
 *                     best_ix = i
 *             i = mid
 *             while i < n - 1:             # <<<<<<<<<<<<<<
 *                 i += 1
 *                 x = array[i]
 */
      while (1) {
        __pyx_t_1 = ((__pyx_v_i < (__pyx_v_n - 1)) != 0);
        if (!__pyx_t_1) break;

        /* "ms_peak_picker/_c/peak_set.pyx":527
 *             i = mid
 *             while i < n - 1:
 *                 i += 1             # <<<<<<<<<<<<<<
 *                 x = array[i]
 *                 err = (x - value) / value
 */
        __pyx_v_i = (__pyx_v_i + 1);

        /* "ms_peak_picker/_c/peak_set.pyx":528
 *             while i < n - 1:
 *                 i += 1
 *                 x = array[i]             # <<<<<<<<<<<<<<
 *                 err = (x - value) / value
 *                 abs_err = abs(err)
 */
        __pyx_v_x = (__pyx_v_array[__pyx_v_i]);

        /* "ms_peak_picker/_c/peak_set.pyx":529
 *                 i += 1
 *                 x = array[i]
 *                 err = (x - value) / value             # <<<<<<<<<<<<<<
 *                 abs_err = abs(err)
 *                 if abs_err > tolerance:
 */
        __pyx_v_err = ((__pyx_v_x - __pyx_v_value) / __pyx_v_value);

        /* "ms_peak_picker/_c/peak_set.pyx":530
 *                 x = array[i]
 *                 err = (x - value) / value
 *                 abs_err = abs(err)             # <<<<<<<<<<<<<<
 *                 if abs_err > tolerance:
 *                     break
 */
        __pyx_v_abs_err = fabs(__pyx_v_err);

        /* "ms_peak_picker/_c/peak_set.pyx":531
 *                 err = (x - value) / value
 *                 abs_err = abs(err)
 *                 if abs_err > tolerance:             # <<<<<<<<<<<<<<
 *                     break
 *                 elif abs_err < best_error:
 */
        __pyx_t_1 = ((__pyx_v_abs_err > __pyx_v_tolerance) != 0);
        if (__pyx_t_1) {

          /* "ms_peak_picker/_c/peak_set.pyx":532
 *                 abs_err = abs(err)
 *                 if abs_err > tolerance:
 *                     break             # <<<<<<<<<<<<<<
 *                 elif abs_err < best_error:
 *                     best_error = abs_err
 */
          goto __pyx_L10_break;

          /* "ms_peak_picker/_c/peak_set.pyx":531
 *                 err = (x - value) / value
 *                 abs_err = abs(err)
 *                 if abs_err > tolerance:             # <<<<<<<<<<<<<<
 *                     break
 *                 elif abs_err < best_error:
 */
        }

        /* "ms_peak_picker/_c/peak_set.pyx":533
 *                 if abs_err > tolerance:
 *                     break
 *                 elif abs_err < best_error:             # <<<<<<<<<<<<<<
 *                     best_error = abs_err
 *                     best_ix = i
 */
        __pyx_t_1 = ((__pyx_v_abs_err < __pyx_v_best_error) != 0);
        if (__pyx_t_1) {

          /* "ms_peak_picker/_c/peak_set.pyx":534
 *                     break
 *                 elif abs_err < best_error:
 *                     best_error = abs_err             # <<<<<<<<<<<<<<
 *                     best_ix = i
 *             return best_ix
 */
          __pyx_v_best_error = __pyx_v_abs_err;

          /* "ms_peak_picker/_c/peak_set.pyx":535
 *                 elif abs_err < best_error:
 *                     best_error = abs_err
 *                     best_ix = i             # <<<<<<<<<<<<<<
 *             return best_ix
 *         elif (hi - 1) == lo:
 */
          __pyx_v_best_ix = __pyx_v_i;

          /* "ms_peak_picker/_c/peak_set.pyx":533
 *                 if abs_err > tolerance:
 *                     break
 *                 elif abs_err < best_error:             # <<<<<<<<<<<<<<
 *                     best_error = abs_err
 *                     best_ix = i
 */
        }
      }
      __pyx_L10_break:;

      /* "ms_peak_picker/_c/peak_set.pyx":536
 *                     best_error = abs_err
 *                     best_ix = i
 *             return best_ix             # <<<<<<<<<<<<<<
 *         elif (hi - 1) == lo:
 *             return mid
 */
      __pyx_r = __pyx_v_best_ix;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":511
 *         err = (x - value) / value
 *         abs_err = abs(err)
 *         if abs_err < tolerance:             # <<<<<<<<<<<<<<
 *             i = mid
 *             best_error = abs_err
 */
    }

    /* "ms_peak_picker/_c/peak_set.pyx":537
 *                     best_ix = i
 *             return best_ix
 *         elif (hi - 1) == lo:             # <<<<<<<<<<<<<<
 *             return mid
 *         elif err > 0:
 */
    __pyx_t_1 = (((__pyx_v_hi - 1) == __pyx_v_lo) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":538
 *             return best_ix
 *         elif (hi - 1) == lo:
 *             return mid             # <<<<<<<<<<<<<<
 *         elif err > 0:
 *             hi = mid
 */
      __pyx_r = __pyx_v_mid;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":537
 *                     best_ix = i
 *             return best_ix
 *         elif (hi - 1) == lo:             # <<<<<<<<<<<<<<
 *             return mid
 *         elif err > 0:
 */
    }

    /* "ms_peak_picker/_c/peak_set.pyx":539
 *         elif (hi - 1) == lo:
 *             return mid
 *         elif err > 0:             # <<<<<<<<<<<<<<
 *             hi = mid
 *         else:
 */
    __pyx_t_1 = ((__pyx_v_err > 0.0) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":540
 *             return mid
 *         elif err > 0:
 *             hi = mid             # <<<<<<<<<<<<<<
 *         else:
 *             lo = mid
 */
      __pyx_v_hi = __pyx_v_mid;

      /* "ms_peak_picker/_c/peak_set.pyx":539
 *         elif (hi - 1) == lo:
 *             return mid
 *         elif err > 0:             # <<<<<<<<<<<<<<
 *             hi = mid
 *         else:
 */
      goto __pyx_L5;
    }

    /* "ms_peak_picker/_c/peak_set.pyx":542
 *             hi = mid
 *         else:
 *             lo = mid             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    /*else*/ {
      __pyx_v_lo = __pyx_v_mid;
    }
    __pyx_L5:;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":543
 *         else:
 *             lo = mid
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":499
 * 
 * @cython.cdivision(True)
 * cdef size_t double_binary_search_ppm(double* array, double value, double tolerance, size_t n):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t lo, hi, mid
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":547
 * 
 * @cython.cdivision(True)
 * cdef size_t double_binary_search_nearest_match(double* array, double value, size_t n):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t lo, hi, mid
 */

static size_t __pyx_f_14ms_peak_picker_2_c_8peak_set_double_binary_search_nearest_match(double *__pyx_v_array, double __pyx_v_value, size_t __pyx_v_n) {
  size_t __pyx_v_lo;
  size_t __pyx_v_hi;
  size_t __pyx_v_mid;
  size_t __pyx_v_i;
  size_t __pyx_v_best_ix;
  double __pyx_v_x;
  double __pyx_v_err;
  double __pyx_v_best_err;
  double __pyx_v_abs_err;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("double_binary_search_nearest_match", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":552
 *         size_t i, best_ix
 *         double x, err, best_err, abs_err
 *     lo = 0             # <<<<<<<<<<<<<<
 *     hi = n
 *     while hi != lo:
 */
  __pyx_v_lo = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":553
 *         double x, err, best_err, abs_err
 *     lo = 0
 *     hi = n             # <<<<<<<<<<<<<<
 *     while hi != lo:
 *         mid = (hi + lo) // 2
 */
  __pyx_v_hi = __pyx_v_n;

  /* "ms_peak_picker/_c/peak_set.pyx":554
 *     lo = 0
 *     hi = n
 *     while hi != lo:             # <<<<<<<<<<<<<<
 *         mid = (hi + lo) // 2
 *         x = array[mid]
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_hi != __pyx_v_lo) != 0);
    if (!__pyx_t_1) break;

    /* "ms_peak_picker/_c/peak_set.pyx":555
 *     hi = n
 *     while hi != lo:
 *         mid = (hi + lo) // 2             # <<<<<<<<<<<<<<
 *         x = array[mid]
 *         err = x - value
 */
    __pyx_v_mid = ((__pyx_v_hi + __pyx_v_lo) / 2);

    /* "ms_peak_picker/_c/peak_set.pyx":556
 *     while hi != lo:
 *         mid = (hi + lo) // 2
 *         x = array[mid]             # <<<<<<<<<<<<<<
 *         err = x - value
 *         if err == 0 or ((hi - 1) == lo):
 */
    __pyx_v_x = (__pyx_v_array[__pyx_v_mid]);

    /* "ms_peak_picker/_c/peak_set.pyx":557
 *         mid = (hi + lo) // 2
 *         x = array[mid]
 *         err = x - value             # <<<<<<<<<<<<<<
 *         if err == 0 or ((hi - 1) == lo):
 *             i = mid
 */
    __pyx_v_err = (__pyx_v_x - __pyx_v_value);

    /* "ms_peak_picker/_c/peak_set.pyx":558
 *         x = array[mid]
 *         err = x - value
 *         if err == 0 or ((hi - 1) == lo):             # <<<<<<<<<<<<<<
 *             i = mid
 *             best_err = abs(err)
 */
    __pyx_t_2 = ((__pyx_v_err == 0.0) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_2 = (((__pyx_v_hi - 1) == __pyx_v_lo) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":559
 *         err = x - value
 *         if err == 0 or ((hi - 1) == lo):
 *             i = mid             # <<<<<<<<<<<<<<
 *             best_err = abs(err)
 *             best_ix = mid
 */
      __pyx_v_i = __pyx_v_mid;

      /* "ms_peak_picker/_c/peak_set.pyx":560
 *         if err == 0 or ((hi - 1) == lo):
 *             i = mid
 *             best_err = abs(err)             # <<<<<<<<<<<<<<
 *             best_ix = mid
 *             while i > 0:
 */
      __pyx_v_best_err = fabs(__pyx_v_err);

      /* "ms_peak_picker/_c/peak_set.pyx":561
 *             i = mid
 *             best_err = abs(err)
 *             best_ix = mid             # <<<<<<<<<<<<<<
 *             while i > 0:
 *                 i -= 1
 */
      __pyx_v_best_ix = __pyx_v_mid;

      /* "ms_peak_picker/_c/peak_set.pyx":562
 *             best_err = abs(err)
 *             best_ix = mid
 *             while i > 0:             # <<<<<<<<<<<<<<
 *                 i -= 1
 *                 x = array[i]
 */
      while (1) {
        __pyx_t_1 = ((__pyx_v_i > 0) != 0);
        if (!__pyx_t_1) break;

        /* "ms_peak_picker/_c/peak_set.pyx":563
 *             best_ix = mid
 *             while i > 0:
 *                 i -= 1             # <<<<<<<<<<<<<<
 *                 x = array[i]
 *                 err = (x - value)
 */
        __pyx_v_i = (__pyx_v_i - 1);

        /* "ms_peak_picker/_c/peak_set.pyx":564
 *             while i > 0:
 *                 i -= 1
 *                 x = array[i]             # <<<<<<<<<<<<<<
 *                 err = (x - value)
 *                 abs_err = abs(err)
 */
        __pyx_v_x = (__pyx_v_array[__pyx_v_i]);

        /* "ms_peak_picker/_c/peak_set.pyx":565
 *                 i -= 1
 *                 x = array[i]
 *                 err = (x - value)             # <<<<<<<<<<<<<<
 *                 abs_err = abs(err)
 *                 if abs_err > best_err:
 */
        __pyx_v_err = (__pyx_v_x - __pyx_v_value);

        /* "ms_peak_picker/_c/peak_set.pyx":566
 *                 x = array[i]
 *                 err = (x - value)
 *                 abs_err = abs(err)             # <<<<<<<<<<<<<<
 *                 if abs_err > best_err:
 *                     break
 */
        __pyx_v_abs_err = fabs(__pyx_v_err);

        /* "ms_peak_picker/_c/peak_set.pyx":567
 *                 err = (x - value)
 *                 abs_err = abs(err)
 *                 if abs_err > best_err:             # <<<<<<<<<<<<<<
 *                     break
 *                 elif abs_err < best_err:
 */
        __pyx_t_1 = ((__pyx_v_abs_err > __pyx_v_best_err) != 0);
        if (__pyx_t_1) {

          /* "ms_peak_picker/_c/peak_set.pyx":568
 *                 abs_err = abs(err)
 *                 if abs_err > best_err:
 *                     break             # <<<<<<<<<<<<<<
 *                 elif abs_err < best_err:
 *                     best_err = abs_err
 */
          goto __pyx_L9_break;

          /* "ms_peak_picker/_c/peak_set.pyx":567
 *                 err = (x - value)
 *                 abs_err = abs(err)
 *                 if abs_err > best_err:             # <<<<<<<<<<<<<<
 *                     break
 *                 elif abs_err < best_err:
 */
        }

        /* "ms_peak_picker/_c/peak_set.pyx":569
 *                 if abs_err > best_err:
 *                     break
 *                 elif abs_err < best_err:             # <<<<<<<<<<<<<<
 *                     best_err = abs_err
 *                     best_ix = i
 */
        __pyx_t_1 = ((__pyx_v_abs_err < __pyx_v_best_err) != 0);
        if (__pyx_t_1) {

          /* "ms_peak_picker/_c/peak_set.pyx":570
 *                     break
 *                 elif abs_err < best_err:
 *                     best_err = abs_err             # <<<<<<<<<<<<<<
 *                     best_ix = i
 *             i = mid
 */
          __pyx_v_best_err = __pyx_v_abs_err;

          /* "ms_peak_picker/_c/peak_set.pyx":571
 *                 elif abs_err < best_err:
 *                     best_err = abs_err
 *                     best_ix = i             # <<<<<<<<<<<<<<
 *             i = mid
 *             while i < n - 1:
 */
          __pyx_v_best_ix = __pyx_v_i;

          /* "ms_peak_picker/_c/peak_set.pyx":569
 *                 if abs_err > best_err:
 *                     break
 *                 elif abs_err < best_err:             # <<<<<<<<<<<<<<
 *                     best_err = abs_err
 *                     best_ix = i
 */
        }
      }
      __pyx_L9_break:;

      /* "ms_peak_picker/_c/peak_set.pyx":572
 *                     best_err = abs_err
 *                     best_ix = i
 *             i = mid             # <<<<<<<<<<<<<<
 *             while i < n - 1:
 *                 i += 1
 */
      __pyx_v_i = __pyx_v_mid;

      /* "ms_peak_picker/_c/peak_set.pyx":573
 *                     best_ix = i
 *             i = mid
 *             while i < n - 1:             # <<<<<<<<<<<<<<
 *                 i += 1
 *                 x = array[i]
 */
      while (1) {
        __pyx_t_1 = ((__pyx_v_i < (__pyx_v_n - 1)) != 0);
        if (!__pyx_t_1) break;

        /* "ms_peak_picker/_c/peak_set.pyx":574
 *             i = mid
 *             while i < n - 1:
 *                 i += 1             # <<<<<<<<<<<<<<
 *                 x = array[i]
 *                 err = (x - value)
 */
        __pyx_v_i = (__pyx_v_i + 1);

        /* "ms_peak_picker/_c/peak_set.pyx":575
 *             while i < n - 1:
 *                 i += 1
 *                 x = array[i]             # <<<<<<<<<<<<<<
 *                 err = (x - value)
 *                 abs_err = abs(err)
 */
        __pyx_v_x = (__pyx_v_array[__pyx_v_i]);

        /* "ms_peak_picker/_c/peak_set.pyx":576
 *                 i += 1
 *                 x = array[i]
 *                 err = (x - value)             # <<<<<<<<<<<<<<
 *                 abs_err = abs(err)
 *                 if abs_err > best_err:
 */
        __pyx_v_err = (__pyx_v_x - __pyx_v_value);

        /* "ms_peak_picker/_c/peak_set.pyx":577
 *                 x = array[i]
 *                 err = (x - value)
 *                 abs_err = abs(err)             # <<<<<<<<<<<<<<
 *                 if abs_err > best_err:
 *                     break
 */
        __pyx_v_abs_err = fabs(__pyx_v_err);

        /* "ms_peak_picker/_c/peak_set.pyx":578
 *                 err = (x - value)
 *                 abs_err = abs(err)
 *                 if abs_err > best_err:             # <<<<<<<<<<<<<<
 *                     break
 *                 elif abs_err < best_err:
 */
        __pyx_t_1 = ((__pyx_v_abs_err > __pyx_v_best_err) != 0);
        if (__pyx_t_1) {

          /* "ms_peak_picker/_c/peak_set.pyx":579
 *                 abs_err = abs(err)
 *                 if abs_err > best_err:
 *                     break             # <<<<<<<<<<<<<<
 *                 elif abs_err < best_err:
 *                     best_err = abs_err
 */
          goto __pyx_L12_break;

          /* "ms_peak_picker/_c/peak_set.pyx":578
 *                 err = (x - value)
 *                 abs_err = abs(err)
 *                 if abs_err > best_err:             # <<<<<<<<<<<<<<
 *                     break
 *                 elif abs_err < best_err:
 */
        }

        /* "ms_peak_picker/_c/peak_set.pyx":580
 *                 if abs_err > best_err:
 *                     break
 *                 elif abs_err < best_err:             # <<<<<<<<<<<<<<
 *                     best_err = abs_err
 *                     best_ix = i
 */
        __pyx_t_1 = ((__pyx_v_abs_err < __pyx_v_best_err) != 0);
        if (__pyx_t_1) {

          /* "ms_peak_picker/_c/peak_set.pyx":581
 *                     break
 *                 elif abs_err < best_err:
 *                     best_err = abs_err             # <<<<<<<<<<<<<<
 *                     best_ix = i
 *             return best_ix
 */
          __pyx_v_best_err = __pyx_v_abs_err;

          /* "ms_peak_picker/_c/peak_set.pyx":582
 *                 elif abs_err < best_err:
 *                     best_err = abs_err
 *                     best_ix = i             # <<<<<<<<<<<<<<
 *             return best_ix
 *         elif err > 0:
 */
          __pyx_v_best_ix = __pyx_v_i;

          /* "ms_peak_picker/_c/peak_set.pyx":580
 *                 if abs_err > best_err:
 *                     break
 *                 elif abs_err < best_err:             # <<<<<<<<<<<<<<
 *                     best_err = abs_err
 *                     best_ix = i
 */
        }
      }
      __pyx_L12_break:;

      /* "ms_peak_picker/_c/peak_set.pyx":583
 *                     best_err = abs_err
 *                     best_ix = i
 *             return best_ix             # <<<<<<<<<<<<<<
 *         elif err > 0:
 *             hi = mid
 */
      __pyx_r = __pyx_v_best_ix;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":558
 *         x = array[mid]
 *         err = x - value
 *         if err == 0 or ((hi - 1) == lo):             # <<<<<<<<<<<<<<
 *             i = mid
 *             best_err = abs(err)
 */
    }

    /* "ms_peak_picker/_c/peak_set.pyx":584
 *                     best_ix = i
 *             return best_ix
 *         elif err > 0:             # <<<<<<<<<<<<<<
 *             hi = mid
 *         else:
 */
    __pyx_t_1 = ((__pyx_v_err > 0.0) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":585
 *             return best_ix
 *         elif err > 0:
 *             hi = mid             # <<<<<<<<<<<<<<
 *         else:
 *             lo = mid
 */
      __pyx_v_hi = __pyx_v_mid;

      /* "ms_peak_picker/_c/peak_set.pyx":584
 *                     best_ix = i
 *             return best_ix
 *         elif err > 0:             # <<<<<<<<<<<<<<
 *             hi = mid
 *         else:
 */
      goto __pyx_L5;
    }

    /* "ms_peak_picker/_c/peak_set.pyx":587
 *             hi = mid
 *         else:
 *             lo = mid             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    /*else*/ {
      __pyx_v_lo = __pyx_v_mid;
    }
    __pyx_L5:;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":588
 *         else:
 *             lo = mid
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":547
 * 
 * @cython.cdivision(True)
 * cdef size_t double_binary_search_nearest_match(double* array, double value, size_t n):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t lo, hi, mid
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":592
 * 
 * @cython.cdivision(True)
 * cdef size_t double_binary_search_ppm_with_hint(double* array, double value, double tolerance, size_t n, size_t hint):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t lo, hi, mid
 */

static size_t __pyx_f_14ms_peak_picker_2_c_8peak_set_double_binary_search_ppm_with_hint(double *__pyx_v_array, double __pyx_v_value, double __pyx_v_tolerance, size_t __pyx_v_n, size_t __pyx_v_hint) {
  size_t __pyx_v_lo;
  size_t __pyx_v_hi;
  size_t __pyx_v_mid;
  size_t __pyx_v_i;
  size_t __pyx_v_best_ix;
  double __pyx_v_x;
  double __pyx_v_err;
  double __pyx_v_best_err;
  double __pyx_v_abs_err;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("double_binary_search_ppm_with_hint", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":597
 *         size_t i, best_ix
 *         double x, err, best_err, abs_err
 *     lo = hint             # <<<<<<<<<<<<<<
 *     hi = n
 *     while hi != lo:
 */
  __pyx_v_lo = __pyx_v_hint;

  /* "ms_peak_picker/_c/peak_set.pyx":598
 *         double x, err, best_err, abs_err
 *     lo = hint
 *     hi = n             # <<<<<<<<<<<<<<
 *     while hi != lo:
 *         mid = (hi + lo) // 2
 */
  __pyx_v_hi = __pyx_v_n;

  /* "ms_peak_picker/_c/peak_set.pyx":599
 *     lo = hint
 *     hi = n
 *     while hi != lo:             # <<<<<<<<<<<<<<
 *         mid = (hi + lo) // 2
 *         x = array[mid]
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_hi != __pyx_v_lo) != 0);
    if (!__pyx_t_1) break;

    /* "ms_peak_picker/_c/peak_set.pyx":600
 *     hi = n
 *     while hi != lo:
 *         mid = (hi + lo) // 2             # <<<<<<<<<<<<<<
 *         x = array[mid]
 *         err = (x - value) / value
 */
    __pyx_v_mid = ((__pyx_v_hi + __pyx_v_lo) / 2);

    /* "ms_peak_picker/_c/peak_set.pyx":601
 *     while hi != lo:
 *         mid = (hi + lo) // 2
 *         x = array[mid]             # <<<<<<<<<<<<<<
 *         err = (x - value) / value
 *         abs_err = abs(err)
 */
    __pyx_v_x = (__pyx_v_array[__pyx_v_mid]);

    /* "ms_peak_picker/_c/peak_set.pyx":602
 *         mid = (hi + lo) // 2
 *         x = array[mid]
 *         err = (x - value) / value             # <<<<<<<<<<<<<<
 *         abs_err = abs(err)
 *         if abs_err < tolerance:
 */
    __pyx_v_err = ((__pyx_v_x - __pyx_v_value) / __pyx_v_value);

    /* "ms_peak_picker/_c/peak_set.pyx":603
 *         x = array[mid]
 *         err = (x - value) / value
 *         abs_err = abs(err)             # <<<<<<<<<<<<<<
 *         if abs_err < tolerance:
 *             i = mid
 */
    __pyx_v_abs_err = fabs(__pyx_v_err);

    /* "ms_peak_picker/_c/peak_set.pyx":604
 *         err = (x - value) / value
 *         abs_err = abs(err)
 *         if abs_err < tolerance:             # <<<<<<<<<<<<<<
 *             i = mid
 *             best_err = abs_err
 */
    __pyx_t_1 = ((__pyx_v_abs_err < __pyx_v_tolerance) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":605
 *         abs_err = abs(err)
 *         if abs_err < tolerance:
 *             i = mid             # <<<<<<<<<<<<<<
 *             best_err = abs_err
 *             best_ix = mid
 */
      __pyx_v_i = __pyx_v_mid;

      /* "ms_peak_picker/_c/peak_set.pyx":606
 *         if abs_err < tolerance:
 *             i = mid
 *             best_err = abs_err             # <<<<<<<<<<<<<<
 *             best_ix = mid
 *             while i > 0:
 */
      __pyx_v_best_err = __pyx_v_abs_err;

      /* "ms_peak_picker/_c/peak_set.pyx":607
 *             i = mid
 *             best_err = abs_err
 *             best_ix = mid             # <<<<<<<<<<<<<<
 *             while i > 0:
 *                 i -= 1
 */
      __pyx_v_best_ix = __pyx_v_mid;

      /* "ms_peak_picker/_c/peak_set.pyx":608
 *             best_err = abs_err
 *             best_ix = mid
 *             while i > 0:             # <<<<<<<<<<<<<<
 *                 i -= 1
 *                 x = array[i]
 */
      while (1) {
        __pyx_t_1 = ((__pyx_v_i > 0) != 0);
        if (!__pyx_t_1) break;

        /* "ms_peak_picker/_c/peak_set.pyx":609
 *             best_ix = mid
 *             while i > 0:
 *                 i -= 1             # <<<<<<<<<<<<<<
 *                 x = array[i]
 *                 err = (x - value) / value
 */
        __pyx_v_i = (__pyx_v_i - 1);

        /* "ms_peak_picker/_c/peak_set.pyx":610
 *             while i > 0:
 *                 i -= 1
 *                 x = array[i]             # <<<<<<<<<<<<<<
 *                 err = (x - value) / value
 *                 abs_err = abs(err)
 */
        __pyx_v_x = (__pyx_v_array[__pyx_v_i]);

        /* "ms_peak_picker/_c/peak_set.pyx":611
 *                 i -= 1
 *                 x = array[i]
 *                 err = (x - value) / value             # <<<<<<<<<<<<<<
 *                 abs_err = abs(err)
 *                 if abs_err > tolerance:
 */
        __pyx_v_err = ((__pyx_v_x - __pyx_v_value) / __pyx_v_value);

        /* "ms_peak_picker/_c/peak_set.pyx":612
 *                 x = array[i]
 *                 err = (x - value) / value
 *                 abs_err = abs(err)             # <<<<<<<<<<<<<<
 *                 if abs_err > tolerance:
 *                     break
 */
        __pyx_v_abs_err = fabs(__pyx_v_err);

        /* "ms_peak_picker/_c/peak_set.pyx":613
 *                 err = (x - value) / value
 *                 abs_err = abs(err)
 *                 if abs_err > tolerance:             # <<<<<<<<<<<<<<
 *                     break
 *                 elif abs_err < best_err:
 */
        __pyx_t_1 = ((__pyx_v_abs_err > __pyx_v_tolerance) != 0);
        if (__pyx_t_1) {

          /* "ms_peak_picker/_c/peak_set.pyx":614
 *                 abs_err = abs(err)
 *                 if abs_err > tolerance:
 *                     break             # <<<<<<<<<<<<<<
 *                 elif abs_err < best_err:
 *                     best_err = abs_err
 */
          goto __pyx_L7_break;

          /* "ms_peak_picker/_c/peak_set.pyx":613
 *                 err = (x - value) / value
 *                 abs_err = abs(err)
 *                 if abs_err > tolerance:             # <<<<<<<<<<<<<<
 *                     break
 *                 elif abs_err < best_err:
 */
        }

        /* "ms_peak_picker/_c/peak_set.pyx":615
 *                 if abs_err > tolerance:
 *                     break
 *                 elif abs_err < best_err:             # <<<<<<<<<<<<<<
 *                     best_err = abs_err
 *                     best_ix = i
 */
        __pyx_t_1 = ((__pyx_v_abs_err < __pyx_v_best_err) != 0);
        if (__pyx_t_1) {

          /* "ms_peak_picker/_c/peak_set.pyx":616
 *                     break
 *                 elif abs_err < best_err:
 *                     best_err = abs_err             # <<<<<<<<<<<<<<
 *                     best_ix = i
 *             i = mid
 */
          __pyx_v_best_err = __pyx_v_abs_err;

          /* "ms_peak_picker/_c/peak_set.pyx":617
 *                 elif abs_err < best_err:
 *                     best_err = abs_err
 *                     best_ix = i             # <<<<<<<<<<<<<<
 *             i = mid
 *             while i < n - 1:
 */
          __pyx_v_best_ix = __pyx_v_i;

          /* "ms_peak_picker/_c/peak_set.pyx":615
 *                 if abs_err > tolerance:
 *                     break
 *                 elif abs_err < best_err:             # <<<<<<<<<<<<<<
 *                     best_err = abs_err
 *                     best_ix = i
 */
        }
      }
      __pyx_L7_break:;

      /* "ms_peak_picker/_c/peak_set.pyx":618
 *                     best_err = abs_err
 *                     best_ix = i
 *             i = mid             # <<<<<<<<<<<<<<
 *             while i < n - 1:
 *                 i += 1
 */
      __pyx_v_i = __pyx_v_mid;

      /* "ms_peak_picker/_c/peak_set.pyx":619
 *                     best_ix = i
 *             i = mid
 *             while i < n - 1:             # <<<<<<<<<<<<<<
 *                 i += 1
 *                 x = array[i]
 */
      while (1) {
        __pyx_t_1 = ((__pyx_v_i < (__pyx_v_n - 1)) != 0);
        if (!__pyx_t_1) break;

        /* "ms_peak_picker/_c/peak_set.pyx":620
 *             i = mid
 *             while i < n - 1:
 *                 i += 1             # <<<<<<<<<<<<<<
 *                 x = array[i]
 *                 err = (x - value) / value
 */
        __pyx_v_i = (__pyx_v_i + 1);

        /* "ms_peak_picker/_c/peak_set.pyx":621
 *             while i < n - 1:
 *                 i += 1
 *                 x = array[i]             # <<<<<<<<<<<<<<
 *                 err = (x - value) / value
 *                 abs_err = abs(err)
 */
        __pyx_v_x = (__pyx_v_array[__pyx_v_i]);

        /* "ms_peak_picker/_c/peak_set.pyx":622
 *                 i += 1
 *                 x = array[i]
 *                 err = (x - value) / value             # <<<<<<<<<<<<<<
 *                 abs_err = abs(err)
 *                 if abs_err > tolerance:
 */
        __pyx_v_err = ((__pyx_v_x - __pyx_v_value) / __pyx_v_value);

        /* "ms_peak_picker/_c/peak_set.pyx":623
 *                 x = array[i]
 *                 err = (x - value) / value
 *                 abs_err = abs(err)             # <<<<<<<<<<<<<<
 *                 if abs_err > tolerance:
 *                     break
 */
        __pyx_v_abs_err = fabs(__pyx_v_err);

        /* "ms_peak_picker/_c/peak_set.pyx":624
 *                 err = (x - value) / value
 *                 abs_err = abs(err)
 *                 if abs_err > tolerance:             # <<<<<<<<<<<<<<
 *                     break
 *                 elif abs_err < best_err:
 */
        __pyx_t_1 = ((__pyx_v_abs_err > __pyx_v_tolerance) != 0);
        if (__pyx_t_1) {

          /* "ms_peak_picker/_c/peak_set.pyx":625
 *                 abs_err = abs(err)
 *                 if abs_err > tolerance:
 *                     break             # <<<<<<<<<<<<<<
 *                 elif abs_err < best_err:
 *                     best_err = abs_err
 */
          goto __pyx_L10_break;

          /* "ms_peak_picker/_c/peak_set.pyx":624
 *                 err = (x - value) / value
 *                 abs_err = abs(err)
 *                 if abs_err > tolerance:             # <<<<<<<<<<<<<<
 *                     break
 *                 elif abs_err < best_err:
 */
        }

        /* "ms_peak_picker/_c/peak_set.pyx":626
 *                 if abs_err > tolerance:
 *                     break
 *                 elif abs_err < best_err:             # <<<<<<<<<<<<<<
 *                     best_err = abs_err
 *                     best_ix = i
 */
        __pyx_t_1 = ((__pyx_v_abs_err < __pyx_v_best_err) != 0);
        if (__pyx_t_1) {

          /* "ms_peak_picker/_c/peak_set.pyx":627
 *                     break
 *                 elif abs_err < best_err:
 *                     best_err = abs_err             # <<<<<<<<<<<<<<
 *                     best_ix = i
 *             return best_ix
 */
          __pyx_v_best_err = __pyx_v_abs_err;

          /* "ms_peak_picker/_c/peak_set.pyx":628
 *                 elif abs_err < best_err:
 *                     best_err = abs_err
 *                     best_ix = i             # <<<<<<<<<<<<<<
 *             return best_ix
 *         elif (hi - 1) == lo:
 */
          __pyx_v_best_ix = __pyx_v_i;

          /* "ms_peak_picker/_c/peak_set.pyx":626
 *                 if abs_err > tolerance:
 *                     break
 *                 elif abs_err < best_err:             # <<<<<<<<<<<<<<
 *                     best_err = abs_err
 *                     best_ix = i
 */
        }
      }
      __pyx_L10_break:;

      /* "ms_peak_picker/_c/peak_set.pyx":629
 *                     best_err = abs_err
 *                     best_ix = i
 *             return best_ix             # <<<<<<<<<<<<<<
 *         elif (hi - 1) == lo:
 *             return mid
 */
      __pyx_r = __pyx_v_best_ix;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":604
 *         err = (x - value) / value
 *         abs_err = abs(err)
 *         if abs_err < tolerance:             # <<<<<<<<<<<<<<
 *             i = mid
 *             best_err = abs_err
 */
    }

    /* "ms_peak_picker/_c/peak_set.pyx":630
 *                     best_ix = i
 *             return best_ix
 *         elif (hi - 1) == lo:             # <<<<<<<<<<<<<<
 *             return mid
 *         elif err > 0:
 */
    __pyx_t_1 = (((__pyx_v_hi - 1) == __pyx_v_lo) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":631
 *             return best_ix
 *         elif (hi - 1) == lo:
 *             return mid             # <<<<<<<<<<<<<<
 *         elif err > 0:
 *             hi = mid
 */
      __pyx_r = __pyx_v_mid;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":630
 *                     best_ix = i
 *             return best_ix
 *         elif (hi - 1) == lo:             # <<<<<<<<<<<<<<
 *             return mid
 *         elif err > 0:
 */
    }

    /* "ms_peak_picker/_c/peak_set.pyx":632
 *         elif (hi - 1) == lo:
 *             return mid
 *         elif err > 0:             # <<<<<<<<<<<<<<
 *             hi = mid
 *         else:
 */
    __pyx_t_1 = ((__pyx_v_err > 0.0) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":633
 *             return mid
 *         elif err > 0:
 *             hi = mid             # <<<<<<<<<<<<<<
 *         else:
 *             lo = mid
 */
      __pyx_v_hi = __pyx_v_mid;

      /* "ms_peak_picker/_c/peak_set.pyx":632
 *         elif (hi - 1) == lo:
 *             return mid
 *         elif err > 0:             # <<<<<<<<<<<<<<
 *             hi = mid
 *         else:
 */
      goto __pyx_L5;
    }

    /* "ms_peak_picker/_c/peak_set.pyx":635
 *             hi = mid
 *         else:
 *             lo = mid             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    /*else*/ {
      __pyx_v_lo = __pyx_v_mid;
    }
    __pyx_L5:;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":636
 *         else:
 *             lo = mid
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":592
 * 
 * @cython.cdivision(True)
 * cdef size_t double_binary_search_ppm_with_hint(double* array, double value, double tolerance, size_t n, size_t hint):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t lo, hi, mid
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":639
 * 
 * @cython.cdivision(True)
 * cdef size_t double_binary_search_nearest_match_with_hint(double* array, double value, size_t n, size_t hint):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t lo, hi, mid
 */

static size_t __pyx_f_14ms_peak_picker_2_c_8peak_set_double_binary_search_nearest_match_with_hint(double *__pyx_v_array, double __pyx_v_value, size_t __pyx_v_n, size_t __pyx_v_hint) {
  size_t __pyx_v_lo;
  size_t __pyx_v_hi;
  size_t __pyx_v_mid;
  size_t __pyx_v_i;
  size_t __pyx_v_best_ix;
  double __pyx_v_x;
  double __pyx_v_err;
  double __pyx_v_best_err;
  double __pyx_v_abs_err;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("double_binary_search_nearest_match_with_hint", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":644
 *         size_t i, best_ix
 *         double x, err, best_err, abs_err
 *     lo = hint             # <<<<<<<<<<<<<<
 *     hi = n
 *     while hi != lo:
 */
  __pyx_v_lo = __pyx_v_hint;

  /* "ms_peak_picker/_c/peak_set.pyx":645
 *         double x, err, best_err, abs_err
 *     lo = hint
 *     hi = n             # <<<<<<<<<<<<<<
 *     while hi != lo:
 *         mid = (hi + lo) // 2
 */
  __pyx_v_hi = __pyx_v_n;

  /* "ms_peak_picker/_c/peak_set.pyx":646
 *     lo = hint
 *     hi = n
 *     while hi != lo:             # <<<<<<<<<<<<<<
 *         mid = (hi + lo) // 2
 *         x = array[mid]
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_hi != __pyx_v_lo) != 0);
    if (!__pyx_t_1) break;

    /* "ms_peak_picker/_c/peak_set.pyx":647
 *     hi = n
 *     while hi != lo:
 *         mid = (hi + lo) // 2             # <<<<<<<<<<<<<<
 *         x = array[mid]
 *         err = x - value
 */
    __pyx_v_mid = ((__pyx_v_hi + __pyx_v_lo) / 2);

    /* "ms_peak_picker/_c/peak_set.pyx":648
 *     while hi != lo:
 *         mid = (hi + lo) // 2
 *         x = array[mid]             # <<<<<<<<<<<<<<
 *         err = x - value
 *         if err == 0 or ((hi - 1) == lo):
 */
    __pyx_v_x = (__pyx_v_array[__pyx_v_mid]);

    /* "ms_peak_picker/_c/peak_set.pyx":649
 *         mid = (hi + lo) // 2
 *         x = array[mid]
 *         err = x - value             # <<<<<<<<<<<<<<
 *         if err == 0 or ((hi - 1) == lo):
 *             i = mid
 */
    __pyx_v_err = (__pyx_v_x - __pyx_v_value);

    /* "ms_peak_picker/_c/peak_set.pyx":650
 *         x = array[mid]
 *         err = x - value
 *         if err == 0 or ((hi - 1) == lo):             # <<<<<<<<<<<<<<
 *             i = mid
 *             best_err = abs(err)
 */
    __pyx_t_2 = ((__pyx_v_err == 0.0) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_2 = (((__pyx_v_hi - 1) == __pyx_v_lo) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":651
 *         err = x - value
 *         if err == 0 or ((hi - 1) == lo):
 *             i = mid             # <<<<<<<<<<<<<<
 *             best_err = abs(err)
 *             best_ix = mid
 */
      __pyx_v_i = __pyx_v_mid;

      /* "ms_peak_picker/_c/peak_set.pyx":652
 *         if err == 0 or ((hi - 1) == lo):
 *             i = mid
 *             best_err = abs(err)             # <<<<<<<<<<<<<<
 *             best_ix = mid
 *             while i > 0:
 */
      __pyx_v_best_err = fabs(__pyx_v_err);

      /* "ms_peak_picker/_c/peak_set.pyx":653
 *             i = mid
 *             best_err = abs(err)
 *             best_ix = mid             # <<<<<<<<<<<<<<
 *             while i > 0:
 *                 i -= 1
 */
      __pyx_v_best_ix = __pyx_v_mid;

      /* "ms_peak_picker/_c/peak_set.pyx":654
 *             best_err = abs(err)
 *             best_ix = mid
 *             while i > 0:             # <<<<<<<<<<<<<<
 *                 i -= 1
 *                 x = array[i]
 */
      while (1) {
        __pyx_t_1 = ((__pyx_v_i > 0) != 0);
        if (!__pyx_t_1) break;

        /* "ms_peak_picker/_c/peak_set.pyx":655
 *             best_ix = mid
 *             while i > 0:
 *                 i -= 1             # <<<<<<<<<<<<<<
 *                 x = array[i]
 *                 err = (x - value)
 */
        __pyx_v_i = (__pyx_v_i - 1);

        /* "ms_peak_picker/_c/peak_set.pyx":656
 *             while i > 0:
 *                 i -= 1
 *                 x = array[i]             # <<<<<<<<<<<<<<
 *                 err = (x - value)
 *                 abs_err = abs(err)
 */
        __pyx_v_x = (__pyx_v_array[__pyx_v_i]);

        /* "ms_peak_picker/_c/peak_set.pyx":657
 *                 i -= 1
 *                 x = array[i]
 *                 err = (x - value)             # <<<<<<<<<<<<<<
 *                 abs_err = abs(err)
 *                 if abs_err > best_err:
 */
        __pyx_v_err = (__pyx_v_x - __pyx_v_value);

        /* "ms_peak_picker/_c/peak_set.pyx":658
 *                 x = array[i]
 *                 err = (x - value)
 *                 abs_err = abs(err)             # <<<<<<<<<<<<<<
 *                 if abs_err > best_err:
 *                     break
 */
        __pyx_v_abs_err = fabs(__pyx_v_err);

        /* "ms_peak_picker/_c/peak_set.pyx":659
 *                 err = (x - value)
 *                 abs_err = abs(err)
 *                 if abs_err > best_err:             # <<<<<<<<<<<<<<
 *                     break
 *                 elif abs_err < best_err:
 */
        __pyx_t_1 = ((__pyx_v_abs_err > __pyx_v_best_err) != 0);
        if (__pyx_t_1) {

          /* "ms_peak_picker/_c/peak_set.pyx":660
 *                 abs_err = abs(err)
 *                 if abs_err > best_err:
 *                     break             # <<<<<<<<<<<<<<
 *                 elif abs_err < best_err:
 *                     best_err = abs_err
 */
          goto __pyx_L9_break;

          /* "ms_peak_picker/_c/peak_set.pyx":659
 *                 err = (x - value)
 *                 abs_err = abs(err)
 *                 if abs_err > best_err:             # <<<<<<<<<<<<<<
 *                     break
 *                 elif abs_err < best_err:
 */
        }

        /* "ms_peak_picker/_c/peak_set.pyx":661
 *                 if abs_err > best_err:
 *                     break
 *                 elif abs_err < best_err:             # <<<<<<<<<<<<<<
 *                     best_err = abs_err
 *                     best_ix = i
 */
        __pyx_t_1 = ((__pyx_v_abs_err < __pyx_v_best_err) != 0);
        if (__pyx_t_1) {

          /* "ms_peak_picker/_c/peak_set.pyx":662
 *                     break
 *                 elif abs_err < best_err:
 *                     best_err = abs_err             # <<<<<<<<<<<<<<
 *                     best_ix = i
 *             i = mid
 */
          __pyx_v_best_err = __pyx_v_abs_err;

          /* "ms_peak_picker/_c/peak_set.pyx":663
 *                 elif abs_err < best_err:
 *                     best_err = abs_err
 *                     best_ix = i             # <<<<<<<<<<<<<<
 *             i = mid
 *             while i < n - 1:
 */
          __pyx_v_best_ix = __pyx_v_i;

          /* "ms_peak_picker/_c/peak_set.pyx":661
 *                 if abs_err > best_err:
 *                     break
 *                 elif abs_err < best_err:             # <<<<<<<<<<<<<<
 *                     best_err = abs_err
 *                     best_ix = i
 */
        }
      }
      __pyx_L9_break:;

      /* "ms_peak_picker/_c/peak_set.pyx":664
 *                     best_err = abs_err
 *                     best_ix = i
 *             i = mid             # <<<<<<<<<<<<<<
 *             while i < n - 1:
 *                 i += 1
 */
      __pyx_v_i = __pyx_v_mid;

      /* "ms_peak_picker/_c/peak_set.pyx":665
 *                     best_ix = i
 *             i = mid
 *             while i < n - 1:             # <<<<<<<<<<<<<<
 *                 i += 1
 *                 x = array[i]
 */
      while (1) {
        __pyx_t_1 = ((__pyx_v_i < (__pyx_v_n - 1)) != 0);
        if (!__pyx_t_1) break;

        /* "ms_peak_picker/_c/peak_set.pyx":666
 *             i = mid
 *             while i < n - 1:
 *                 i += 1             # <<<<<<<<<<<<<<
 *                 x = array[i]
 *                 err = (x - value)
 */
        __pyx_v_i = (__pyx_v_i + 1);

        /* "ms_peak_picker/_c/peak_set.pyx":667
 *             while i < n - 1:
 *                 i += 1
 *                 x = array[i]             # <<<<<<<<<<<<<<
 *                 err = (x - value)
 *                 abs_err = abs(err)
 */
        __pyx_v_x = (__pyx_v_array[__pyx_v_i]);

        /* "ms_peak_picker/_c/peak_set.pyx":668
 *                 i += 1
 *                 x = array[i]
 *                 err = (x - value)             # <<<<<<<<<<<<<<
 *                 abs_err = abs(err)
 *                 if abs_err > best_err:
 */
        __pyx_v_err = (__pyx_v_x - __pyx_v_value);

        /* "ms_peak_picker/_c/peak_set.pyx":669
 *                 x = array[i]
 *                 err = (x - value)
 *                 abs_err = abs(err)             # <<<<<<<<<<<<<<
 *                 if abs_err > best_err:
 *                     break
 */
        __pyx_v_abs_err = fabs(__pyx_v_err);

        /* "ms_peak_picker/_c/peak_set.pyx":670
 *                 err = (x - value)
 *                 abs_err = abs(err)
 *                 if abs_err > best_err:             # <<<<<<<<<<<<<<
 *                     break
 *                 elif abs_err < best_err:
 */
        __pyx_t_1 = ((__pyx_v_abs_err > __pyx_v_best_err) != 0);
        if (__pyx_t_1) {

          /* "ms_peak_picker/_c/peak_set.pyx":671
 *                 abs_err = abs(err)
 *                 if abs_err > best_err:
 *                     break             # <<<<<<<<<<<<<<
 *                 elif abs_err < best_err:
 *                     best_err = abs_err
 */
          goto __pyx_L12_break;

          /* "ms_peak_picker/_c/peak_set.pyx":670
 *                 err = (x - value)
 *                 abs_err = abs(err)
 *                 if abs_err > best_err:             # <<<<<<<<<<<<<<
 *                     break
 *                 elif abs_err < best_err:
 */
        }

        /* "ms_peak_picker/_c/peak_set.pyx":672
 *                 if abs_err > best_err:
 *                     break
 *                 elif abs_err < best_err:             # <<<<<<<<<<<<<<
 *                     best_err = abs_err
 *                     best_ix = i
 */
        __pyx_t_1 = ((__pyx_v_abs_err < __pyx_v_best_err) != 0);
        if (__pyx_t_1) {

          /* "ms_peak_picker/_c/peak_set.pyx":673
 *                     break
 *                 elif abs_err < best_err:
 *                     best_err = abs_err             # <<<<<<<<<<<<<<
 *                     best_ix = i
 *             return best_ix
 */
          __pyx_v_best_err = __pyx_v_abs_err;

          /* "ms_peak_picker/_c/peak_set.pyx":674
 *                 elif abs_err < best_err:
 *                     best_err = abs_err
 *                     best_ix = i             # <<<<<<<<<<<<<<
 *             return best_ix
 *         elif err > 0:
 */
          __pyx_v_best_ix = __pyx_v_i;

          /* "ms_peak_picker/_c/peak_set.pyx":672
 *                 if abs_err > best_err:
 *                     break
 *                 elif abs_err < best_err:             # <<<<<<<<<<<<<<
 *                     best_err = abs_err
 *                     best_ix = i
 */
        }
      }
      __pyx_L12_break:;

      /* "ms_peak_picker/_c/peak_set.pyx":675
 *                     best_err = abs_err
 *                     best_ix = i
 *             return best_ix             # <<<<<<<<<<<<<<
 *         elif err > 0:
 *             hi = mid
 */
      __pyx_r = __pyx_v_best_ix;
      goto __pyx_L0;

      /* "ms_peak_picker/_c/peak_set.pyx":650
 *         x = array[mid]
 *         err = x - value
 *         if err == 0 or ((hi - 1) == lo):             # <<<<<<<<<<<<<<
 *             i = mid
 *             best_err = abs(err)
 */
    }

    /* "ms_peak_picker/_c/peak_set.pyx":676
 *                     best_ix = i
 *             return best_ix
 *         elif err > 0:             # <<<<<<<<<<<<<<
 *             hi = mid
 *         else:
 */
    __pyx_t_1 = ((__pyx_v_err > 0.0) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":677
 *             return best_ix
 *         elif err > 0:
 *             hi = mid             # <<<<<<<<<<<<<<
 *         else:
 *             lo = mid
 */
      __pyx_v_hi = __pyx_v_mid;

      /* "ms_peak_picker/_c/peak_set.pyx":676
 *                     best_ix = i
 *             return best_ix
 *         elif err > 0:             # <<<<<<<<<<<<<<
 *             hi = mid
 *         else:
 */
      goto __pyx_L5;
    }

    /* "ms_peak_picker/_c/peak_set.pyx":679
 *             hi = mid
 *         else:
 *             lo = mid             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    /*else*/ {
      __pyx_v_lo = __pyx_v_mid;
    }
    __pyx_L5:;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":680
 *         else:
 *             lo = mid
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":639
 * 
 * @cython.cdivision(True)
 * cdef size_t double_binary_search_nearest_match_with_hint(double* array, double value, size_t n, size_t hint):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t lo, hi, mid
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":688
 * 
 *     @staticmethod
 *     cdef PeakSet _create(tuple peaks):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PeakSetIndexed inst
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__create(PyObject *__pyx_v_peaks) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_inst = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":691
 *         cdef:
 *             PeakSetIndexed inst
 *         inst = PeakSetIndexed.__new__(PeakSetIndexed)             # <<<<<<<<<<<<<<
 *         inst.peaks = peaks
 *         inst.mz_index = NULL
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed(((PyTypeObject *)__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_inst = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":692
 *             PeakSetIndexed inst
 *         inst = PeakSetIndexed.__new__(PeakSetIndexed)
 *         inst.peaks = peaks             # <<<<<<<<<<<<<<
 *         inst.mz_index = NULL
 *         inst.interval_index = NULL
 */
  __Pyx_INCREF(__pyx_v_peaks);
  __Pyx_GIVEREF(__pyx_v_peaks);
  __Pyx_GOTREF(__pyx_v_inst->__pyx_base.peaks);
  __Pyx_DECREF(__pyx_v_inst->__pyx_base.peaks);
  __pyx_v_inst->__pyx_base.peaks = __pyx_v_peaks;

  /* "ms_peak_picker/_c/peak_set.pyx":693
 *         inst = PeakSetIndexed.__new__(PeakSetIndexed)
 *         inst.peaks = peaks
 *         inst.mz_index = NULL             # <<<<<<<<<<<<<<
 *         inst.interval_index = NULL
 *         return inst
 */
  __pyx_v_inst->mz_index = NULL;

  /* "ms_peak_picker/_c/peak_set.pyx":694
 *         inst.peaks = peaks
 *         inst.mz_index = NULL
 *         inst.interval_index = NULL             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_v_inst->interval_index = NULL;

  /* "ms_peak_picker/_c/peak_set.pyx":695
 *         inst.mz_index = NULL
 *         inst.interval_index = NULL
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, peaks):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_inst));
  __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_inst);
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":688
 * 
 *     @staticmethod
 *     cdef PeakSet _create(tuple peaks):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PeakSetIndexed inst
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":697
 *         return inst
 * 
 *     def __init__(self, peaks):             # <<<<<<<<<<<<<<
 *         PeakSet.__init__(self, peaks)
 *         self.mz_index = NULL
 */

/* Python wrapper */
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_peaks = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaks,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaks)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 697, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_peaks = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 697, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed___init__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self), __pyx_v_peaks);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed___init__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, PyObject *__pyx_v_peaks) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":698
 * 
 *     def __init__(self, peaks):
 *         PeakSet.__init__(self, peaks)             # <<<<<<<<<<<<<<
 *         self.mz_index = NULL
 *         self.interval_index = NULL
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet), __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_self), __pyx_v_peaks};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_self), __pyx_v_peaks};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, ((PyObject *)__pyx_v_self));
    __Pyx_INCREF(__pyx_v_peaks);
    __Pyx_GIVEREF(__pyx_v_peaks);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_peaks);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":699
 *     def __init__(self, peaks):
 *         PeakSet.__init__(self, peaks)
 *         self.mz_index = NULL             # <<<<<<<<<<<<<<
 *         self.interval_index = NULL
 * 
 */
  __pyx_v_self->mz_index = NULL;

  /* "ms_peak_picker/_c/peak_set.pyx":700
 *         PeakSet.__init__(self, peaks)
 *         self.mz_index = NULL
 *         self.interval_index = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->interval_index = NULL;

  /* "ms_peak_picker/_c/peak_set.pyx":697
 *         return inst
 * 
 *     def __init__(self, peaks):             # <<<<<<<<<<<<<<
 *         PeakSet.__init__(self, peaks)
 *         self.mz_index = NULL
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":702
 *         self.interval_index = NULL
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.mz_index != NULL:
 *             free(self.mz_index)
 */

/* Python wrapper */
static void __pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_2__dealloc__(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_2__dealloc__(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":703
 * 
 *     def __dealloc__(self):
 *         if self.mz_index != NULL:             # <<<<<<<<<<<<<<
 *             free(self.mz_index)
 *         if self.interval_index != NULL:
 */
  __pyx_t_1 = ((__pyx_v_self->mz_index != NULL) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":704
 *     def __dealloc__(self):
 *         if self.mz_index != NULL:
 *             free(self.mz_index)             # <<<<<<<<<<<<<<
 *         if self.interval_index != NULL:
 *             free_index_list(self.interval_index)
 */
    free(__pyx_v_self->mz_index);

    /* "ms_peak_picker/_c/peak_set.pyx":703
 * 
 *     def __dealloc__(self):
 *         if self.mz_index != NULL:             # <<<<<<<<<<<<<<
 *             free(self.mz_index)
 *         if self.interval_index != NULL:
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":705
 *         if self.mz_index != NULL:
 *             free(self.mz_index)
 *         if self.interval_index != NULL:             # <<<<<<<<<<<<<<
 *             free_index_list(self.interval_index)
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->interval_index != NULL) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":706
 *             free(self.mz_index)
 *         if self.interval_index != NULL:
 *             free_index_list(self.interval_index)             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakSet clone(self):
 */
    __pyx_f_14ms_peak_picker_2_c_8peak_set_free_index_list(__pyx_v_self->interval_index);

    /* "ms_peak_picker/_c/peak_set.pyx":705
 *         if self.mz_index != NULL:
 *             free(self.mz_index)
 *         if self.interval_index != NULL:             # <<<<<<<<<<<<<<
 *             free_index_list(self.interval_index)
 * 
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":702
 *         self.interval_index = NULL
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.mz_index != NULL:
 *             free(self.mz_index)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "ms_peak_picker/_c/peak_set.pyx":708
 *             free_index_list(self.interval_index)
 * 
 *     cpdef PeakSet clone(self):             # <<<<<<<<<<<<<<
 *         cdef PeakSetIndexed inst = PeakSetIndexed(tuple([p.clone() for p in self]))
 *         if self.indexed:
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_5clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_clone(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_inst = 0;
  PyObject *__pyx_v_p = NULL;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_clone); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 708, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_5clone)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 708, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet))))) __PYX_ERR(0, 708, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_peak_picker/_c/peak_set.pyx":709
 * 
 *     cpdef PeakSet clone(self):
 *         cdef PeakSetIndexed inst = PeakSetIndexed(tuple([p.clone() for p in self]))             # <<<<<<<<<<<<<<
 *         if self.indexed:
 *             inst.reindex()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(((PyObject *)__pyx_v_self))) || PyTuple_CheckExact(((PyObject *)__pyx_v_self))) {
    __pyx_t_2 = ((PyObject *)__pyx_v_self); __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 709, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 709, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 709, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 709, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 709, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 709, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 709, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_p, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_p, __pyx_n_s_clone); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 709, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 709, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 709, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_AsTuple(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed), __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_inst = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":710
 *     cpdef PeakSet clone(self):
 *         cdef PeakSetIndexed inst = PeakSetIndexed(tuple([p.clone() for p in self]))
 *         if self.indexed:             # <<<<<<<<<<<<<<
 *             inst.reindex()
 *         return inst
 */
  __pyx_t_8 = (__pyx_v_self->__pyx_base.indexed != 0);
  if (__pyx_t_8) {

    /* "ms_peak_picker/_c/peak_set.pyx":711
 *         cdef PeakSetIndexed inst = PeakSetIndexed(tuple([p.clone() for p in self]))
 *         if self.indexed:
 *             inst.reindex()             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_inst->__pyx_base.__pyx_vtab)->__pyx_base.reindex(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_inst), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 711, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ms_peak_picker/_c/peak_set.pyx":710
 *     cpdef PeakSet clone(self):
 *         cdef PeakSetIndexed inst = PeakSetIndexed(tuple([p.clone() for p in self]))
 *         if self.indexed:             # <<<<<<<<<<<<<<
 *             inst.reindex()
 *         return inst
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":712
 *         if self.indexed:
 *             inst.reindex()
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     cpdef _allocate_index(self, size_t interval_index_size):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_inst));
  __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_inst);
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":708
 *             free_index_list(self.interval_index)
 * 
 *     cpdef PeakSet clone(self):             # <<<<<<<<<<<<<<
 *         cdef PeakSetIndexed inst = PeakSetIndexed(tuple([p.clone() for p in self]))
 *         if self.indexed:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XDECREF(__pyx_v_p);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_5clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_4clone[] = "PeakSetIndexed.clone(self) -> PeakSet";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_5clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_4clone(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_4clone(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_clone(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":714
 *         return inst
 * 
 *     cpdef _allocate_index(self, size_t interval_index_size):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_7_allocate_index(PyObject *__pyx_v_self, PyObject *__pyx_arg_interval_index_size); /*proto*/
static PyObject *__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__allocate_index(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, size_t __pyx_v_interval_index_size, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_p = 0;
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *__pyx_v_interval_index;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  double __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_allocate_index", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_allocate_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 714, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_7_allocate_index)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_interval_index_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 714, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 714, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_peak_picker/_c/peak_set.pyx":719
 *             FittedPeak p
 *             index_list* interval_index
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         if self.mz_index != NULL:
 *             free(self.mz_index)
 */
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_size(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self));

  /* "ms_peak_picker/_c/peak_set.pyx":720
 *             index_list* interval_index
 *         n = self.get_size()
 *         if self.mz_index != NULL:             # <<<<<<<<<<<<<<
 *             free(self.mz_index)
 *             self.mz_index = NULL
 */
  __pyx_t_6 = ((__pyx_v_self->mz_index != NULL) != 0);
  if (__pyx_t_6) {

    /* "ms_peak_picker/_c/peak_set.pyx":721
 *         n = self.get_size()
 *         if self.mz_index != NULL:
 *             free(self.mz_index)             # <<<<<<<<<<<<<<
 *             self.mz_index = NULL
 *         self.mz_index = <double*>malloc(sizeof(double) * n)
 */
    free(__pyx_v_self->mz_index);

    /* "ms_peak_picker/_c/peak_set.pyx":722
 *         if self.mz_index != NULL:
 *             free(self.mz_index)
 *             self.mz_index = NULL             # <<<<<<<<<<<<<<
 *         self.mz_index = <double*>malloc(sizeof(double) * n)
 *         for i in range(n):
 */
    __pyx_v_self->mz_index = NULL;

    /* "ms_peak_picker/_c/peak_set.pyx":720
 *             index_list* interval_index
 *         n = self.get_size()
 *         if self.mz_index != NULL:             # <<<<<<<<<<<<<<
 *             free(self.mz_index)
 *             self.mz_index = NULL
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":723
 *             free(self.mz_index)
 *             self.mz_index = NULL
 *         self.mz_index = <double*>malloc(sizeof(double) * n)             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             p = self.getitem(i)
 */
  __pyx_v_self->mz_index = ((double *)malloc(((sizeof(double)) * __pyx_v_n)));

  /* "ms_peak_picker/_c/peak_set.pyx":724
 *             self.mz_index = NULL
 *         self.mz_index = <double*>malloc(sizeof(double) * n)
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             p = self.getitem(i)
 *             self.mz_index[i] = p.mz
 */
  __pyx_t_7 = __pyx_v_n;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "ms_peak_picker/_c/peak_set.pyx":725
 *         self.mz_index = <double*>malloc(sizeof(double) * n)
 *         for i in range(n):
 *             p = self.getitem(i)             # <<<<<<<<<<<<<<
 *             self.mz_index[i] = p.mz
 * 
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.getitem(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 725, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_p, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "ms_peak_picker/_c/peak_set.pyx":726
 *         for i in range(n):
 *             p = self.getitem(i)
 *             self.mz_index[i] = p.mz             # <<<<<<<<<<<<<<
 * 
 *         if n > 2 and interval_index_size > 0:
 */
    __pyx_t_10 = __pyx_v_p->__pyx_base.mz;
    (__pyx_v_self->mz_index[__pyx_v_i]) = __pyx_t_10;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":728
 *             self.mz_index[i] = p.mz
 * 
 *         if n > 2 and interval_index_size > 0:             # <<<<<<<<<<<<<<
 *         # if False:
 *             if self.interval_index != NULL:
 */
  __pyx_t_11 = ((__pyx_v_n > 2) != 0);
  if (__pyx_t_11) {
  } else {
    __pyx_t_6 = __pyx_t_11;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_11 = ((__pyx_v_interval_index_size > 0) != 0);
  __pyx_t_6 = __pyx_t_11;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_6) {

    /* "ms_peak_picker/_c/peak_set.pyx":730
 *         if n > 2 and interval_index_size > 0:
 *         # if False:
 *             if self.interval_index != NULL:             # <<<<<<<<<<<<<<
 *                 free_index_list(self.interval_index)
 *                 self.interval_index = NULL
 */
    __pyx_t_6 = ((__pyx_v_self->interval_index != NULL) != 0);
    if (__pyx_t_6) {

      /* "ms_peak_picker/_c/peak_set.pyx":731
 *         # if False:
 *             if self.interval_index != NULL:
 *                 free_index_list(self.interval_index)             # <<<<<<<<<<<<<<
 *                 self.interval_index = NULL
 *             interval_index = <index_list*>malloc(sizeof(index_list))
 */
      __pyx_f_14ms_peak_picker_2_c_8peak_set_free_index_list(__pyx_v_self->interval_index);

      /* "ms_peak_picker/_c/peak_set.pyx":732
 *             if self.interval_index != NULL:
 *                 free_index_list(self.interval_index)
 *                 self.interval_index = NULL             # <<<<<<<<<<<<<<
 *             interval_index = <index_list*>malloc(sizeof(index_list))
 *             build_interval_index(self, interval_index, interval_index_size)
 */
      __pyx_v_self->interval_index = NULL;

      /* "ms_peak_picker/_c/peak_set.pyx":730
 *         if n > 2 and interval_index_size > 0:
 *         # if False:
 *             if self.interval_index != NULL:             # <<<<<<<<<<<<<<
 *                 free_index_list(self.interval_index)
 *                 self.interval_index = NULL
 */
    }

    /* "ms_peak_picker/_c/peak_set.pyx":733
 *                 free_index_list(self.interval_index)
 *                 self.interval_index = NULL
 *             interval_index = <index_list*>malloc(sizeof(index_list))             # <<<<<<<<<<<<<<
 *             build_interval_index(self, interval_index, interval_index_size)
 *             if check_index(interval_index) != 0:
 */
    __pyx_v_interval_index = ((struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *)malloc((sizeof(struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list))));

    /* "ms_peak_picker/_c/peak_set.pyx":734
 *                 self.interval_index = NULL
 *             interval_index = <index_list*>malloc(sizeof(index_list))
 *             build_interval_index(self, interval_index, interval_index_size)             # <<<<<<<<<<<<<<
 *             if check_index(interval_index) != 0:
 *                 free_index_list(interval_index)
 */
    (void)(__pyx_f_14ms_peak_picker_2_c_8peak_set_build_interval_index(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), __pyx_v_interval_index, __pyx_v_interval_index_size));

    /* "ms_peak_picker/_c/peak_set.pyx":735
 *             interval_index = <index_list*>malloc(sizeof(index_list))
 *             build_interval_index(self, interval_index, interval_index_size)
 *             if check_index(interval_index) != 0:             # <<<<<<<<<<<<<<
 *                 free_index_list(interval_index)
 *             else:
 */
    __pyx_t_6 = ((__pyx_f_14ms_peak_picker_2_c_8peak_set_check_index(__pyx_v_interval_index) != 0) != 0);
    if (__pyx_t_6) {

      /* "ms_peak_picker/_c/peak_set.pyx":736
 *             build_interval_index(self, interval_index, interval_index_size)
 *             if check_index(interval_index) != 0:
 *                 free_index_list(interval_index)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.interval_index = interval_index
 */
      __pyx_f_14ms_peak_picker_2_c_8peak_set_free_index_list(__pyx_v_interval_index);

      /* "ms_peak_picker/_c/peak_set.pyx":735
 *             interval_index = <index_list*>malloc(sizeof(index_list))
 *             build_interval_index(self, interval_index, interval_index_size)
 *             if check_index(interval_index) != 0:             # <<<<<<<<<<<<<<
 *                 free_index_list(interval_index)
 *             else:
 */
      goto __pyx_L10;
    }

    /* "ms_peak_picker/_c/peak_set.pyx":738
 *                 free_index_list(interval_index)
 *             else:
 *                 self.interval_index = interval_index             # <<<<<<<<<<<<<<
 *         else:
 *             if self.interval_index != NULL:
 */
    /*else*/ {
      __pyx_v_self->interval_index = __pyx_v_interval_index;
    }
    __pyx_L10:;

    /* "ms_peak_picker/_c/peak_set.pyx":728
 *             self.mz_index[i] = p.mz
 * 
 *         if n > 2 and interval_index_size > 0:             # <<<<<<<<<<<<<<
 *         # if False:
 *             if self.interval_index != NULL:
 */
    goto __pyx_L6;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":740
 *                 self.interval_index = interval_index
 *         else:
 *             if self.interval_index != NULL:             # <<<<<<<<<<<<<<
 *                 free_index_list(self.interval_index)
 *                 self.interval_index = NULL
 */
  /*else*/ {
    __pyx_t_6 = ((__pyx_v_self->interval_index != NULL) != 0);
    if (__pyx_t_6) {

      /* "ms_peak_picker/_c/peak_set.pyx":741
 *         else:
 *             if self.interval_index != NULL:
 *                 free_index_list(self.interval_index)             # <<<<<<<<<<<<<<
 *                 self.interval_index = NULL
 * 
 */
      __pyx_f_14ms_peak_picker_2_c_8peak_set_free_index_list(__pyx_v_self->interval_index);

      /* "ms_peak_picker/_c/peak_set.pyx":742
 *             if self.interval_index != NULL:
 *                 free_index_list(self.interval_index)
 *                 self.interval_index = NULL             # <<<<<<<<<<<<<<
 * 
 *     cpdef size_t _index(self):
 */
      __pyx_v_self->interval_index = NULL;

      /* "ms_peak_picker/_c/peak_set.pyx":740
 *                 self.interval_index = interval_index
 *         else:
 *             if self.interval_index != NULL:             # <<<<<<<<<<<<<<
 *                 free_index_list(self.interval_index)
 *                 self.interval_index = NULL
 */
    }
  }
  __pyx_L6:;

  /* "ms_peak_picker/_c/peak_set.pyx":714
 *         return inst
 * 
 *     cpdef _allocate_index(self, size_t interval_index_size):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed._allocate_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_p);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_7_allocate_index(PyObject *__pyx_v_self, PyObject *__pyx_arg_interval_index_size); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_6_allocate_index[] = "PeakSetIndexed._allocate_index(self, size_t interval_index_size)";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_7_allocate_index(PyObject *__pyx_v_self, PyObject *__pyx_arg_interval_index_size) {
  size_t __pyx_v_interval_index_size;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_allocate_index (wrapper)", 0);
  assert(__pyx_arg_interval_index_size); {
    __pyx_v_interval_index_size = __Pyx_PyInt_As_size_t(__pyx_arg_interval_index_size); if (unlikely((__pyx_v_interval_index_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 714, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed._allocate_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_6_allocate_index(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self), ((size_t)__pyx_v_interval_index_size));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_6_allocate_index(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, size_t __pyx_v_interval_index_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_allocate_index", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__allocate_index(__pyx_v_self, __pyx_v_interval_index_size, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed._allocate_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":744
 *                 self.interval_index = NULL
 * 
 *     cpdef size_t _index(self):             # <<<<<<<<<<<<<<
 *         i = PeakSet._index(self)
 *         self._allocate_index(INTERVAL_INDEX_SIZE)
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_9_index(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static size_t __pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__index(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_index", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_index_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 744, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_9_index)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_5 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 744, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_peak_picker/_c/peak_set.pyx":745
 * 
 *     cpdef size_t _index(self):
 *         i = PeakSet._index(self)             # <<<<<<<<<<<<<<
 *         self._allocate_index(INTERVAL_INDEX_SIZE)
 *         return i
 */
  __pyx_v_i = __pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__index(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), 1);

  /* "ms_peak_picker/_c/peak_set.pyx":746
 *     cpdef size_t _index(self):
 *         i = PeakSet._index(self)
 *         self._allocate_index(INTERVAL_INDEX_SIZE)             # <<<<<<<<<<<<<<
 *         return i
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self->__pyx_base.__pyx_vtab)->_allocate_index(__pyx_v_self, __pyx_v_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":747
 *         i = PeakSet._index(self)
 *         self._allocate_index(INTERVAL_INDEX_SIZE)
 *         return i             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
  __pyx_r = __pyx_v_i;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":744
 *                 self.interval_index = NULL
 * 
 *     cpdef size_t _index(self):             # <<<<<<<<<<<<<<
 *         i = PeakSet._index(self)
 *         self._allocate_index(INTERVAL_INDEX_SIZE)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("ms_peak_picker._c.peak_set.PeakSetIndexed._index", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_9_index(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_8_index[] = "PeakSetIndexed._index(self) -> size_t";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_9_index(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_index (wrapper)", 0);
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_8_index(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_8_index(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_index", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__index(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed._index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":750
 * 
 *     @cython.cdivision(True)
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=1e-5):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n, s
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__has_peak(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, double __pyx_v_mz, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__has_peak *__pyx_optional_args) {
  double __pyx_v_tolerance = ((double)1e-5);
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  size_t __pyx_v_s;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_has_peak", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_tolerance = __pyx_optional_args->tolerance;
    }
  }

  /* "ms_peak_picker/_c/peak_set.pyx":754
 *             size_t i, n, s
 *             FittedPeak peak
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         if n == 0:
 *             return _null_peak
 */
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_size(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self));

  /* "ms_peak_picker/_c/peak_set.pyx":755
 *             FittedPeak peak
 *         n = self.get_size()
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return _null_peak
 *         if self.interval_index != NULL:
 */
  __pyx_t_1 = ((__pyx_v_n == 0) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":756
 *         n = self.get_size()
 *         if n == 0:
 *             return _null_peak             # <<<<<<<<<<<<<<
 *         if self.interval_index != NULL:
 *             find_search_interval(self.interval_index, mz, &s, &n)
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak));
    __pyx_r = __pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":755
 *             FittedPeak peak
 *         n = self.get_size()
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return _null_peak
 *         if self.interval_index != NULL:
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":757
 *         if n == 0:
 *             return _null_peak
 *         if self.interval_index != NULL:             # <<<<<<<<<<<<<<
 *             find_search_interval(self.interval_index, mz, &s, &n)
 *             i = double_binary_search_ppm_with_hint(self.mz_index, mz, tolerance, n, s)
 */
  __pyx_t_1 = ((__pyx_v_self->interval_index != NULL) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":758
 *             return _null_peak
 *         if self.interval_index != NULL:
 *             find_search_interval(self.interval_index, mz, &s, &n)             # <<<<<<<<<<<<<<
 *             i = double_binary_search_ppm_with_hint(self.mz_index, mz, tolerance, n, s)
 *         else:
 */
    (void)(__pyx_f_14ms_peak_picker_2_c_8peak_set_find_search_interval(__pyx_v_self->interval_index, __pyx_v_mz, (&__pyx_v_s), (&__pyx_v_n)));

    /* "ms_peak_picker/_c/peak_set.pyx":759
 *         if self.interval_index != NULL:
 *             find_search_interval(self.interval_index, mz, &s, &n)
 *             i = double_binary_search_ppm_with_hint(self.mz_index, mz, tolerance, n, s)             # <<<<<<<<<<<<<<
 *         else:
 *             i = double_binary_search_ppm(self.mz_index, mz, tolerance, n)
 */
    __pyx_v_i = __pyx_f_14ms_peak_picker_2_c_8peak_set_double_binary_search_ppm_with_hint(__pyx_v_self->mz_index, __pyx_v_mz, __pyx_v_tolerance, __pyx_v_n, __pyx_v_s);

    /* "ms_peak_picker/_c/peak_set.pyx":757
 *         if n == 0:
 *             return _null_peak
 *         if self.interval_index != NULL:             # <<<<<<<<<<<<<<
 *             find_search_interval(self.interval_index, mz, &s, &n)
 *             i = double_binary_search_ppm_with_hint(self.mz_index, mz, tolerance, n, s)
 */
    goto __pyx_L4;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":761
 *             i = double_binary_search_ppm_with_hint(self.mz_index, mz, tolerance, n, s)
 *         else:
 *             i = double_binary_search_ppm(self.mz_index, mz, tolerance, n)             # <<<<<<<<<<<<<<
 *         peak = self.getitem(i)
 *         if abs((peak.mz - mz) / mz) < tolerance:
 */
  /*else*/ {
    __pyx_v_i = __pyx_f_14ms_peak_picker_2_c_8peak_set_double_binary_search_ppm(__pyx_v_self->mz_index, __pyx_v_mz, __pyx_v_tolerance, __pyx_v_n);
  }
  __pyx_L4:;

  /* "ms_peak_picker/_c/peak_set.pyx":762
 *         else:
 *             i = double_binary_search_ppm(self.mz_index, mz, tolerance, n)
 *         peak = self.getitem(i)             # <<<<<<<<<<<<<<
 *         if abs((peak.mz - mz) / mz) < tolerance:
 *             return peak
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.getitem(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), __pyx_v_i)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":763
 *             i = double_binary_search_ppm(self.mz_index, mz, tolerance, n)
 *         peak = self.getitem(i)
 *         if abs((peak.mz - mz) / mz) < tolerance:             # <<<<<<<<<<<<<<
 *             return peak
 *         else:
 */
  __pyx_t_1 = ((fabs(((__pyx_v_peak->__pyx_base.mz - __pyx_v_mz) / __pyx_v_mz)) < __pyx_v_tolerance) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":764
 *         peak = self.getitem(i)
 *         if abs((peak.mz - mz) / mz) < tolerance:
 *             return peak             # <<<<<<<<<<<<<<
 *         else:
 *             return _null_peak
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_peak));
    __pyx_r = __pyx_v_peak;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":763
 *             i = double_binary_search_ppm(self.mz_index, mz, tolerance, n)
 *         peak = self.getitem(i)
 *         if abs((peak.mz - mz) / mz) < tolerance:             # <<<<<<<<<<<<<<
 *             return peak
 *         else:
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":766
 *             return peak
 *         else:
 *             return _null_peak             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
  /*else*/ {
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak));
    __pyx_r = __pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak;
    goto __pyx_L0;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":750
 * 
 *     @cython.cdivision(True)
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=1e-5):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n, s
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed._has_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":769
 * 
 *     @cython.cdivision(True)
 *     def get_nearest_peak(self, double mz):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double errout
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_11get_nearest_peak(PyObject *__pyx_v_self, PyObject *__pyx_arg_mz); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_10get_nearest_peak[] = "PeakSetIndexed.get_nearest_peak(self, double mz)";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_11get_nearest_peak(PyObject *__pyx_v_self, PyObject *__pyx_arg_mz) {
  double __pyx_v_mz;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_nearest_peak (wrapper)", 0);
  assert(__pyx_arg_mz); {
    __pyx_v_mz = __pyx_PyFloat_AsDouble(__pyx_arg_mz); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 769, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed.get_nearest_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_10get_nearest_peak(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self), ((double)__pyx_v_mz));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_10get_nearest_peak(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, double __pyx_v_mz) {
  double __pyx_v_errout;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_nearest_peak", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":773
 *             double errout
 *             FittedPeak peak
 *         peak = self._get_nearest_peak(mz, &errout)             # <<<<<<<<<<<<<<
 *         return peak, errout
 * 
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._get_nearest_peak(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), __pyx_v_mz, (&__pyx_v_errout))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":774
 *             FittedPeak peak
 *         peak = self._get_nearest_peak(mz, &errout)
 *         return peak, errout             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_errout); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_peak));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_peak));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_peak));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":769
 * 
 *     @cython.cdivision(True)
 *     def get_nearest_peak(self, double mz):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double errout
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed.get_nearest_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":777
 * 
 *     @cython.cdivision(True)
 *     cdef FittedPeak _get_nearest_peak(self, double mz, double* errout):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n, s
 */

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__get_nearest_peak(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, double __pyx_v_mz, double *__pyx_v_errout) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  size_t __pyx_v_s;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_nearest_peak", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":781
 *             size_t i, n, s
 *             FittedPeak peak
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         if n == 0:
 *             errout[0] = INF
 */
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_size(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self));

  /* "ms_peak_picker/_c/peak_set.pyx":782
 *             FittedPeak peak
 *         n = self.get_size()
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             errout[0] = INF
 *             return _null_peak
 */
  __pyx_t_1 = ((__pyx_v_n == 0) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":783
 *         n = self.get_size()
 *         if n == 0:
 *             errout[0] = INF             # <<<<<<<<<<<<<<
 *             return _null_peak
 *         if self.interval_index != NULL:
 */
    (__pyx_v_errout[0]) = __pyx_v_14ms_peak_picker_2_c_8peak_set_INF;

    /* "ms_peak_picker/_c/peak_set.pyx":784
 *         if n == 0:
 *             errout[0] = INF
 *             return _null_peak             # <<<<<<<<<<<<<<
 *         if self.interval_index != NULL:
 *             find_search_interval(self.interval_index, mz, &s, &n)
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak));
    __pyx_r = __pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":782
 *             FittedPeak peak
 *         n = self.get_size()
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             errout[0] = INF
 *             return _null_peak
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":785
 *             errout[0] = INF
 *             return _null_peak
 *         if self.interval_index != NULL:             # <<<<<<<<<<<<<<
 *             find_search_interval(self.interval_index, mz, &s, &n)
 *             i = double_binary_search_nearest_match_with_hint(self.mz_index, mz, n, s)
 */
  __pyx_t_1 = ((__pyx_v_self->interval_index != NULL) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":786
 *             return _null_peak
 *         if self.interval_index != NULL:
 *             find_search_interval(self.interval_index, mz, &s, &n)             # <<<<<<<<<<<<<<
 *             i = double_binary_search_nearest_match_with_hint(self.mz_index, mz, n, s)
 *         else:
 */
    (void)(__pyx_f_14ms_peak_picker_2_c_8peak_set_find_search_interval(__pyx_v_self->interval_index, __pyx_v_mz, (&__pyx_v_s), (&__pyx_v_n)));

    /* "ms_peak_picker/_c/peak_set.pyx":787
 *         if self.interval_index != NULL:
 *             find_search_interval(self.interval_index, mz, &s, &n)
 *             i = double_binary_search_nearest_match_with_hint(self.mz_index, mz, n, s)             # <<<<<<<<<<<<<<
 *         else:
 *             i = double_binary_search_nearest_match(self.mz_index, mz, n)
 */
    __pyx_v_i = __pyx_f_14ms_peak_picker_2_c_8peak_set_double_binary_search_nearest_match_with_hint(__pyx_v_self->mz_index, __pyx_v_mz, __pyx_v_n, __pyx_v_s);

    /* "ms_peak_picker/_c/peak_set.pyx":785
 *             errout[0] = INF
 *             return _null_peak
 *         if self.interval_index != NULL:             # <<<<<<<<<<<<<<
 *             find_search_interval(self.interval_index, mz, &s, &n)
 *             i = double_binary_search_nearest_match_with_hint(self.mz_index, mz, n, s)
 */
    goto __pyx_L4;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":789
 *             i = double_binary_search_nearest_match_with_hint(self.mz_index, mz, n, s)
 *         else:
 *             i = double_binary_search_nearest_match(self.mz_index, mz, n)             # <<<<<<<<<<<<<<
 *         peak = self.getitem(i)
 *         errout[0] = (peak.mz - mz) / mz
 */
  /*else*/ {
    __pyx_v_i = __pyx_f_14ms_peak_picker_2_c_8peak_set_double_binary_search_nearest_match(__pyx_v_self->mz_index, __pyx_v_mz, __pyx_v_n);
  }
  __pyx_L4:;

  /* "ms_peak_picker/_c/peak_set.pyx":790
 *         else:
 *             i = double_binary_search_nearest_match(self.mz_index, mz, n)
 *         peak = self.getitem(i)             # <<<<<<<<<<<<<<
 *         errout[0] = (peak.mz - mz) / mz
 *         return peak
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.getitem(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), __pyx_v_i)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":791
 *             i = double_binary_search_nearest_match(self.mz_index, mz, n)
 *         peak = self.getitem(i)
 *         errout[0] = (peak.mz - mz) / mz             # <<<<<<<<<<<<<<
 *         return peak
 * 
 */
  (__pyx_v_errout[0]) = ((__pyx_v_peak->__pyx_base.mz - __pyx_v_mz) / __pyx_v_mz);

  /* "ms_peak_picker/_c/peak_set.pyx":792
 *         peak = self.getitem(i)
 *         errout[0] = (peak.mz - mz) / mz
 *         return peak             # <<<<<<<<<<<<<<
 * 
 *     cpdef FittedPeak has_peak_hinted(self, double mz, double tolerance=2e-5, size_t hint_start=0, size_t hint_end=-1):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_peak));
  __pyx_r = __pyx_v_peak;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":777
 * 
 *     @cython.cdivision(True)
 *     cdef FittedPeak _get_nearest_peak(self, double mz, double* errout):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n, s
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed._get_nearest_peak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":794
 *         return peak
 * 
 *     cpdef FittedPeak has_peak_hinted(self, double mz, double tolerance=2e-5, size_t hint_start=0, size_t hint_end=-1):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_13has_peak_hinted(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, double __pyx_v_mz, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted *__pyx_optional_args) {
  double __pyx_v_tolerance = ((double)2e-5);
  size_t __pyx_v_hint_start = ((size_t)0);
  size_t __pyx_v_hint_end = ((size_t)-1L);
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  double __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("has_peak_hinted", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_tolerance = __pyx_optional_args->tolerance;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_hint_start = __pyx_optional_args->hint_start;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_hint_end = __pyx_optional_args->hint_end;
        }
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_has_peak_hinted); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 794, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_13has_peak_hinted)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 794, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_tolerance); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 794, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_hint_start); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 794, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyInt_FromSize_t(__pyx_v_hint_end); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 794, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_8 = NULL;
        __pyx_t_9 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_9 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 794, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 794, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        {
          __pyx_t_10 = PyTuple_New(4+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 794, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (__pyx_t_8) {
            __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_9, __pyx_t_6);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 794, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak))))) __PYX_ERR(0, 794, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "ms_peak_picker/_c/peak_set.pyx":798
 *             size_t i, n
 *             FittedPeak peak
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         if hint_start > n :
 *             raise ValueError("Hinted start index %d cannot be greater than the set size %d" % (hint_start, n))
 */
  __pyx_v_n = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_size(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self));

  /* "ms_peak_picker/_c/peak_set.pyx":799
 *             FittedPeak peak
 *         n = self.get_size()
 *         if hint_start > n :             # <<<<<<<<<<<<<<
 *             raise ValueError("Hinted start index %d cannot be greater than the set size %d" % (hint_start, n))
 *         if hint_end > n:
 */
  __pyx_t_11 = ((__pyx_v_hint_start > __pyx_v_n) != 0);
  if (unlikely(__pyx_t_11)) {

    /* "ms_peak_picker/_c/peak_set.pyx":800
 *         n = self.get_size()
 *         if hint_start > n :
 *             raise ValueError("Hinted start index %d cannot be greater than the set size %d" % (hint_start, n))             # <<<<<<<<<<<<<<
 *         if hint_end > n:
 *             raise ValueError("Hinted start index %d cannot be greater than the set size %d" % (hint_end, n))
 */
    __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_hint_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 800, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 800, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 800, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_Hinted_start_index_d_cannot_be_g, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 800, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 800, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 800, __pyx_L1_error)

    /* "ms_peak_picker/_c/peak_set.pyx":799
 *             FittedPeak peak
 *         n = self.get_size()
 *         if hint_start > n :             # <<<<<<<<<<<<<<
 *             raise ValueError("Hinted start index %d cannot be greater than the set size %d" % (hint_start, n))
 *         if hint_end > n:
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":801
 *         if hint_start > n :
 *             raise ValueError("Hinted start index %d cannot be greater than the set size %d" % (hint_start, n))
 *         if hint_end > n:             # <<<<<<<<<<<<<<
 *             raise ValueError("Hinted start index %d cannot be greater than the set size %d" % (hint_end, n))
 *         if n == 0:
 */
  __pyx_t_11 = ((__pyx_v_hint_end > __pyx_v_n) != 0);
  if (unlikely(__pyx_t_11)) {

    /* "ms_peak_picker/_c/peak_set.pyx":802
 *             raise ValueError("Hinted start index %d cannot be greater than the set size %d" % (hint_start, n))
 *         if hint_end > n:
 *             raise ValueError("Hinted start index %d cannot be greater than the set size %d" % (hint_end, n))             # <<<<<<<<<<<<<<
 *         if n == 0:
 *             return _null_peak
 */
    __pyx_t_7 = __Pyx_PyInt_FromSize_t(__pyx_v_hint_end); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_7 = 0;
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_Hinted_start_index_d_cannot_be_g, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 802, __pyx_L1_error)

    /* "ms_peak_picker/_c/peak_set.pyx":801
 *         if hint_start > n :
 *             raise ValueError("Hinted start index %d cannot be greater than the set size %d" % (hint_start, n))
 *         if hint_end > n:             # <<<<<<<<<<<<<<
 *             raise ValueError("Hinted start index %d cannot be greater than the set size %d" % (hint_end, n))
 *         if n == 0:
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":803
 *         if hint_end > n:
 *             raise ValueError("Hinted start index %d cannot be greater than the set size %d" % (hint_end, n))
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return _null_peak
 *         i = double_binary_search_ppm_with_hint(self.mz_index, mz, tolerance, hint_end, hint_start)
 */
  __pyx_t_11 = ((__pyx_v_n == 0) != 0);
  if (__pyx_t_11) {

    /* "ms_peak_picker/_c/peak_set.pyx":804
 *             raise ValueError("Hinted start index %d cannot be greater than the set size %d" % (hint_end, n))
 *         if n == 0:
 *             return _null_peak             # <<<<<<<<<<<<<<
 *         i = double_binary_search_ppm_with_hint(self.mz_index, mz, tolerance, hint_end, hint_start)
 *         peak = self.getitem(i)
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak));
    __pyx_r = __pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":803
 *         if hint_end > n:
 *             raise ValueError("Hinted start index %d cannot be greater than the set size %d" % (hint_end, n))
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return _null_peak
 *         i = double_binary_search_ppm_with_hint(self.mz_index, mz, tolerance, hint_end, hint_start)
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":805
 *         if n == 0:
 *             return _null_peak
 *         i = double_binary_search_ppm_with_hint(self.mz_index, mz, tolerance, hint_end, hint_start)             # <<<<<<<<<<<<<<
 *         peak = self.getitem(i)
 *         if abs((peak.mz - mz) / mz) < tolerance:
 */
  __pyx_v_i = __pyx_f_14ms_peak_picker_2_c_8peak_set_double_binary_search_ppm_with_hint(__pyx_v_self->mz_index, __pyx_v_mz, __pyx_v_tolerance, __pyx_v_hint_end, __pyx_v_hint_start);

  /* "ms_peak_picker/_c/peak_set.pyx":806
 *             return _null_peak
 *         i = double_binary_search_ppm_with_hint(self.mz_index, mz, tolerance, hint_end, hint_start)
 *         peak = self.getitem(i)             # <<<<<<<<<<<<<<
 *         if abs((peak.mz - mz) / mz) < tolerance:
 *             return peak
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.getitem(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self), __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":807
 *         i = double_binary_search_ppm_with_hint(self.mz_index, mz, tolerance, hint_end, hint_start)
 *         peak = self.getitem(i)
 *         if abs((peak.mz - mz) / mz) < tolerance:             # <<<<<<<<<<<<<<
 *             return peak
 *         else:
 */
  __pyx_t_12 = (__pyx_v_peak->__pyx_base.mz - __pyx_v_mz);
  if (unlikely(__pyx_v_mz == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 807, __pyx_L1_error)
  }
  __pyx_t_11 = ((fabs((__pyx_t_12 / __pyx_v_mz)) < __pyx_v_tolerance) != 0);
  if (__pyx_t_11) {

    /* "ms_peak_picker/_c/peak_set.pyx":808
 *         peak = self.getitem(i)
 *         if abs((peak.mz - mz) / mz) < tolerance:
 *             return peak             # <<<<<<<<<<<<<<
 *         else:
 *             return _null_peak
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_peak));
    __pyx_r = __pyx_v_peak;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":807
 *         i = double_binary_search_ppm_with_hint(self.mz_index, mz, tolerance, hint_end, hint_start)
 *         peak = self.getitem(i)
 *         if abs((peak.mz - mz) / mz) < tolerance:             # <<<<<<<<<<<<<<
 *             return peak
 *         else:
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":810
 *             return peak
 *         else:
 *             return _null_peak             # <<<<<<<<<<<<<<
 * 
 *     def test_interval(self, double value):
 */
  /*else*/ {
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak));
    __pyx_r = __pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak;
    goto __pyx_L0;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":794
 *         return peak
 * 
 *     cpdef FittedPeak has_peak_hinted(self, double mz, double tolerance=2e-5, size_t hint_start=0, size_t hint_end=-1):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed.has_peak_hinted", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_13has_peak_hinted(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_12has_peak_hinted[] = "PeakSetIndexed.has_peak_hinted(self, double mz, double tolerance=2e-5, size_t hint_start=0, size_t hint_end=-1) -> FittedPeak";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_13has_peak_hinted(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_mz;
  double __pyx_v_tolerance;
  size_t __pyx_v_hint_start;
  size_t __pyx_v_hint_end;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("has_peak_hinted (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_tolerance,&__pyx_n_s_hint_start,&__pyx_n_s_hint_end,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tolerance);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hint_start);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hint_end);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "has_peak_hinted") < 0)) __PYX_ERR(0, 794, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_mz = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 794, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 794, __pyx_L3_error)
    } else {
      __pyx_v_tolerance = ((double)2e-5);
    }
    if (values[2]) {
      __pyx_v_hint_start = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_hint_start == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 794, __pyx_L3_error)
    } else {
      __pyx_v_hint_start = ((size_t)0);
    }
    if (values[3]) {
      __pyx_v_hint_end = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_hint_end == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 794, __pyx_L3_error)
    } else {
      __pyx_v_hint_end = ((size_t)-1L);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("has_peak_hinted", 0, 1, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 794, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed.has_peak_hinted", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_12has_peak_hinted(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self), __pyx_v_mz, __pyx_v_tolerance, __pyx_v_hint_start, __pyx_v_hint_end);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_12has_peak_hinted(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, double __pyx_v_mz, double __pyx_v_tolerance, size_t __pyx_v_hint_start, size_t __pyx_v_hint_end) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("has_peak_hinted", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 3;
  __pyx_t_2.tolerance = __pyx_v_tolerance;
  __pyx_t_2.hint_start = __pyx_v_hint_start;
  __pyx_t_2.hint_end = __pyx_v_hint_end;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed->has_peak_hinted(__pyx_v_self, __pyx_v_mz, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed.has_peak_hinted", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":812
 *             return _null_peak
 * 
 *     def test_interval(self, double value):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int status
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_15test_interval(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_14test_interval[] = "PeakSetIndexed.test_interval(self, double value)";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_15test_interval(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("test_interval (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 812, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed.test_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_14test_interval(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self), ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_14test_interval(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, double __pyx_v_value) {
  CYTHON_UNUSED int __pyx_v_status;
  size_t __pyx_v_start;
  size_t __pyx_v_end;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("test_interval", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":817
 *             size_t start, end
 *             size_t i
 *         if self.interval_index == NULL:             # <<<<<<<<<<<<<<
 *             return None, None, None
 *         status = find_search_interval(self.interval_index, value, &start, &end)
 */
  __pyx_t_1 = ((__pyx_v_self->interval_index == NULL) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":818
 *             size_t i
 *         if self.interval_index == NULL:
 *             return None, None, None             # <<<<<<<<<<<<<<
 *         status = find_search_interval(self.interval_index, value, &start, &end)
 *         return start, end, interpolate_index(self.interval_index, value)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_tuple_);
    __pyx_r = __pyx_tuple_;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":817
 *             size_t start, end
 *             size_t i
 *         if self.interval_index == NULL:             # <<<<<<<<<<<<<<
 *             return None, None, None
 *         status = find_search_interval(self.interval_index, value, &start, &end)
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":819
 *         if self.interval_index == NULL:
 *             return None, None, None
 *         status = find_search_interval(self.interval_index, value, &start, &end)             # <<<<<<<<<<<<<<
 *         return start, end, interpolate_index(self.interval_index, value)
 * 
 */
  __pyx_v_status = __pyx_f_14ms_peak_picker_2_c_8peak_set_find_search_interval(__pyx_v_self->interval_index, __pyx_v_value, (&__pyx_v_start), (&__pyx_v_end));

  /* "ms_peak_picker/_c/peak_set.pyx":820
 *             return None, None, None
 *         status = find_search_interval(self.interval_index, value, &start, &end)
 *         return start, end, interpolate_index(self.interval_index, value)             # <<<<<<<<<<<<<<
 * 
 *     def find_interval_for(self, double value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_start); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_f_14ms_peak_picker_2_c_8peak_set_interpolate_index(__pyx_v_self->interval_index, __pyx_v_value)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":812
 *             return _null_peak
 * 
 *     def test_interval(self, double value):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int status
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed.test_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":822
 *         return start, end, interpolate_index(self.interval_index, value)
 * 
 *     def find_interval_for(self, double value):             # <<<<<<<<<<<<<<
 *         if self.interval_index == NULL:
 *             return None
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_17find_interval_for(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_16find_interval_for[] = "PeakSetIndexed.find_interval_for(self, double value)";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_17find_interval_for(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_interval_for (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 822, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed.find_interval_for", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_16find_interval_for(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self), ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_16find_interval_for(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, double __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_interval_for", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":823
 * 
 *     def find_interval_for(self, double value):
 *         if self.interval_index == NULL:             # <<<<<<<<<<<<<<
 *             return None
 *         return interpolate_index(self.interval_index, value)
 */
  __pyx_t_1 = ((__pyx_v_self->interval_index == NULL) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":824
 *     def find_interval_for(self, double value):
 *         if self.interval_index == NULL:
 *             return None             # <<<<<<<<<<<<<<
 *         return interpolate_index(self.interval_index, value)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":823
 * 
 *     def find_interval_for(self, double value):
 *         if self.interval_index == NULL:             # <<<<<<<<<<<<<<
 *             return None
 *         return interpolate_index(self.interval_index, value)
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":825
 *         if self.interval_index == NULL:
 *             return None
 *         return interpolate_index(self.interval_index, value)             # <<<<<<<<<<<<<<
 * 
 *     def check_interval(self, size_t i):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_f_14ms_peak_picker_2_c_8peak_set_interpolate_index(__pyx_v_self->interval_index, __pyx_v_value)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":822
 *         return start, end, interpolate_index(self.interval_index, value)
 * 
 *     def find_interval_for(self, double value):             # <<<<<<<<<<<<<<
 *         if self.interval_index == NULL:
 *             return None
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed.find_interval_for", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":827
 *         return interpolate_index(self.interval_index, value)
 * 
 *     def check_interval(self, size_t i):             # <<<<<<<<<<<<<<
 *         cdef:
 *             index_cell cell
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_19check_interval(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_18check_interval[] = "PeakSetIndexed.check_interval(self, size_t i)";
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_19check_interval(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  size_t __pyx_v_i;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("check_interval (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_size_t(__pyx_arg_i); if (unlikely((__pyx_v_i == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 827, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed.check_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_18check_interval(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)__pyx_v_self), ((size_t)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_18check_interval(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_v_self, size_t __pyx_v_i) {
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell __pyx_v_cell;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_interval", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":830
 *         cdef:
 *             index_cell cell
 *         if self.interval_index == NULL:             # <<<<<<<<<<<<<<
 *             return None
 *         cell = self.interval_index.index[i]
 */
  __pyx_t_1 = ((__pyx_v_self->interval_index == NULL) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":831
 *             index_cell cell
 *         if self.interval_index == NULL:
 *             return None             # <<<<<<<<<<<<<<
 *         cell = self.interval_index.index[i]
 *         return cell
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":830
 *         cdef:
 *             index_cell cell
 *         if self.interval_index == NULL:             # <<<<<<<<<<<<<<
 *             return None
 *         cell = self.interval_index.index[i]
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":832
 *         if self.interval_index == NULL:
 *             return None
 *         cell = self.interval_index.index[i]             # <<<<<<<<<<<<<<
 *         return cell
 * 
 */
  __pyx_v_cell = (__pyx_v_self->interval_index->index[__pyx_v_i]);

  /* "ms_peak_picker/_c/peak_set.pyx":833
 *             return None
 *         cell = self.interval_index.index[i]
 *         return cell             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_convert__to_py_struct____pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell(__pyx_v_cell); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":827
 *         return interpolate_index(self.interval_index, value)
 * 
 *     def check_interval(self, size_t i):             # <<<<<<<<<<<<<<
 *         cdef:
 *             index_cell cell
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.PeakSetIndexed.check_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":836
 * 
 * 
 * cdef double* build_linear_spaced_array(double low, double high, size_t n):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* array
 */

static double *__pyx_f_14ms_peak_picker_2_c_8peak_set_build_linear_spaced_array(double __pyx_v_low, double __pyx_v_high, size_t __pyx_v_n) {
  double *__pyx_v_array;
  double __pyx_v_delta;
  size_t __pyx_v_i;
  double *__pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("build_linear_spaced_array", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":842
 *         size_t i
 * 
 *     delta = (high - low) / (n - 1)             # <<<<<<<<<<<<<<
 * 
 *     array = <double*>malloc(sizeof(double) * n)
 */
  __pyx_t_1 = (__pyx_v_high - __pyx_v_low);
  __pyx_t_2 = (__pyx_v_n - 1);
  if (unlikely(__pyx_t_2 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 842, __pyx_L1_error)
  }
  __pyx_v_delta = (__pyx_t_1 / __pyx_t_2);

  /* "ms_peak_picker/_c/peak_set.pyx":844
 *     delta = (high - low) / (n - 1)
 * 
 *     array = <double*>malloc(sizeof(double) * n)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(n):
 */
  __pyx_v_array = ((double *)malloc(((sizeof(double)) * __pyx_v_n)));

  /* "ms_peak_picker/_c/peak_set.pyx":846
 *     array = <double*>malloc(sizeof(double) * n)
 * 
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         array[i] = low + i * delta
 *     return array
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "ms_peak_picker/_c/peak_set.pyx":847
 * 
 *     for i in range(n):
 *         array[i] = low + i * delta             # <<<<<<<<<<<<<<
 *     return array
 * 
 */
    (__pyx_v_array[__pyx_v_i]) = (__pyx_v_low + (__pyx_v_i * __pyx_v_delta));
  }

  /* "ms_peak_picker/_c/peak_set.pyx":848
 *     for i in range(n):
 *         array[i] = low + i * delta
 *     return array             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_array;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":836
 * 
 * 
 * cdef double* build_linear_spaced_array(double low, double high, size_t n):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* array
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("ms_peak_picker._c.peak_set.build_linear_spaced_array", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":851
 * 
 * 
 * cdef void free_index_list(index_list* index):             # <<<<<<<<<<<<<<
 *     free(index.index)
 *     free(index)
 */

static void __pyx_f_14ms_peak_picker_2_c_8peak_set_free_index_list(struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *__pyx_v_index) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("free_index_list", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":852
 * 
 * cdef void free_index_list(index_list* index):
 *     free(index.index)             # <<<<<<<<<<<<<<
 *     free(index)
 * 
 */
  free(__pyx_v_index->index);

  /* "ms_peak_picker/_c/peak_set.pyx":853
 * cdef void free_index_list(index_list* index):
 *     free(index.index)
 *     free(index)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  free(__pyx_v_index);

  /* "ms_peak_picker/_c/peak_set.pyx":851
 * 
 * 
 * cdef void free_index_list(index_list* index):             # <<<<<<<<<<<<<<
 *     free(index.index)
 *     free(index)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "ms_peak_picker/_c/peak_set.pyx":856
 * 
 * 
 * cdef int check_index(index_list* index) nogil:             # <<<<<<<<<<<<<<
 *     if index.size == 0:
 *         return 1
 */

static int __pyx_f_14ms_peak_picker_2_c_8peak_set_check_index(struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *__pyx_v_index) {
  int __pyx_r;
  int __pyx_t_1;

  /* "ms_peak_picker/_c/peak_set.pyx":857
 * 
 * cdef int check_index(index_list* index) nogil:
 *     if index.size == 0:             # <<<<<<<<<<<<<<
 *         return 1
 *     elif (index.high - index.low) == 0:
 */
  __pyx_t_1 = ((__pyx_v_index->size == 0) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":858
 * cdef int check_index(index_list* index) nogil:
 *     if index.size == 0:
 *         return 1             # <<<<<<<<<<<<<<
 *     elif (index.high - index.low) == 0:
 *         return 2
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":857
 * 
 * cdef int check_index(index_list* index) nogil:
 *     if index.size == 0:             # <<<<<<<<<<<<<<
 *         return 1
 *     elif (index.high - index.low) == 0:
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":859
 *     if index.size == 0:
 *         return 1
 *     elif (index.high - index.low) == 0:             # <<<<<<<<<<<<<<
 *         return 2
 *     else:
 */
  __pyx_t_1 = (((__pyx_v_index->high - __pyx_v_index->low) == 0.0) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":860
 *         return 1
 *     elif (index.high - index.low) == 0:
 *         return 2             # <<<<<<<<<<<<<<
 *     else:
 *         return 0
 */
    __pyx_r = 2;
    goto __pyx_L0;

    /* "ms_peak_picker/_c/peak_set.pyx":859
 *     if index.size == 0:
 *         return 1
 *     elif (index.high - index.low) == 0:             # <<<<<<<<<<<<<<
 *         return 2
 *     else:
 */
  }

  /* "ms_peak_picker/_c/peak_set.pyx":862
 *         return 2
 *     else:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":856
 * 
 * 
 * cdef int check_index(index_list* index) nogil:             # <<<<<<<<<<<<<<
 *     if index.size == 0:
 *         return 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":865
 * 
 * 
 * cdef int build_interval_index(PeakSet peaks, index_list* index, size_t index_size):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* linear_spacing
 */

static int __pyx_f_14ms_peak_picker_2_c_8peak_set_build_interval_index(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks, struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *__pyx_v_index, size_t __pyx_v_index_size) {
  double *__pyx_v_linear_spacing;
  double __pyx_v_current_value;
  double __pyx_v_err;
  double __pyx_v_next_value;
  size_t __pyx_v_i;
  size_t __pyx_v_start_i;
  size_t __pyx_v_end_i;
  size_t __pyx_v_index_i;
  size_t __pyx_v_peaks_size;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  double __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  long __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("build_interval_index", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":871
 *         size_t i, start_i, end_i, index_i, peaks_size
 *         FittedPeak peak
 *     peaks_size = peaks.get_size()             # <<<<<<<<<<<<<<
 *     if peaks_size > 0:
 *         index.low = peaks.getitem(0).mz
 */
  __pyx_v_peaks_size = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->get_size(__pyx_v_peaks);

  /* "ms_peak_picker/_c/peak_set.pyx":872
 *         FittedPeak peak
 *     peaks_size = peaks.get_size()
 *     if peaks_size > 0:             # <<<<<<<<<<<<<<
 *         index.low = peaks.getitem(0).mz
 *         index.high = peaks.getitem(peaks_size - 1).mz
 */
  __pyx_t_1 = ((__pyx_v_peaks_size > 0) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":873
 *     peaks_size = peaks.get_size()
 *     if peaks_size > 0:
 *         index.low = peaks.getitem(0).mz             # <<<<<<<<<<<<<<
 *         index.high = peaks.getitem(peaks_size - 1).mz
 *     else:
 */
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2)->__pyx_base.mz;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_index->low = __pyx_t_3;

    /* "ms_peak_picker/_c/peak_set.pyx":874
 *     if peaks_size > 0:
 *         index.low = peaks.getitem(0).mz
 *         index.high = peaks.getitem(peaks_size - 1).mz             # <<<<<<<<<<<<<<
 *     else:
 *         index.low = 0
 */
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, (__pyx_v_peaks_size - 1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2)->__pyx_base.mz;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_index->high = __pyx_t_3;

    /* "ms_peak_picker/_c/peak_set.pyx":872
 *         FittedPeak peak
 *     peaks_size = peaks.get_size()
 *     if peaks_size > 0:             # <<<<<<<<<<<<<<
 *         index.low = peaks.getitem(0).mz
 *         index.high = peaks.getitem(peaks_size - 1).mz
 */
    goto __pyx_L3;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":876
 *         index.high = peaks.getitem(peaks_size - 1).mz
 *     else:
 *         index.low = 0             # <<<<<<<<<<<<<<
 *         index.high = 1
 *     index.size = index_size
 */
  /*else*/ {
    __pyx_v_index->low = 0.0;

    /* "ms_peak_picker/_c/peak_set.pyx":877
 *     else:
 *         index.low = 0
 *         index.high = 1             # <<<<<<<<<<<<<<
 *     index.size = index_size
 *     linear_spacing = build_linear_spaced_array(
 */
    __pyx_v_index->high = 1.0;
  }
  __pyx_L3:;

  /* "ms_peak_picker/_c/peak_set.pyx":878
 *         index.low = 0
 *         index.high = 1
 *     index.size = index_size             # <<<<<<<<<<<<<<
 *     linear_spacing = build_linear_spaced_array(
 *         index.low,
 */
  __pyx_v_index->size = __pyx_v_index_size;

  /* "ms_peak_picker/_c/peak_set.pyx":879
 *         index.high = 1
 *     index.size = index_size
 *     linear_spacing = build_linear_spaced_array(             # <<<<<<<<<<<<<<
 *         index.low,
 *         index.high,
 */
  __pyx_v_linear_spacing = __pyx_f_14ms_peak_picker_2_c_8peak_set_build_linear_spaced_array(__pyx_v_index->low, __pyx_v_index->high, __pyx_v_index_size);

  /* "ms_peak_picker/_c/peak_set.pyx":884
 *         index_size)
 * 
 *     index.index = <index_cell*>malloc(sizeof(index_cell) * index_size)             # <<<<<<<<<<<<<<
 * 
 *     for index_i in range(index_size):
 */
  __pyx_v_index->index = ((struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell *)malloc(((sizeof(struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell)) * __pyx_v_index_size)));

  /* "ms_peak_picker/_c/peak_set.pyx":886
 *     index.index = <index_cell*>malloc(sizeof(index_cell) * index_size)
 * 
 *     for index_i in range(index_size):             # <<<<<<<<<<<<<<
 *         current_value = linear_spacing[index_i]
 *         peak = peaks._get_nearest_peak(current_value, &err)
 */
  __pyx_t_4 = __pyx_v_index_size;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_index_i = __pyx_t_6;

    /* "ms_peak_picker/_c/peak_set.pyx":887
 * 
 *     for index_i in range(index_size):
 *         current_value = linear_spacing[index_i]             # <<<<<<<<<<<<<<
 *         peak = peaks._get_nearest_peak(current_value, &err)
 *         if peaks_size > 0:
 */
    __pyx_v_current_value = (__pyx_v_linear_spacing[__pyx_v_index_i]);

    /* "ms_peak_picker/_c/peak_set.pyx":888
 *     for index_i in range(index_size):
 *         current_value = linear_spacing[index_i]
 *         peak = peaks._get_nearest_peak(current_value, &err)             # <<<<<<<<<<<<<<
 *         if peaks_size > 0:
 *             start_i = peak.peak_count
 */
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->_get_nearest_peak(__pyx_v_peaks, __pyx_v_current_value, (&__pyx_v_err))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 888, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "ms_peak_picker/_c/peak_set.pyx":889
 *         current_value = linear_spacing[index_i]
 *         peak = peaks._get_nearest_peak(current_value, &err)
 *         if peaks_size > 0:             # <<<<<<<<<<<<<<
 *             start_i = peak.peak_count
 *             if index_i > 0:
 */
    __pyx_t_1 = ((__pyx_v_peaks_size > 0) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":890
 *         peak = peaks._get_nearest_peak(current_value, &err)
 *         if peaks_size > 0:
 *             start_i = peak.peak_count             # <<<<<<<<<<<<<<
 *             if index_i > 0:
 *                 start_i = index.index[index_i - 1].end - 1
 */
      __pyx_t_7 = __pyx_v_peak->peak_count;
      __pyx_v_start_i = __pyx_t_7;

      /* "ms_peak_picker/_c/peak_set.pyx":891
 *         if peaks_size > 0:
 *             start_i = peak.peak_count
 *             if index_i > 0:             # <<<<<<<<<<<<<<
 *                 start_i = index.index[index_i - 1].end - 1
 *             if index_i == index_size - 1:
 */
      __pyx_t_1 = ((__pyx_v_index_i > 0) != 0);
      if (__pyx_t_1) {

        /* "ms_peak_picker/_c/peak_set.pyx":892
 *             start_i = peak.peak_count
 *             if index_i > 0:
 *                 start_i = index.index[index_i - 1].end - 1             # <<<<<<<<<<<<<<
 *             if index_i == index_size - 1:
 *                 end_i = peaks_size - 1
 */
        __pyx_v_start_i = ((__pyx_v_index->index[(__pyx_v_index_i - 1)]).end - 1);

        /* "ms_peak_picker/_c/peak_set.pyx":891
 *         if peaks_size > 0:
 *             start_i = peak.peak_count
 *             if index_i > 0:             # <<<<<<<<<<<<<<
 *                 start_i = index.index[index_i - 1].end - 1
 *             if index_i == index_size - 1:
 */
      }

      /* "ms_peak_picker/_c/peak_set.pyx":893
 *             if index_i > 0:
 *                 start_i = index.index[index_i - 1].end - 1
 *             if index_i == index_size - 1:             # <<<<<<<<<<<<<<
 *                 end_i = peaks_size - 1
 *             else:
 */
      __pyx_t_1 = ((__pyx_v_index_i == (__pyx_v_index_size - 1)) != 0);
      if (__pyx_t_1) {

        /* "ms_peak_picker/_c/peak_set.pyx":894
 *                 start_i = index.index[index_i - 1].end - 1
 *             if index_i == index_size - 1:
 *                 end_i = peaks_size - 1             # <<<<<<<<<<<<<<
 *             else:
 *                 next_value = linear_spacing[index_i + 1]
 */
        __pyx_v_end_i = (__pyx_v_peaks_size - 1);

        /* "ms_peak_picker/_c/peak_set.pyx":893
 *             if index_i > 0:
 *                 start_i = index.index[index_i - 1].end - 1
 *             if index_i == index_size - 1:             # <<<<<<<<<<<<<<
 *                 end_i = peaks_size - 1
 *             else:
 */
        goto __pyx_L8;
      }

      /* "ms_peak_picker/_c/peak_set.pyx":896
 *                 end_i = peaks_size - 1
 *             else:
 *                 next_value = linear_spacing[index_i + 1]             # <<<<<<<<<<<<<<
 *                 i = peak.peak_count
 *                 while i < peaks_size:
 */
      /*else*/ {
        __pyx_v_next_value = (__pyx_v_linear_spacing[(__pyx_v_index_i + 1)]);

        /* "ms_peak_picker/_c/peak_set.pyx":897
 *             else:
 *                 next_value = linear_spacing[index_i + 1]
 *                 i = peak.peak_count             # <<<<<<<<<<<<<<
 *                 while i < peaks_size:
 *                     peak = peaks.getitem(i)
 */
        __pyx_t_7 = __pyx_v_peak->peak_count;
        __pyx_v_i = __pyx_t_7;

        /* "ms_peak_picker/_c/peak_set.pyx":898
 *                 next_value = linear_spacing[index_i + 1]
 *                 i = peak.peak_count
 *                 while i < peaks_size:             # <<<<<<<<<<<<<<
 *                     peak = peaks.getitem(i)
 *                     if abs(current_value - peak.mz) > abs(next_value - peak.mz):
 */
        while (1) {
          __pyx_t_1 = ((__pyx_v_i < __pyx_v_peaks_size) != 0);
          if (!__pyx_t_1) break;

          /* "ms_peak_picker/_c/peak_set.pyx":899
 *                 i = peak.peak_count
 *                 while i < peaks_size:
 *                     peak = peaks.getitem(i)             # <<<<<<<<<<<<<<
 *                     if abs(current_value - peak.mz) > abs(next_value - peak.mz):
 *                         break
 */
          __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_i)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 899, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2));
          __pyx_t_2 = 0;

          /* "ms_peak_picker/_c/peak_set.pyx":900
 *                 while i < peaks_size:
 *                     peak = peaks.getitem(i)
 *                     if abs(current_value - peak.mz) > abs(next_value - peak.mz):             # <<<<<<<<<<<<<<
 *                         break
 *                     i += 1
 */
          __pyx_t_1 = ((fabs((__pyx_v_current_value - __pyx_v_peak->__pyx_base.mz)) > fabs((__pyx_v_next_value - __pyx_v_peak->__pyx_base.mz))) != 0);
          if (__pyx_t_1) {

            /* "ms_peak_picker/_c/peak_set.pyx":901
 *                     peak = peaks.getitem(i)
 *                     if abs(current_value - peak.mz) > abs(next_value - peak.mz):
 *                         break             # <<<<<<<<<<<<<<
 *                     i += 1
 *                 end_i = i
 */
            goto __pyx_L10_break;

            /* "ms_peak_picker/_c/peak_set.pyx":900
 *                 while i < peaks_size:
 *                     peak = peaks.getitem(i)
 *                     if abs(current_value - peak.mz) > abs(next_value - peak.mz):             # <<<<<<<<<<<<<<
 *                         break
 *                     i += 1
 */
          }

          /* "ms_peak_picker/_c/peak_set.pyx":902
 *                     if abs(current_value - peak.mz) > abs(next_value - peak.mz):
 *                         break
 *                     i += 1             # <<<<<<<<<<<<<<
 *                 end_i = i
 *         else:
 */
          __pyx_v_i = (__pyx_v_i + 1);
        }
        __pyx_L10_break:;

        /* "ms_peak_picker/_c/peak_set.pyx":903
 *                         break
 *                     i += 1
 *                 end_i = i             # <<<<<<<<<<<<<<
 *         else:
 *             start_i = 0
 */
        __pyx_v_end_i = __pyx_v_i;
      }
      __pyx_L8:;

      /* "ms_peak_picker/_c/peak_set.pyx":889
 *         current_value = linear_spacing[index_i]
 *         peak = peaks._get_nearest_peak(current_value, &err)
 *         if peaks_size > 0:             # <<<<<<<<<<<<<<
 *             start_i = peak.peak_count
 *             if index_i > 0:
 */
      goto __pyx_L6;
    }

    /* "ms_peak_picker/_c/peak_set.pyx":905
 *                 end_i = i
 *         else:
 *             start_i = 0             # <<<<<<<<<<<<<<
 *             end_i = 0
 *         index.index[index_i].center_value = current_value
 */
    /*else*/ {
      __pyx_v_start_i = 0;

      /* "ms_peak_picker/_c/peak_set.pyx":906
 *         else:
 *             start_i = 0
 *             end_i = 0             # <<<<<<<<<<<<<<
 *         index.index[index_i].center_value = current_value
 *         index.index[index_i].start = start_i
 */
      __pyx_v_end_i = 0;
    }
    __pyx_L6:;

    /* "ms_peak_picker/_c/peak_set.pyx":907
 *             start_i = 0
 *             end_i = 0
 *         index.index[index_i].center_value = current_value             # <<<<<<<<<<<<<<
 *         index.index[index_i].start = start_i
 *         index.index[index_i].end = end_i
 */
    (__pyx_v_index->index[__pyx_v_index_i]).center_value = __pyx_v_current_value;

    /* "ms_peak_picker/_c/peak_set.pyx":908
 *             end_i = 0
 *         index.index[index_i].center_value = current_value
 *         index.index[index_i].start = start_i             # <<<<<<<<<<<<<<
 *         index.index[index_i].end = end_i
 * 
 */
    (__pyx_v_index->index[__pyx_v_index_i]).start = __pyx_v_start_i;

    /* "ms_peak_picker/_c/peak_set.pyx":909
 *         index.index[index_i].center_value = current_value
 *         index.index[index_i].start = start_i
 *         index.index[index_i].end = end_i             # <<<<<<<<<<<<<<
 * 
 *     free(linear_spacing)
 */
    (__pyx_v_index->index[__pyx_v_index_i]).end = __pyx_v_end_i;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":911
 *         index.index[index_i].end = end_i
 * 
 *     free(linear_spacing)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  free(__pyx_v_linear_spacing);

  /* "ms_peak_picker/_c/peak_set.pyx":912
 * 
 *     free(linear_spacing)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":865
 * 
 * 
 * cdef int build_interval_index(PeakSet peaks, index_list* index, size_t index_size):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* linear_spacing
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("ms_peak_picker._c.peak_set.build_interval_index", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":915
 * 
 * 
 * cdef size_t interpolate_index(index_list* index, double value):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double v
 */

static size_t __pyx_f_14ms_peak_picker_2_c_8peak_set_interpolate_index(struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *__pyx_v_index, double __pyx_v_value) {
  double __pyx_v_v;
  size_t __pyx_v_i;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  double __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("interpolate_index", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":919
 *         double v
 *         size_t i
 *     v = (((value - index.low) / (index.high - index.low)) * index.size)             # <<<<<<<<<<<<<<
 *     i = <size_t>v
 *     return i
 */
  __pyx_t_1 = (__pyx_v_value - __pyx_v_index->low);
  __pyx_t_2 = (__pyx_v_index->high - __pyx_v_index->low);
  if (unlikely(__pyx_t_2 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 919, __pyx_L1_error)
  }
  __pyx_v_v = ((__pyx_t_1 / __pyx_t_2) * __pyx_v_index->size);

  /* "ms_peak_picker/_c/peak_set.pyx":920
 *         size_t i
 *     v = (((value - index.low) / (index.high - index.low)) * index.size)
 *     i = <size_t>v             # <<<<<<<<<<<<<<
 *     return i
 * 
 */
  __pyx_v_i = ((size_t)__pyx_v_v);

  /* "ms_peak_picker/_c/peak_set.pyx":921
 *     v = (((value - index.low) / (index.high - index.low)) * index.size)
 *     i = <size_t>v
 *     return i             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_i;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":915
 * 
 * 
 * cdef size_t interpolate_index(index_list* index, double value):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double v
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("ms_peak_picker._c.peak_set.interpolate_index", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ms_peak_picker/_c/peak_set.pyx":924
 * 
 * 
 * cdef int find_search_interval(index_list* index, double value, size_t* start, size_t* end):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static int __pyx_f_14ms_peak_picker_2_c_8peak_set_find_search_interval(struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *__pyx_v_index, double __pyx_v_value, size_t *__pyx_v_start, size_t *__pyx_v_end) {
  size_t __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  size_t __pyx_t_2;
  __Pyx_RefNannySetupContext("find_search_interval", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":927
 *     cdef:
 *         size_t i
 *     if value > index.high:             # <<<<<<<<<<<<<<
 *         i = index.size - 1
 *     elif value < index.low:
 */
  __pyx_t_1 = ((__pyx_v_value > __pyx_v_index->high) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":928
 *         size_t i
 *     if value > index.high:
 *         i = index.size - 1             # <<<<<<<<<<<<<<
 *     elif value < index.low:
 *         i = 0
 */
    __pyx_v_i = (__pyx_v_index->size - 1);

    /* "ms_peak_picker/_c/peak_set.pyx":927
 *     cdef:
 *         size_t i
 *     if value > index.high:             # <<<<<<<<<<<<<<
 *         i = index.size - 1
 *     elif value < index.low:
 */
    goto __pyx_L3;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":929
 *     if value > index.high:
 *         i = index.size - 1
 *     elif value < index.low:             # <<<<<<<<<<<<<<
 *         i = 0
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_value < __pyx_v_index->low) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":930
 *         i = index.size - 1
 *     elif value < index.low:
 *         i = 0             # <<<<<<<<<<<<<<
 *     else:
 *         i = interpolate_index(index, value)
 */
    __pyx_v_i = 0;

    /* "ms_peak_picker/_c/peak_set.pyx":929
 *     if value > index.high:
 *         i = index.size - 1
 *     elif value < index.low:             # <<<<<<<<<<<<<<
 *         i = 0
 *     else:
 */
    goto __pyx_L3;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":932
 *         i = 0
 *     else:
 *         i = interpolate_index(index, value)             # <<<<<<<<<<<<<<
 *     if i > 0:
 *         if i < index.size:
 */
  /*else*/ {
    __pyx_v_i = __pyx_f_14ms_peak_picker_2_c_8peak_set_interpolate_index(__pyx_v_index, __pyx_v_value);
  }
  __pyx_L3:;

  /* "ms_peak_picker/_c/peak_set.pyx":933
 *     else:
 *         i = interpolate_index(index, value)
 *     if i > 0:             # <<<<<<<<<<<<<<
 *         if i < index.size:
 *             start[0] = index.index[i - 1].start
 */
  __pyx_t_1 = ((__pyx_v_i > 0) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":934
 *         i = interpolate_index(index, value)
 *     if i > 0:
 *         if i < index.size:             # <<<<<<<<<<<<<<
 *             start[0] = index.index[i - 1].start
 *         else:
 */
    __pyx_t_1 = ((__pyx_v_i < __pyx_v_index->size) != 0);
    if (__pyx_t_1) {

      /* "ms_peak_picker/_c/peak_set.pyx":935
 *     if i > 0:
 *         if i < index.size:
 *             start[0] = index.index[i - 1].start             # <<<<<<<<<<<<<<
 *         else:
 *             # if we're at index.size or greater, act as if we're at the last
 */
      __pyx_t_2 = (__pyx_v_index->index[(__pyx_v_i - 1)]).start;
      (__pyx_v_start[0]) = __pyx_t_2;

      /* "ms_peak_picker/_c/peak_set.pyx":934
 *         i = interpolate_index(index, value)
 *     if i > 0:
 *         if i < index.size:             # <<<<<<<<<<<<<<
 *             start[0] = index.index[i - 1].start
 *         else:
 */
      goto __pyx_L5;
    }

    /* "ms_peak_picker/_c/peak_set.pyx":939
 *             # if we're at index.size or greater, act as if we're at the last
 *             # cell of the index
 *             start[0] = index.index[index.size - 2].start             # <<<<<<<<<<<<<<
 *     else:
 *         # if somehow the index were negative, this could end badly.
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_index->index[(__pyx_v_index->size - 2)]).start;
      (__pyx_v_start[0]) = __pyx_t_2;
    }
    __pyx_L5:;

    /* "ms_peak_picker/_c/peak_set.pyx":933
 *     else:
 *         i = interpolate_index(index, value)
 *     if i > 0:             # <<<<<<<<<<<<<<
 *         if i < index.size:
 *             start[0] = index.index[i - 1].start
 */
    goto __pyx_L4;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":942
 *     else:
 *         # if somehow the index were negative, this could end badly.
 *         start[0] = index.index[i].start             # <<<<<<<<<<<<<<
 *     if i >= (index.size - 1):
 *         end[0] = index.index[index.size - 1].end + 1
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_index->index[__pyx_v_i]).start;
    (__pyx_v_start[0]) = __pyx_t_2;
  }
  __pyx_L4:;

  /* "ms_peak_picker/_c/peak_set.pyx":943
 *         # if somehow the index were negative, this could end badly.
 *         start[0] = index.index[i].start
 *     if i >= (index.size - 1):             # <<<<<<<<<<<<<<
 *         end[0] = index.index[index.size - 1].end + 1
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_i >= (__pyx_v_index->size - 1)) != 0);
  if (__pyx_t_1) {

    /* "ms_peak_picker/_c/peak_set.pyx":944
 *         start[0] = index.index[i].start
 *     if i >= (index.size - 1):
 *         end[0] = index.index[index.size - 1].end + 1             # <<<<<<<<<<<<<<
 *     else:
 *         end[0] = index.index[i + 1].end + 1
 */
    (__pyx_v_end[0]) = ((__pyx_v_index->index[(__pyx_v_index->size - 1)]).end + 1);

    /* "ms_peak_picker/_c/peak_set.pyx":943
 *         # if somehow the index were negative, this could end badly.
 *         start[0] = index.index[i].start
 *     if i >= (index.size - 1):             # <<<<<<<<<<<<<<
 *         end[0] = index.index[index.size - 1].end + 1
 *     else:
 */
    goto __pyx_L6;
  }

  /* "ms_peak_picker/_c/peak_set.pyx":946
 *         end[0] = index.index[index.size - 1].end + 1
 *     else:
 *         end[0] = index.index[i + 1].end + 1             # <<<<<<<<<<<<<<
 *     return 0
 */
  /*else*/ {
    (__pyx_v_end[0]) = ((__pyx_v_index->index[(__pyx_v_i + 1)]).end + 1);
  }
  __pyx_L6:;

  /* "ms_peak_picker/_c/peak_set.pyx":947
 *     else:
 *         end[0] = index.index[i + 1].end + 1
 *     return 0             # <<<<<<<<<<<<<<
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "ms_peak_picker/_c/peak_set.pyx":924
 * 
 * 
 * cdef int find_search_interval(index_list* index, double value, size_t* start, size_t* end):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_PeakBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_3__pyx_unpickle_PeakBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14ms_peak_picker_2_c_8peak_set_2__pyx_unpickle_PeakBase[] = "__pyx_unpickle_PeakBase(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_14ms_peak_picker_2_c_8peak_set_3__pyx_unpickle_PeakBase = {"__pyx_unpickle_PeakBase", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14ms_peak_picker_2_c_8peak_set_3__pyx_unpickle_PeakBase, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_2__pyx_unpickle_PeakBase};
static PyObject *__pyx_pw_14ms_peak_picker_2_c_8peak_set_3__pyx_unpickle_PeakBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_PeakBase (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PeakBase", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PeakBase", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_PeakBase") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PeakBase", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.__pyx_unpickle_PeakBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14ms_peak_picker_2_c_8peak_set_2__pyx_unpickle_PeakBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14ms_peak_picker_2_c_8peak_set_2__pyx_unpickle_PeakBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PeakBase", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x4f329c4, 0x9e542aa, 0x4c1f989):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x4f329c4, 0x9e542aa, 0x4c1f989) = (area, intensity, mz))" % __pyx_checksum)
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__2, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x4f329c4, 0x9e542aa, 0x4c1f989):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x4f329c4, 0x9e542aa, 0x4c1f989) = (area, intensity, mz))" % __pyx_checksum)
 *     __pyx_result = PeakBase.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x4f329c4, 0x9e542aa, 0x4c1f989):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x4f329c4, 0x9e542aa, 0x4c1f989) = (area, intensity, mz))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = PeakBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x4f329c4, 0x9e542aa, 0x4c1f989):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x4f329c4, 0x9e542aa, 0x4c1f989) = (area, intensity, mz))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x4f329c4, 0x9e542aa, 0x4c1f989) = (area, intensity, mz))" % __pyx_checksum)
 *     __pyx_result = PeakBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PeakBase__set_state(<PeakBase> __pyx_result, __pyx_state)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x4f329c4, 0x9e542aa, 0x4c1f989) = (area, intensity, mz))" % __pyx_checksum)
 *     __pyx_result = PeakBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PeakBase__set_state(<PeakBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = PeakBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PeakBase__set_state(<PeakBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_PeakBase__set_state(PeakBase __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_4 = __pyx_f_14ms_peak_picker_2_c_8peak_set___pyx_unpickle_PeakBase__set_state(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x4f329c4, 0x9e542aa, 0x4c1f989) = (area, intensity, mz))" % __pyx_checksum)
 *     __pyx_result = PeakBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PeakBase__set_state(<PeakBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PeakBase__set_state(<PeakBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_PeakBase__set_state(PeakBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.area = __pyx_state[0]; __pyx_result.intensity = __pyx_state[1]; __pyx_result.mz = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PeakBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.__pyx_unpickle_PeakBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_PeakBase__set_state(<PeakBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PeakBase__set_state(PeakBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.area = __pyx_state[0]; __pyx_result.intensity = __pyx_state[1]; __pyx_result.mz = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_14ms_peak_picker_2_c_8peak_set___pyx_unpickle_PeakBase__set_state(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PeakBase__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PeakBase__set_state(PeakBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.area = __pyx_state[0]; __pyx_result.intensity = __pyx_state[1]; __pyx_result.mz = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->area = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->intensity = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->mz = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_PeakBase__set_state(PeakBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.area = __pyx_state[0]; __pyx_result.intensity = __pyx_state[1]; __pyx_result.mz = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_4 > 3) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 0);
  __pyx_t_3 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.area = __pyx_state[0]; __pyx_result.intensity = __pyx_state[1]; __pyx_result.mz = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_PeakBase__set_state(PeakBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.area = __pyx_state[0]; __pyx_result.intensity = __pyx_state[1]; __pyx_result.mz = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_PeakBase__set_state(<PeakBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PeakBase__set_state(PeakBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.area = __pyx_state[0]; __pyx_result.intensity = __pyx_state[1]; __pyx_result.mz = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("ms_peak_picker._c.peak_set.__pyx_unpickle_PeakBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakBase;

static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_freelist_14ms_peak_picker_2_c_8peak_set_PeakBase[1000000];
static int __pyx_freecount_14ms_peak_picker_2_c_8peak_set_PeakBase = 0;

static PyObject *__pyx_tp_new_14ms_peak_picker_2_c_8peak_set_PeakBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *p;
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_14ms_peak_picker_2_c_8peak_set_PeakBase > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase)) & ((t->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)) == 0))) {
    o = (PyObject*)__pyx_freelist_14ms_peak_picker_2_c_8peak_set_PeakBase[--__pyx_freecount_14ms_peak_picker_2_c_8peak_set_PeakBase];
    memset(o, 0, sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase));
    (void) PyObject_INIT(o, t);
  } else {
    if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  p = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)o);
  p->__pyx_vtab = __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase;
  return o;
}

static void __pyx_tp_dealloc_14ms_peak_picker_2_c_8peak_set_PeakBase(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_14ms_peak_picker_2_c_8peak_set_PeakBase < 1000000) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase)) & ((Py_TYPE(o)->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)) == 0))) {
    __pyx_freelist_14ms_peak_picker_2_c_8peak_set_PeakBase[__pyx_freecount_14ms_peak_picker_2_c_8peak_set_PeakBase++] = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyObject *__pyx_getprop_14ms_peak_picker_2_c_8peak_set_8PeakBase_mz(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_2mz_1__get__(o);
}

static int __pyx_setprop_14ms_peak_picker_2_c_8peak_set_8PeakBase_mz(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_2mz_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14ms_peak_picker_2_c_8peak_set_8PeakBase_intensity(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_9intensity_1__get__(o);
}

static int __pyx_setprop_14ms_peak_picker_2_c_8peak_set_8PeakBase_intensity(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_9intensity_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14ms_peak_picker_2_c_8peak_set_8PeakBase_area(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_4area_1__get__(o);
}

static int __pyx_setprop_14ms_peak_picker_2_c_8peak_set_8PeakBase_area(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_4area_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_14ms_peak_picker_2_c_8peak_set_PeakBase[] = {
  {"clone", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_1clone, METH_NOARGS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_8PeakBase_clone},
  {"copy", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_3copy, METH_NOARGS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_8PeakBase_2copy},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_5__reduce_cython__, METH_NOARGS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_8PeakBase_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_8PeakBase_7__setstate_cython__, METH_O, __pyx_doc_14ms_peak_picker_2_c_8peak_set_8PeakBase_6__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_14ms_peak_picker_2_c_8peak_set_PeakBase[] = {
  {(char *)"mz", __pyx_getprop_14ms_peak_picker_2_c_8peak_set_8PeakBase_mz, __pyx_setprop_14ms_peak_picker_2_c_8peak_set_8PeakBase_mz, (char *)"mz: 'double'", 0},
  {(char *)"intensity", __pyx_getprop_14ms_peak_picker_2_c_8peak_set_8PeakBase_intensity, __pyx_setprop_14ms_peak_picker_2_c_8peak_set_8PeakBase_intensity, (char *)"intensity: 'double'", 0},
  {(char *)"area", __pyx_getprop_14ms_peak_picker_2_c_8peak_set_8PeakBase_area, __pyx_setprop_14ms_peak_picker_2_c_8peak_set_8PeakBase_area, (char *)"area: 'double'", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_14ms_peak_picker_2_c_8peak_set_PeakBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "ms_peak_picker._c.peak_set.PeakBase", /*tp_name*/
  sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14ms_peak_picker_2_c_8peak_set_PeakBase, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_14ms_peak_picker_2_c_8peak_set_PeakBase, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_14ms_peak_picker_2_c_8peak_set_PeakBase, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14ms_peak_picker_2_c_8peak_set_PeakBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak __pyx_vtable_14ms_peak_picker_2_c_8peak_set_FittedPeak;

static PyObject *__pyx_tp_new_14ms_peak_picker_2_c_8peak_set_FittedPeak(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *p;
  PyObject *o = __pyx_tp_new_14ms_peak_picker_2_c_8peak_set_PeakBase(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase*)__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak;
  return o;
}

static PyObject *__pyx_getprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_signal_to_noise(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_15signal_to_noise_1__get__(o);
}

static int __pyx_setprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_signal_to_noise(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_15signal_to_noise_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_full_width_at_half_max(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_22full_width_at_half_max_1__get__(o);
}

static int __pyx_setprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_full_width_at_half_max(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_22full_width_at_half_max_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_left_width(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10left_width_1__get__(o);
}

static int __pyx_setprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_left_width(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10left_width_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_right_width(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_11right_width_1__get__(o);
}

static int __pyx_setprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_right_width(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_11right_width_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_peak_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10peak_count_1__get__(o);
}

static int __pyx_setprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_peak_count(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10peak_count_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_5index_1__get__(o);
}

static int __pyx_setprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_5index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_14ms_peak_picker_2_c_8peak_set_FittedPeak[] = {
  {"clone", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_3clone, METH_NOARGS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_10FittedPeak_2clone},
  {"__getstate__", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_7__getstate__, METH_NOARGS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_10FittedPeak_6__getstate__},
  {"__setstate__", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_9__setstate__, METH_O, __pyx_doc_14ms_peak_picker_2_c_8peak_set_10FittedPeak_8__setstate__},
  {"__reduce__", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_11__reduce__, METH_NOARGS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_10FittedPeak_10__reduce__},
  {"_eq", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_15_eq, METH_O, __pyx_doc_14ms_peak_picker_2_c_8peak_set_10FittedPeak_14_eq},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_14ms_peak_picker_2_c_8peak_set_FittedPeak[] = {
  {(char *)"signal_to_noise", __pyx_getprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_signal_to_noise, __pyx_setprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_signal_to_noise, (char *)"signal_to_noise: 'double'", 0},
  {(char *)"full_width_at_half_max", __pyx_getprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_full_width_at_half_max, __pyx_setprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_full_width_at_half_max, (char *)"full_width_at_half_max: 'double'", 0},
  {(char *)"left_width", __pyx_getprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_left_width, __pyx_setprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_left_width, (char *)"left_width: 'double'", 0},
  {(char *)"right_width", __pyx_getprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_right_width, __pyx_setprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_right_width, (char *)"right_width: 'double'", 0},
  {(char *)"peak_count", __pyx_getprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_peak_count, __pyx_setprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_peak_count, (char *)"peak_count: 'long'", 0},
  {(char *)"index", __pyx_getprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_index, __pyx_setprop_14ms_peak_picker_2_c_8peak_set_10FittedPeak_index, (char *)"index: 'long'", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_14ms_peak_picker_2_c_8peak_set_FittedPeak = {
  PyVarObject_HEAD_INIT(0, 0)
  "ms_peak_picker._c.peak_set.FittedPeak", /*tp_name*/
  sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14ms_peak_picker_2_c_8peak_set_PeakBase, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_13__hash__, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "FittedPeak(mz, intensity, signal_to_noise, peak_count, index, full_width_at_half_max, area, left_width=0, right_width=0)\nRepresent a single centroided mass spectral peak.\n\n    FittedPeak instances are comparable for equality and\n    hashable.\n\n    Attributes\n    ----------\n    mz : float\n        The m/z value at which the peak achieves its maximal\n        abundance\n    intensity : float\n        The apex height of the peak\n    signal_to_noise : float\n        The signal to noise ratio of the peak\n    full_width_at_half_max : double\n        The symmetric full width at half of the\n        peak's maximum height\n    index : int\n        The index at which the peak was found in the m/z array\n        it was picked from\n    peak_count : int\n        The order in which the peak was picked\n    area : float\n        The total area of the peak, as determined\n        by trapezoid integration\n    left_width : float\n        The left-sided width at half of max\n    right_width : float\n        The right-sided width at half of max\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_17__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_14ms_peak_picker_2_c_8peak_set_FittedPeak, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_14ms_peak_picker_2_c_8peak_set_FittedPeak, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_14ms_peak_picker_2_c_8peak_set_10FittedPeak_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14ms_peak_picker_2_c_8peak_set_FittedPeak, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSet;

static PyObject *__pyx_tp_new_14ms_peak_picker_2_c_8peak_set_PeakSet(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)o);
  p->__pyx_vtab = __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet;
  p->peaks = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_14ms_peak_picker_2_c_8peak_set_PeakSet(PyObject *o) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *p = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->peaks);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_14ms_peak_picker_2_c_8peak_set_PeakSet(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *p = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)o;
  if (p->peaks) {
    e = (*v)(p->peaks, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_14ms_peak_picker_2_c_8peak_set_PeakSet(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *p = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)o;
  tmp = ((PyObject*)p->peaks);
  p->peaks = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_14ms_peak_picker_2_c_8peak_set_PeakSet(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_getprop_14ms_peak_picker_2_c_8peak_set_7PeakSet_peaks(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks_1__get__(o);
}

static int __pyx_setprop_14ms_peak_picker_2_c_8peak_set_7PeakSet_peaks(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks_3__set__(o, v);
  }
  else {
    return __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_5peaks_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14ms_peak_picker_2_c_8peak_set_7PeakSet_indexed(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_7indexed_1__get__(o);
}

static int __pyx_setprop_14ms_peak_picker_2_c_8peak_set_7PeakSet_indexed(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_7indexed_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_14ms_peak_picker_2_c_8peak_set_PeakSet[] = {
  {"reindex", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_5reindex, METH_NOARGS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_4reindex},
  {"_index", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_7_index, METH_NOARGS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_6_index},
  {"has_peak", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_9has_peak, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_8has_peak},
  {"get_nearest_peak", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_11get_nearest_peak, METH_O, __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_10get_nearest_peak},
  {"clone", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_17clone, METH_NOARGS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_16clone},
  {"copy", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_19copy, METH_NOARGS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_18copy},
  {"between", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_21between, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_20between},
  {"all_peaks_for", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_23all_peaks_for, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_22all_peaks_for},
  {"__getstate__", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_25__getstate__, METH_NOARGS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_24__getstate__},
  {"__setstate__", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_27__setstate__, METH_O, __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_26__setstate__},
  {"__reduce__", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_29__reduce__, METH_NOARGS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_28__reduce__},
  {"_eq", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_31_eq, METH_O, __pyx_doc_14ms_peak_picker_2_c_8peak_set_7PeakSet_30_eq},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_14ms_peak_picker_2_c_8peak_set_PeakSet[] = {
  {(char *)"peaks", __pyx_getprop_14ms_peak_picker_2_c_8peak_set_7PeakSet_peaks, __pyx_setprop_14ms_peak_picker_2_c_8peak_set_7PeakSet_peaks, (char *)"peaks: tuple", 0},
  {(char *)"indexed", __pyx_getprop_14ms_peak_picker_2_c_8peak_set_7PeakSet_indexed, __pyx_setprop_14ms_peak_picker_2_c_8peak_set_7PeakSet_indexed, (char *)"indexed: 'bool'", 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_PeakSet = {
  __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_14ms_peak_picker_2_c_8peak_set_PeakSet, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_PeakSet = {
  __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_3__len__, /*mp_length*/
  __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_15__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSet = {
  PyVarObject_HEAD_INIT(0, 0)
  "ms_peak_picker._c.peak_set.PeakSet", /*tp_name*/
  sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14ms_peak_picker_2_c_8peak_set_PeakSet, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_13__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_PeakSet, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_PeakSet, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "PeakSet(peaks)\nA sequence of :class:`FittedPeak` instances, ordered by m/z,\n    providing efficient search and retrieval of individual peaks or\n    whole intervals of the m/z domain.\n\n    This collection is not meant to be updated once created, as it\n    it is indexed for ease of connecting individual peak objects to\n    their position in the underlying sequence.\n\n    One :class:`PeakSet` is considered equal to another if all of their\n    contained :class:`FittedPeak` members are equal to each other\n\n    Attributes\n    ----------\n    peaks : tuple\n        The :class:`FittedPeak` instances, stored\n    ", /*tp_doc*/
  __pyx_tp_traverse_14ms_peak_picker_2_c_8peak_set_PeakSet, /*tp_traverse*/
  __pyx_tp_clear_14ms_peak_picker_2_c_8peak_set_PeakSet, /*tp_clear*/
  __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_33__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_14ms_peak_picker_2_c_8peak_set_PeakSet, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_14ms_peak_picker_2_c_8peak_set_PeakSet, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14ms_peak_picker_2_c_8peak_set_PeakSet, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;

static PyObject *__pyx_tp_new_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *p;
  PyObject *o = __pyx_tp_new_14ms_peak_picker_2_c_8peak_set_PeakSet(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet*)__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;
  return o;
}

static void __pyx_tp_dealloc_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_14ms_peak_picker_2_c_8peak_set_PeakSet(o);
}

static PyMethodDef __pyx_methods_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed[] = {
  {"clone", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_5clone, METH_NOARGS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_4clone},
  {"_allocate_index", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_7_allocate_index, METH_O, __pyx_doc_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_6_allocate_index},
  {"_index", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_9_index, METH_NOARGS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_8_index},
  {"get_nearest_peak", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_11get_nearest_peak, METH_O, __pyx_doc_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_10get_nearest_peak},
  {"has_peak_hinted", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_13has_peak_hinted, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_12has_peak_hinted},
  {"test_interval", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_15test_interval, METH_O, __pyx_doc_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_14test_interval},
  {"find_interval_for", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_17find_interval_for, METH_O, __pyx_doc_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_16find_interval_for},
  {"check_interval", (PyCFunction)__pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_19check_interval, METH_O, __pyx_doc_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_18check_interval},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = {
  PyVarObject_HEAD_INIT(0, 0)
  "ms_peak_picker._c.peak_set.PeakSetIndexed", /*tp_name*/
  sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_14ms_peak_picker_2_c_8peak_set_7PeakSet_13__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "PeakSetIndexed(peaks)", /*tp_doc*/
  __pyx_tp_traverse_14ms_peak_picker_2_c_8peak_set_PeakSet, /*tp_traverse*/
  __pyx_tp_clear_14ms_peak_picker_2_c_8peak_set_PeakSet, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"binary_search_ppm_error", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14ms_peak_picker_2_c_8peak_set_1binary_search_ppm_error, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14ms_peak_picker_2_c_8peak_set_binary_search_ppm_error},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_peak_set(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_peak_set},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "peak_set",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_FittedPeak, __pyx_k_FittedPeak, sizeof(__pyx_k_FittedPeak), 0, 0, 1, 1},
  {&__pyx_kp_s_FittedPeak_mz_0_3f_intensity_0_3, __pyx_k_FittedPeak_mz_0_3f_intensity_0_3, sizeof(__pyx_k_FittedPeak_mz_0_3f_intensity_0_3), 0, 0, 1, 0},
  {&__pyx_kp_s_Hinted_start_index_d_cannot_be_g, __pyx_k_Hinted_start_index_d_cannot_be_g, sizeof(__pyx_k_Hinted_start_index_d_cannot_be_g), 0, 0, 1, 0},
  {&__pyx_n_s_INTERVAL_INDEX_SIZE, __pyx_k_INTERVAL_INDEX_SIZE, sizeof(__pyx_k_INTERVAL_INDEX_SIZE), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
  {&__pyx_n_s_NotImplemented, __pyx_k_NotImplemented, sizeof(__pyx_k_NotImplemented), 0, 0, 1, 1},
  {&__pyx_n_s_PeakBase, __pyx_k_PeakBase, sizeof(__pyx_k_PeakBase), 0, 0, 1, 1},
  {&__pyx_n_s_PeakSet, __pyx_k_PeakSet, sizeof(__pyx_k_PeakSet), 0, 0, 1, 1},
  {&__pyx_n_s_PeakSetIndexed, __pyx_k_PeakSetIndexed, sizeof(__pyx_k_PeakSetIndexed), 0, 0, 1, 1},
  {&__pyx_kp_s_PeakSet_d_Peaks, __pyx_k_PeakSet_d_Peaks, sizeof(__pyx_k_PeakSet_d_Peaks), 0, 0, 1, 0},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_all_peaks_for, __pyx_k_all_peaks_for, sizeof(__pyx_k_all_peaks_for), 0, 0, 1, 1},
  {&__pyx_n_s_allocate_index, __pyx_k_allocate_index, sizeof(__pyx_k_allocate_index), 0, 0, 1, 1},
  {&__pyx_n_s_area, __pyx_k_area, sizeof(__pyx_k_area), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_attrgetter, __pyx_k_attrgetter, sizeof(__pyx_k_attrgetter), 0, 0, 1, 1},
  {&__pyx_n_s_center_value, __pyx_k_center_value, sizeof(__pyx_k_center_value), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_clone, __pyx_k_clone, sizeof(__pyx_k_clone), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_eq, __pyx_k_eq, sizeof(__pyx_k_eq), 0, 0, 1, 1},
  {&__pyx_n_s_error_tolerance, __pyx_k_error_tolerance, sizeof(__pyx_k_error_tolerance), 0, 0, 1, 1},
  {&__pyx_n_s_full_width_at_half_max, __pyx_k_full_width_at_half_max, sizeof(__pyx_k_full_width_at_half_max), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_has_peak_hinted, __pyx_k_has_peak_hinted, sizeof(__pyx_k_has_peak_hinted), 0, 0, 1, 1},
  {&__pyx_n_s_hint_end, __pyx_k_hint_end, sizeof(__pyx_k_hint_end), 0, 0, 1, 1},
  {&__pyx_n_s_hint_start, __pyx_k_hint_start, sizeof(__pyx_k_hint_start), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
  {&__pyx_n_s_index_2, __pyx_k_index_2, sizeof(__pyx_k_index_2), 0, 0, 1, 1},
  {&__pyx_n_s_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_intensity, __pyx_k_intensity, sizeof(__pyx_k_intensity), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_left_width, __pyx_k_left_width, sizeof(__pyx_k_left_width), 0, 0, 1, 1},
  {&__pyx_n_s_m1, __pyx_k_m1, sizeof(__pyx_k_m1), 0, 0, 1, 1},
  {&__pyx_n_s_m2, __pyx_k_m2, sizeof(__pyx_k_m2), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_ms_peak_picker__c_peak_set, __pyx_k_ms_peak_picker__c_peak_set, sizeof(__pyx_k_ms_peak_picker__c_peak_set), 0, 0, 1, 1},
  {&__pyx_n_s_mz, __pyx_k_mz, sizeof(__pyx_k_mz), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_operator, __pyx_k_operator, sizeof(__pyx_k_operator), 0, 0, 1, 1},
  {&__pyx_n_s_peak_count, __pyx_k_peak_count, sizeof(__pyx_k_peak_count), 0, 0, 1, 1},
  {&__pyx_n_s_peaks, __pyx_k_peaks, sizeof(__pyx_k_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_capi, __pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_PeakBase, __pyx_k_pyx_unpickle_PeakBase, sizeof(__pyx_k_pyx_unpickle_PeakBase), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_reindex, __pyx_k_reindex, sizeof(__pyx_k_reindex), 0, 0, 1, 1},
  {&__pyx_n_s_right_width, __pyx_k_right_width, sizeof(__pyx_k_right_width), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_signal_to_noise, __pyx_k_signal_to_noise, sizeof(__pyx_k_signal_to_noise), 0, 0, 1, 1},
  {&__pyx_n_s_sorted, __pyx_k_sorted, sizeof(__pyx_k_sorted), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_tolerance, __pyx_k_tolerance, sizeof(__pyx_k_tolerance), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_sorted = __Pyx_GetBuiltinName(__pyx_n_s_sorted); if (!__pyx_builtin_sorted) __PYX_ERR(0, 186, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 189, __pyx_L1_error)
  __pyx_builtin_NotImplemented = __Pyx_GetBuiltinName(__pyx_n_s_NotImplemented); if (!__pyx_builtin_NotImplemented) __PYX_ERR(0, 390, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 800, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "ms_peak_picker/_c/peak_set.pyx":818
 *             size_t i
 *         if self.interval_index == NULL:
 *             return None, None, None             # <<<<<<<<<<<<<<
 *         status = find_search_interval(self.interval_index, value, &start, &end)
 *         return start, end, interpolate_index(self.interval_index, value)
 */
  __pyx_tuple_ = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x4f329c4, 0x9e542aa, 0x4c1f989):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x4f329c4, 0x9e542aa, 0x4c1f989) = (area, intensity, mz))" % __pyx_checksum)
 */
  __pyx_tuple__2 = PyTuple_Pack(3, __pyx_int_83044804, __pyx_int_166019754, __pyx_int_79821193); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "ms_peak_picker/_c/peak_set.pyx":410
 * 
 * cdef FittedPeak _null_peak
 * _null_peak = FittedPeak(-1,-1,-1,-1,-1,-1, -1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__3 = PyTuple_Pack(7, __pyx_int_neg_1, __pyx_int_neg_1, __pyx_int_neg_1, __pyx_int_neg_1, __pyx_int_neg_1, __pyx_int_neg_1, __pyx_int_neg_1); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "(tree fragment)":1
 * def __pyx_unpickle_PeakBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__4 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);
  __pyx_codeobj__5 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__4, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_PeakBase, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__5)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_79821193 = PyInt_FromLong(79821193L); if (unlikely(!__pyx_int_79821193)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_83044804 = PyInt_FromLong(83044804L); if (unlikely(!__pyx_int_83044804)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_166019754 = PyInt_FromLong(166019754L); if (unlikely(!__pyx_int_166019754)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_n_s_INTERVAL_INDEX_SIZE, (void *)&__pyx_v_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("binary_search_ppm_error", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_8peak_set_binary_search_ppm_error, "struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(PyObject *, double, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_binary_search_ppm_error", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_8peak_set__binary_search_ppm_error, "struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(PyObject *, double, size_t, size_t, double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_binary_search_nearest_match", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_8peak_set__binary_search_nearest_match, "struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(PyObject *, double, size_t, size_t, double *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("check_index", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_8peak_set_check_index, "int (struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("double_binary_search_ppm", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_8peak_set_double_binary_search_ppm, "size_t (double *, double, double, size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("double_binary_search_nearest_match", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_8peak_set_double_binary_search_nearest_match, "size_t (double *, double, size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("double_binary_search_ppm_with_hint", (void (*)(void))__pyx_f_14ms_peak_picker_2_c_8peak_set_double_binary_search_ppm_with_hint, "size_t (double *, double, double, size_t, size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase = &__pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakBase;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakBase.clone = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_8peak_set_8PeakBase_clone;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakBase.copy = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_8peak_set_8PeakBase_copy;
  if (PyType_Ready(&__pyx_type_14ms_peak_picker_2_c_8peak_set_PeakBase) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14ms_peak_picker_2_c_8peak_set_PeakBase.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14ms_peak_picker_2_c_8peak_set_PeakBase.tp_dictoffset && __pyx_type_14ms_peak_picker_2_c_8peak_set_PeakBase.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14ms_peak_picker_2_c_8peak_set_PeakBase.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_14ms_peak_picker_2_c_8peak_set_PeakBase.tp_dict, __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PeakBase, (PyObject *)&__pyx_type_14ms_peak_picker_2_c_8peak_set_PeakBase) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_14ms_peak_picker_2_c_8peak_set_PeakBase) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = &__pyx_type_14ms_peak_picker_2_c_8peak_set_PeakBase;
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak = &__pyx_vtable_14ms_peak_picker_2_c_8peak_set_FittedPeak;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_FittedPeak.__pyx_base = *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_FittedPeak.__pyx_base.clone = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_8peak_set_10FittedPeak_clone;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_FittedPeak._eq = (int (*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_8peak_set_10FittedPeak__eq;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_FittedPeak._create = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*)(double, double, double, double, double, double, long, long, double))__pyx_f_14ms_peak_picker_2_c_8peak_set_10FittedPeak__create;
  __pyx_type_14ms_peak_picker_2_c_8peak_set_FittedPeak.tp_base = __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase;
  if (PyType_Ready(&__pyx_type_14ms_peak_picker_2_c_8peak_set_FittedPeak) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14ms_peak_picker_2_c_8peak_set_FittedPeak.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14ms_peak_picker_2_c_8peak_set_FittedPeak.tp_dictoffset && __pyx_type_14ms_peak_picker_2_c_8peak_set_FittedPeak.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14ms_peak_picker_2_c_8peak_set_FittedPeak.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_14ms_peak_picker_2_c_8peak_set_FittedPeak.tp_dict, __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FittedPeak, (PyObject *)&__pyx_type_14ms_peak_picker_2_c_8peak_set_FittedPeak) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = &__pyx_type_14ms_peak_picker_2_c_8peak_set_FittedPeak;
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet = &__pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSet;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSet._create = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*)(PyObject *))__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__create;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSet.reindex = (PyObject *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_reindex;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSet._index = (size_t (*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__index;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSet._get_nearest_peak = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double *))__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__get_nearest_peak;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSet._has_peak = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak *__pyx_optional_args))__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSet.clone = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_clone;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSet.copy = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_copy;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSet._between = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double))__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__between;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSet._between_bounds = (int (*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double, size_t *, size_t *))__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__between_bounds;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSet.all_peaks_for = (PyObject *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for *__pyx_optional_args))__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSet.getitem = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, size_t))__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_getitem;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSet.get_size = (size_t (*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *))__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet_get_size;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSet._eq = (int (*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_8peak_set_7PeakSet__eq;
  if (PyType_Ready(&__pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSet) < 0) __PYX_ERR(0, 140, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSet.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSet.tp_dictoffset && __pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSet.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSet.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSet.tp_dict, __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet) < 0) __PYX_ERR(0, 140, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PeakSet, (PyObject *)&__pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSet) < 0) __PYX_ERR(0, 140, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = &__pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSet;
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = &__pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed.__pyx_base = *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed.__pyx_base._create = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*)(PyObject *))__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__create;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed.__pyx_base._index = (size_t (*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__index;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed.__pyx_base._get_nearest_peak = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double *))__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__get_nearest_peak;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed.__pyx_base._has_peak = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak *__pyx_optional_args))__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__has_peak;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed.__pyx_base.clone = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_clone;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed._allocate_index = (PyObject *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, size_t, int __pyx_skip_dispatch))__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed__allocate_index;
  __pyx_vtable_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed.has_peak_hinted = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted *__pyx_optional_args))__pyx_f_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted;
  __pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed.tp_base = __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet;
  if (PyType_Ready(&__pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed) < 0) __PYX_ERR(0, 685, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed.tp_dictoffset && __pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed.tp_dict, __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed) < 0) __PYX_ERR(0, 685, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PeakSetIndexed, (PyObject *)&__pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed) < 0) __PYX_ERR(0, 685, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = &__pyx_type_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(4, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(5, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initpeak_set(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initpeak_set(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_peak_set(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_peak_set(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_peak_set(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'peak_set' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_peak_set(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("peak_set", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_ms_peak_picker___c__peak_set) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "ms_peak_picker._c.peak_set")) {
      if (unlikely(PyDict_SetItemString(modules, "ms_peak_picker._c.peak_set", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  if (unlikely(__Pyx_modinit_variable_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "ms_peak_picker/_c/peak_set.pyx":3
 * # cython: embedsignature=True
 * cimport cython
 * import operator             # <<<<<<<<<<<<<<
 * from libc.stdlib cimport malloc, realloc, free
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_operator, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_operator, __pyx_t_1) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":15
 * 
 * 
 * cdef double INF = float('inf')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_AsDouble(__pyx_n_s_inf); if (unlikely(__pyx_t_2 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 15, __pyx_L1_error)
  __pyx_v_14ms_peak_picker_2_c_8peak_set_INF = __pyx_t_2;

  /* "ms_peak_picker/_c/peak_set.pyx":410
 * 
 * cdef FittedPeak _null_peak
 * _null_peak = FittedPeak(-1,-1,-1,-1,-1,-1, -1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak), __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XGOTREF(((PyObject *)__pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak));
  __Pyx_DECREF_SET(__pyx_v_14ms_peak_picker_2_c_8peak_set__null_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":683
 * 
 * 
 * cdef size_t INTERVAL_INDEX_SIZE = 0             # <<<<<<<<<<<<<<
 * 
 * cdef class PeakSetIndexed(PeakSet):
 */
  __pyx_v_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PeakBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_14ms_peak_picker_2_c_8peak_set_3__pyx_unpickle_PeakBase, NULL, __pyx_n_s_ms_peak_picker__c_peak_set); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_PeakBase, __pyx_t_1) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_peak_picker/_c/peak_set.pyx":1
 * # cython: embedsignature=True             # <<<<<<<<<<<<<<
 * cimport cython
 * import operator
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init ms_peak_picker._c.peak_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init ms_peak_picker._c.peak_set");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* pyobject_as_double */
static double __Pyx__PyObject_AsDouble(PyObject* obj) {
    PyObject* float_value;
#if !CYTHON_USE_TYPE_SLOTS
    float_value = PyNumber_Float(obj);  if ((0)) goto bad;
#else
    PyNumberMethods *nb = Py_TYPE(obj)->tp_as_number;
    if (likely(nb) && likely(nb->nb_float)) {
        float_value = nb->nb_float(obj);
        if (likely(float_value) && unlikely(!PyFloat_Check(float_value))) {
            PyErr_Format(PyExc_TypeError,
                "__float__ returned non-float (type %.200s)",
                Py_TYPE(float_value)->tp_name);
            Py_DECREF(float_value);
            goto bad;
        }
    } else if (PyUnicode_CheckExact(obj) || PyBytes_CheckExact(obj)) {
#if PY_MAJOR_VERSION >= 3
        float_value = PyFloat_FromString(obj);
#else
        float_value = PyFloat_FromString(obj, 0);
#endif
    } else {
        PyObject* args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        PyTuple_SET_ITEM(args, 0, obj);
        float_value = PyObject_Call((PyObject*)&PyFloat_Type, args, 0);
        PyTuple_SET_ITEM(args, 0, 0);
        Py_DECREF(args);
    }
#endif
    if (likely(float_value)) {
        double value = PyFloat_AS_DOUBLE(float_value);
        Py_DECREF(float_value);
        return value;
    }
bad:
    return (double)-1;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

static PyObject* __pyx_convert__to_py_struct____pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell(struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell s) {
  PyObject* res;
  PyObject* member;
  res = __Pyx_PyDict_NewPresized(3); if (unlikely(!res)) return NULL;
  member = PyFloat_FromDouble(s.center_value); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_center_value, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __Pyx_PyInt_FromSize_t(s.start); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_start, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __Pyx_PyInt_FromSize_t(s.end); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_end, member) < 0)) goto bad;
  Py_DECREF(member);
  return res;
  bad:
  Py_XDECREF(member);
  Py_DECREF(res);
  return NULL;
}
/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* VoidPtrExport */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    d = PyDict_GetItem(__pyx_d, __pyx_n_s_pyx_capi);
    Py_XINCREF(d);
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_n_s_pyx_capi, d) < 0)
            goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
